(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DP_Framework
imports
  Term_Order
  Trs_Impl
begin

text {*
A \emph{dependency pair problem} (DP problem, for short) is a tuple consisting
of two rewrite systems.
*} 
type_synonym
  ('f, 'v) dpp = "('f, 'v) trs \<times> ('f, 'v) trs"

definition minimal_cond
where "minimal_cond R s t \<sigma> \<equiv> \<forall>i. SN_elt (rstep R) (t i \<cdot> \<sigma> i)"

definition ideriv :: "('f,'v)trs \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> bool"
where "ideriv R t \<equiv> \<forall> i. (t i, t (Suc i)) \<in> rstep R"

fun ichain ::
 "('f,'v)dpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)subst) \<Rightarrow> bool"
where "ichain (P,R) s t \<sigma> = (
 (\<forall>i. ((s i,t i) \<in> P)) \<and> 
 (\<forall>i. ((t i)\<cdot>(\<sigma> i),(s(Suc i))\<cdot>(\<sigma> (Suc i))) \<in> (rstep R)^*) 
)"

text {* A \emph{minimal infinite chain} is an infinite chain where additionally all @{text \<sigma>}-instances
of terms in the sequence~@{text t} are terminating w.r.t.~the system~@{text R}. *}
fun min_ichain ::
 "('f,'v)dpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)subst) \<Rightarrow> bool"
where "min_ichain (P,R) s t \<sigma> = (ichain (P,R) s t \<sigma> \<and> minimal_cond R s t \<sigma>)"

lemma ichain_imp_map_ichain:
  assumes chain: "ichain (P,R) s t \<sigma>"
  shows "ichain (map_trs fg P, map_trs fg R) (\<lambda> i. map_term fg (s i)) (\<lambda> i. map_term fg (t i)) (\<lambda> i. map_subst fg (\<sigma> i))"
  (is "ichain (?lP, ?lR) ?ls ?lt ?lsig")
proof -
  have mem: "\<forall> i. (?ls i, ?lt i) \<in> ?lP"
  proof
    fix i
    from chain have "(s i, t i) \<in> P" by auto
    thus "(?ls i, ?lt i) \<in> ?lP" by (force simp: map_trs.simps)
  qed
  have steps: "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*"
  proof
    fix i
    from chain have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by auto
    hence "(map_term fg (t i \<cdot> \<sigma> i), map_term fg (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (rstep (map_trs fg R))^*" by (rule rsteps_imp_map_rsteps)
    thus "(?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*" by auto
  qed
  from mem steps show ?thesis by auto
qed


text {* An auxiliary function to stack contexts (and substitutions) in order to reconstruct
DP steps in the original TRS. *}
fun D :: "(nat \<Rightarrow> ('f,'v)ctxt) \<Rightarrow> (nat \<Rightarrow> ('f,'v)subst) \<Rightarrow> nat \<Rightarrow> ('f,'v)ctxt"
where "D C \<sigma> 0 = \<box>"
    | "D C \<sigma> (Suc i) = (D C \<sigma> i) \<circ>c ((C i) \<cdot>c (\<sigma> i))"

lemma D_clean_subst_wf:
  fixes F::"'f sig" and f::"('g, 'v) subst \<Rightarrow> ('f, 'v) subst"
  defines "cs \<equiv> \<lambda>\<sigma>. clean_subst F (f \<sigma>)"
  assumes "\<forall>i. C i \<in> wf_ctxts F" shows "(D C (\<lambda>i. cs(\<sigma> i)) i) \<in> wf_ctxts F"
proof (induct i)
  case 0 show ?case by auto
next
  case (Suc i)
  obtain s where s: "\<sigma> i = s" by auto
  from assms have "C i \<in> wf_ctxts F" by simp
  from subst_apply_ctxt_clean_subst_wf_ctxts[OF this]
    have "(C i \<cdot>c (cs(\<sigma> i))) \<in> wf_ctxts F" unfolding cs_def s o_assoc by simp
  with Suc show ?case by simp
qed

lemma [simp]: "(D C \<sigma> i)\<langle>(C i \<cdot>c \<sigma> i)\<langle>t\<rangle>\<rangle> = (D C \<sigma> (Suc i))\<langle>t\<rangle>" by (cases i) simp_all

definition
  sharp_sig :: "('f \<Rightarrow> 'f) \<Rightarrow> 'f sig \<Rightarrow> 'f sig"
where
  "sharp_sig shp F \<equiv> F \<union> {(shp f, n) | f n. (f, n) \<in> F}"

lemma lhs_DP_wf:
  fixes R::"('f, 'v) trs" and shp :: "'f \<Rightarrow> 'f"
  defines "F \<equiv> sharp_sig shp (funas_trs R)"
  assumes DP: "(s,t) \<in> DP shp R" shows "s \<in> wf_terms F"
proof -
  have subset: "funas_trs(DP shp R) \<subseteq> F" using funas_DP_subset[of shp]
    unfolding F_def sharp_sig_def by auto
  have "funas_term s \<subseteq> funas_trs(DP shp R)" using DP unfolding funas_trs_def funas_rule_def by auto
  hence "wf_terms (funas_term s) \<subseteq> wf_terms (funas_trs(DP shp R))" by (rule wf_terms_subset)
  moreover have "s \<in> wf_terms (funas_term s)" by simp
  ultimately have "s \<in> wf_terms (funas_trs(DP shp R))" by auto
  moreover from subset have "wf_terms (funas_trs(DP shp R)) \<subseteq> wf_terms F" by (rule wf_terms_subset)
  ultimately show ?thesis by auto
qed



lemma the_inv_id: "the_inv id = id"
proof (intro ext)
  fix x
  have "the_inv id x = the_inv id (id x)" by simp
  also have "\<dots> = x" by (rule the_inv_f_f, unfold inj_on_def, simp)
  finally show "the_inv id x = id x" by simp
qed

lemma [simp]: "sharp_term id t = t" by (cases t, auto)
lemma [simp]: "sharp_term (the_inv id) t = t" 
proof (cases t, simp)
  case (Fun f ss)
  show ?thesis unfolding Fun  
    by (simp add: the_inv_id)
qed

lemma DP_not_is_var_t: "(s,t) \<in> DP shp R \<Longrightarrow> \<not> is_var t"
  unfolding DP_def by auto

lemma DP_not_is_var_s: "wf_trs R \<Longrightarrow> (s,t) \<in> DP shp R \<Longrightarrow> \<not> is_var s"
  unfolding DP_def wf_trs_def by force

lemma sharp_term_subst_distr_2: "\<not> is_var t \<Longrightarrow> sharp_term shp t \<cdot> \<sigma> = sharp_term shp (t \<cdot> \<sigma>)"
  using sharp_term_subst_distr by (cases t, auto)

(* TODO: perhaps specialize to arbitrary shp-functions instead of id *)
  
theorem clean_ichain_imp_not_SN_wfrstep:
  fixes R::"('f, 'v) trs" and shp :: "'f \<Rightarrow> 'f"
  defines "F \<equiv> sharp_sig shp (funas_trs R)"
  defines "cs \<equiv> clean_subst F"
  assumes chain: "ichain (DP shp R, R) s t (\<lambda>i. cs (\<sigma> i))"
  and inj: "inj shp"
  and id: "shp = id"
  shows "\<not> SN (wfrstep R)"
proof (cases "wf_trs R")
  case False
  with SN_wfrstep_imp_wf_trs show ?thesis by blast
next
  case True
  from chain have P_seq: "\<And> i. (s i, t i) \<in> DP shp R"
    and seq: "\<forall>i. ((t i) \<cdot> cs (\<sigma> i), (s (Suc i)) \<cdot> cs (\<sigma> (Suc i))) \<in> (rstep R)^*"
    by auto
  let ?s = "sharp_term shp"
  let ?us = "sharp_term (the_inv shp)"
  let ?cus = "clean_subst (funas_trs R)"
  have "\<forall>i. \<exists>C. C \<in> wf_ctxts (funas_trs R) \<and> (?us(s i), C\<langle>?us(t i)\<rangle>) \<in> R"
  proof
    fix i from P_seq have "(s i, t i) \<in> DP shp R" by simp
    from DP_step_in_R[OF this inj]  
    show "\<exists>C. C \<in> wf_ctxts (funas_trs R) \<and> (?us (s i), C\<langle>?us(t i)\<rangle>) \<in> R" . 
  qed
  from choice[OF this] obtain C
    where "\<forall>i. C i \<in> wf_ctxts (funas_trs R) \<and> (?us (s i), (C i)\<langle>?us(t i)\<rangle>) \<in> R" by best
  hence wf_ctxt: "\<forall>i. C i \<in> wf_ctxts (funas_trs R)"
    and R: "\<forall>i. (?us(s i), (C i)\<langle>?us(t i)\<rangle>) \<in> R" by auto
  from R have root_steps:
    "\<forall>i. (?us(s i) \<cdot> ?cus (\<sigma> i), (C i \<cdot>c ?cus (\<sigma> i))\<langle>?us(t i) \<cdot> ?cus (\<sigma> i)\<rangle>) \<in> rstep R"
    using rstep.id[THEN rstep.subst, of _ _ R] by force
  let ?C = "D C (\<lambda>i. ?cus (\<sigma> i))"
  let ?s = "\<lambda>i. (?C i)\<langle>?us(s i) \<cdot> ?cus (\<sigma> i)\<rangle>"
  let ?t = "\<lambda>i. (?C (Suc i))\<langle>?us(t i) \<cdot> ?cus (\<sigma> i)\<rangle>"
  { 
    fix i
    note sharp_term_subst_distr_2[OF DP_not_is_var_s[OF True P_seq[of i]]]
  } note si = this
  { 
    fix i
    note sharp_term_subst_distr_2[OF DP_not_is_var_t[OF P_seq[of i]]]
  } note ti = this
  have ctxt_steps: "\<forall>i. (?s i, ?t i) \<in> rstep R" using root_steps by auto
  have
    aux: "\<forall>i.
    ( (?us(t i) \<cdot> cs (\<sigma> i)), (?us(s (Suc i)) \<cdot> cs (\<sigma> (Suc i)))) \<in> (rstep R)^*"
    (is "\<forall>i. ?Q i") 
  proof 
    fix i
    show  "?Q i" unfolding si ti
      using seq[THEN spec[of _i]] unfolding id by simp
  qed
  have "\<forall>i. (?us(t i) \<cdot> ?cus (\<sigma> i), ?us(s (Suc i)) \<cdot> ?cus (\<sigma> (Suc i))) \<in> (rstep R)^*"
    (is "\<forall>i. ?P i")
  proof
    fix i
    from aux have "?Q i" by simp
    thus "?P i"
      unfolding cs_def F_def map_subst_distrib o_def  sharp_sig_def  id by simp 
  qed
  hence steps: "\<forall>i. (?t i, ?s (Suc i)) \<in> (rstep R)^*" using rsteps_closed_ctxt by best
  have big_step: "\<forall>i. (?s i, ?s(Suc i)) \<in> (rstep R)^+"
  proof
    fix i
    from ctxt_steps have "(?s i, ?t i) \<in> rstep R" by best
    moreover from steps have "(?t i, ?s (Suc i)) \<in> (rstep R)^*" by fast
    ultimately show "(?s i, ?s (Suc i)) \<in> (rstep R)^+" by (rule rtrancl_into_trancl2)
  qed
  from D_clean_subst_wf[OF wf_ctxt] have "\<forall>i. ?C i \<in> wf_ctxts (funas_trs R)"
    unfolding clean_subst.simps o_assoc by auto
  moreover have "\<forall>i. ?us(s i) \<cdot> ?cus (\<sigma> i) \<in> wf_terms (funas_trs R)"
  proof
    fix i
    from lhs_DP_wf and P_seq have "s i \<in> wf_terms F" unfolding F_def by best
    from this[unfolded F_def sharp_sig_def]
      have "?us(s i) \<in> wf_terms (funas_trs R)" unfolding id by simp
    from subst_apply_term_clean_wf_terms[OF this,of "\<sigma> i"]
      show "?us(s i) \<cdot> ?cus(\<sigma> i) \<in> wf_terms (funas_trs R)" by (simp add: subst o_assoc o_def)
  qed
  ultimately have "\<forall>i. ?s i \<in> wf_terms (funas_trs R)" using ctxt_apply_wf by best
  with big_step
    have seq: "\<forall>i. (?s i, ?s(Suc i)) \<in> {(x, y) | x y. (x, y) \<in> (rstep R)^+ \<and> x \<in> wf_terms (funas_trs R) \<and> y \<in> wf_terms (funas_trs R)}"
    (is "\<forall>i. _ \<in> ?steps") by blast
  have"?s 0 = ?s 0" by simp
  with seq have "?s 0 = ?s 0 \<and> (\<forall>i. (?s i, ?s (Suc i)) \<in> ?steps)" by best
  hence "\<exists>S. S 0 = ?s 0 \<and> (\<forall>i. (S i, S (Suc i)) \<in> ?steps)" by best
  hence "\<exists>x S. S 0 = x \<and> (\<forall>i. (S i, S (Suc i)) \<in> ?steps)" by auto
  hence "\<not> SN ?steps" unfolding SN_defs by blast
  hence "\<not> SN ((wfrstep R)^+)" unfolding wfrstep_trancl[OF True] .
  with SN_trancl_SN_conv show ?thesis by auto
qed

hide_const D


fun
  funas_dpp :: "('f,'v) dpp \<Rightarrow> 'f sig"
where
  "funas_dpp (P, R) = funas_trs P \<union> funas_trs R"

lemma ichain_imp_clean_ichain:
  fixes P R::"('f,'v)trs" and F::"'f sig"
  defines "c \<equiv> \<lambda>t::('f, 'v) term. clean_term F t"
  defines "cs \<equiv> \<lambda>\<sigma>::('f, 'v) subst. clean_subst F \<sigma>"
  assumes subset: "funas_dpp (P, R) \<subseteq> F"
    and ichain: "ichain (P,R) s t \<sigma>"
  shows "ichain (P,R) s t (\<lambda>i. cs(\<sigma> i))"
proof -
  from ichain have dpstep: "\<forall>i. (s i,t i) \<in> P"
    and seq: "\<forall>i. (t i \<cdot> \<sigma> i,s(Suc i)\<cdot>\<sigma>(Suc i)) \<in> (rstep R)^*" by auto
  from subset have subset_P: "funas_trs P \<subseteq> F" by simp
  from subset have subset_R: "funas_trs R \<subseteq> F" by simp
  have wf_s: "\<forall>i. s i \<in> wf_terms F"
  proof
    fix i
    from dpstep have "(s i,t i) \<in> P" by simp
    from lhs_wf[OF this subset_P] show "s i \<in> wf_terms F" .
  qed
  have wf_t: "\<forall>i. t i \<in> wf_terms F"
  proof
    fix i
    from dpstep have "(s i,t i) \<in> P" by simp
    from rhs_wf[OF this subset_P] show "t i \<in> wf_terms F" .
  qed
  have seq': "\<forall>i. (c(t i\<cdot>(\<sigma> i)),c(s(Suc i)\<cdot>(\<sigma>(Suc i)))) \<in> (rstep R)^*" (is "\<forall>i. (c(?t i),c(?s i)) \<in> _")
    using seq rsteps_imp_clean_rsteps[OF subset_R] unfolding c_def by blast
  have "\<forall>i. (t i \<cdot> cs(\<sigma> i),s(Suc i) \<cdot> cs(\<sigma>(Suc i))) \<in> (rstep R)^*"
  proof
    fix i
    have cts: "c(t i \<cdot> \<sigma> i) = t i \<cdot> cs(\<sigma> i)"
      using clean_subst_apply_term[OF wf_t[THEN spec[of _ i]]]
      unfolding c_def cs_def o_def by auto
    have css: "c(s(Suc i) \<cdot> \<sigma>(Suc i)) = s(Suc i) \<cdot> cs(\<sigma>(Suc i))"
      using clean_subst_apply_term[OF wf_s[THEN spec[of _ "Suc i"]]]
      unfolding c_def cs_def o_def by auto
    from seq'[THEN spec[of _ i]]
      show "(t i \<cdot> cs(\<sigma> i),s(Suc i) \<cdot> cs(\<sigma>(Suc i))) \<in> (rstep R)^*"
      unfolding cts css .
  qed
  with dpstep show ?thesis by auto
qed


lemma left_linear_min_ichain_imp_clean_min_ichain:
  fixes P R::"('f,'v)trs" and F::"('f\<times>nat)set"
  defines "c \<equiv> \<lambda>t::('f, 'v) term. clean_term F t"
  defines "cs \<equiv> \<lambda>\<sigma>::('f, 'v) subst. clean_subst F \<sigma>"
  assumes subset: "funas_dpp (P, R) \<subseteq> F"
    and left_linear: "left_linear_trs R"
    and chain: "min_ichain (P,R) s t \<sigma>"
  shows "min_ichain (P,R) s t (\<lambda>i. cs(\<sigma> i))"
proof -
  from chain have ichain: "ichain (P,R) s t \<sigma>"
    and minimal: "\<forall>i. SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)
  from ichain_imp_clean_ichain[OF subset ichain] have ichain: "ichain (P,R) s t (\<lambda> i. cs(\<sigma> i))" unfolding cs_def .
  from subset have "funas_trs R \<subseteq> F" by simp
  from subset have subset_P: "funas_trs P \<subseteq> F" and subset_R: "funas_trs R \<subseteq> F" by auto
  have wf_t: "\<forall>i. t i \<in> wf_terms F"
  proof
    fix i
    from ichain have "(s i,t i) \<in> P" by simp
    from rhs_wf[OF this subset_P] show "t i \<in> wf_terms F" .
  qed
  have minimal: "\<forall>i. SN_elt (rstep R) (t i \<cdot> (cs(\<sigma> i)))"
  proof
    fix i show "SN_elt (rstep R) ((t i) \<cdot> cs(\<sigma> i))"
    proof -
      from 
        clean_subst_apply_term[of _ F, OF wf_t[THEN spec[of _ i]], of "\<sigma> i"]
        left_linear_SN_elt_imp_clean_SN_elt[OF left_linear minimal[THEN spec[of _ i]], of F]
        show ?thesis unfolding cs_def c_def by simp
    qed
  qed
  with ichain show ?thesis by (auto simp: minimal_cond_def)
qed

definition
  funas_below_term :: "('f, 'v) term \<Rightarrow> 'f sig"
where 
  "funas_below_term t = \<Union> set (map funas_term (args t))"

definition
  funas_below_term_list' :: "('f,'v)term \<Rightarrow> ('f\<times>nat)list"
where
  "funas_below_term_list' t = concat (map funas_term_list' (args t))"

lemma funas_below_term_list'_sound[simp]:
  "set (funas_below_term_list' t) = funas_below_term t"
by (cases t) (simp_all add: funas_below_term_def funas_below_term_list'_def)

definition
  funas_below_rule :: "('f, 'v) rule \<Rightarrow> 'f sig"
where
  "funas_below_rule r \<equiv> funas_below_term (fst r) \<union> funas_below_term (snd r)"

definition
  funas_below_rule_list' :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list"
where
  "funas_below_rule_list' r \<equiv> funas_below_term_list' (fst r) @ funas_below_term_list' (snd r)"

lemma funas_below_rule_list'_sound[simp]:
  "set (funas_below_rule_list' r) = funas_below_rule r"
unfolding funas_below_rule_list'_def funas_below_rule_def by (cases r) simp

definition
  funas_below_trs :: "('f, 'v) trs \<Rightarrow> 'f sig"
where
  "funas_below_trs R \<equiv> \<Union>r\<in>R. funas_below_rule r"

definition
  funas_below_trs_list :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list"
where
  "funas_below_trs_list trs \<equiv> remdups (concat (map funas_below_rule_list' trs))"

lemma funas_below_trs_list_sound[simp]:
  "set (funas_below_trs_list trs) = funas_below_trs (set trs)"
by (induct trs) (auto simp: funas_below_trs_def funas_below_trs_list_def)

lemmas funas_below_defs = funas_below_trs_def funas_below_rule_def funas_below_term_def
                          funas_below_trs_list_def

lemma ichain_imp_clean_ichain_fcc: 
  fixes P R::"('f, 'v) trs" and F::"'f sig"
  defines "c \<equiv> \<lambda>t::('f, 'v) term. clean_term F t"
  defines "cs \<equiv> \<lambda>\<sigma>::('f, 'v) subst. clean_subst F \<sigma>"
  assumes subset: "funas_below_trs P \<union> funas_trs R \<subseteq> F"
  and non_var: "\<And> s t. (s,t) \<in> P \<Longrightarrow> \<not> is_var s \<and> \<not> is_var t \<and> \<not> defined R (the (root t), num_args t)"
  and nvar: "\<forall> lr \<in> R. \<forall> x. fst lr \<noteq> Var x"
  and chain: "ichain (P,R) s t \<sigma>"
  shows "ichain (P,R) s t (\<lambda> i. cs (\<sigma> i))"
proof -
  from chain have dpstep: "\<And> i. (s i,t i) \<in> P"
    and seq: "\<And> i. (t i \<cdot> \<sigma> i,s(Suc i)\<cdot>\<sigma>(Suc i)) \<in> (rstep R)^*"
    by auto
  let ?split = "\<lambda> i f g ss ts. s i = Fun f ss \<and> t i = Fun g ts \<and> \<not> defined R (g, length ts)" 
  have fs: "\<forall> i. \<exists> f g ss ts. ?split i f g ss ts"
  proof
    fix i
    from non_var dpstep have "\<not> is_var (s i) \<and> \<not> is_var (t i)" by auto
    with non_var[OF dpstep[of i]] obtain f ss g ts where "?split i f g ss ts" by (cases "s i", auto, cases "t i", auto) 
    thus "\<exists> f g ss ts. ?split i f g ss ts" by blast
  qed
  from choice[OF fs] obtain f where fs: "\<forall> i. \<exists> g ss ts. ?split i (f i) g ss ts" ..
  from choice[OF fs] obtain g where fs: "\<forall> i. \<exists> ss ts. ?split i (f i) (g i) ss ts" ..
  from choice[OF fs] obtain ss where fs: "\<forall> i. \<exists> ts. ?split i (f i) (g i) (ss i) ts" ..
  from choice[OF fs] obtain ts where fs: "\<And> i. ?split i (f i) (g i) (ss i) (ts i)" by auto
  let ?s = "\<lambda> i. Fun (f i) (map c (ss i))"
  let ?t = "\<lambda> i. Fun (g i) (map c (ts i))"
  let ?ssj = "\<lambda> i j. ss i ! j \<cdot> \<sigma> i"
  let ?tsj = "\<lambda> i j. ts i ! j \<cdot> \<sigma> i"
  let ?ssc = "\<lambda> i j. ss i ! j \<cdot> cs(\<sigma> i)"
  let ?tsc = "\<lambda> i j. ts i ! j \<cdot> cs(\<sigma> i)"
  let ?cond = "\<lambda> i. length (ts i) = length (ss (Suc i)) \<and> g i = f (Suc i) \<and> (\<forall> j < length (ts i). (?tsj i j, ?ssj (Suc i) j) \<in> (rstep R)^*)"
  have argsteps: "\<And> i. ?cond i"
  proof -
    fix i
    from fs have ndef: "(g i, length (ts i)) \<notin> defs R" unfolding defs_def by auto
    from nondef_root_imp_arg_steps[OF seq[of i, simplified fs, simplified] nvar, simplified, OF ndef]
      nth_map[where xs = "ts  i"]
    show "?cond i" by auto
  qed
  from subset have subset_P: "funas_below_trs P \<subseteq> F" by simp
  from subset have subset_R: "funas_trs R \<subseteq> F" by simp
  have main: "\<And> i j. j < length (ts i) \<Longrightarrow> (?tsc i j, ?ssc (Suc i) j) \<in> (rstep R)^*"
  proof -
    fix i j
    assume j: "j < length (ts i)"
    with fs have "?tsj i j \<in> set (args (t i \<cdot> \<sigma> i))" by simp
    from j fs[of "Suc i"] argsteps[of i] have "ss (Suc i) ! j \<in> set (args (s (Suc i)))" by auto 
    with dpstep[of "Suc i"] subset_P have "funas_term (ss (Suc i) ! j) \<subseteq> F"
      unfolding funas_below_defs by force
    hence wf_s: "ss (Suc i) ! j \<in> wf_terms F" by (rule funas_term_wf_terms)
    from j fs[of i] have "ts i ! j \<in> set (args (t i))" by auto 
    with dpstep[of i] subset_P have "funas_term (ts i ! j) \<subseteq> F"
      unfolding funas_below_defs by force
    hence wf_t: "ts i ! j \<in> wf_terms F" by (rule funas_term_wf_terms)
    have c_t: "c(?tsj i j) = (?tsc i j)" unfolding c_def cs_def
      by (simp add: clean_subst_apply_term[OF wf_t])
    have c_s: "c(?ssj (Suc i) j) = (?ssc (Suc i) j)" unfolding c_def cs_def
      by (simp add: clean_subst_apply_term[OF wf_s])
    from argsteps[of i] j have "(?tsj i j, ?ssj (Suc i) j) \<in> (rstep R)^*" by simp
    from rsteps_imp_clean_rsteps[OF subset_R this] show "(?tsc i j, ?ssc (Suc i) j) \<in> (rstep R)^*"
      unfolding c_t[symmetric] c_s[symmetric] c_def .
  qed
  { 
    fix i
    from `?cond i` have len: "(length (ss (Suc i))) = length (ts i)" by simp
    have "(Fun (g i) (map (\<lambda> t. t \<cdot> cs (\<sigma> i)) (ts i)), Fun (g i) (map (\<lambda> t. t \<cdot> cs (\<sigma> (Suc i))) (ss (Suc i)))) \<in> (rstep R)^*" 
    proof (rule mp[OF spec[OF spec[OF spec[OF rstep_rtrancl_all_ctxt_closed[of R, unfolded all_ctxt_closed_def]]]] conjI],
      simp add: `?cond i`, simp add: len, intro allI impI)
      fix j
      assume "j < length (ts i)"
      from main[OF this]
      show "(?tsc i j, ?ssc (Suc i) j) \<in> (rstep R)^*" .
    qed
    with `?cond i` 
    have "(t i \<cdot> cs (\<sigma> i), s (Suc i) \<cdot> cs (\<sigma> (Suc i))) \<in> (rstep R)^*" by (simp add: fs[of i] fs[of "Suc i"])
  } 
  with dpstep show ?thesis by (auto)
qed

lemma left_linear_min_ichain_imp_clean_min_ichain_fcc: 
  fixes P R::"('f,'v)trs" and F::"'f sig"
  defines "c \<equiv> \<lambda>t::('f, 'v) term. clean_term F t"
  defines "cs \<equiv> \<lambda>\<sigma>::('f, 'v) subst. clean_subst F \<sigma>"
  assumes left_linear: "left_linear_trs R"
  and subset: "funas_below_trs P \<union> funas_trs R \<subseteq> F"
  and non_var: "\<And> s t. (s,t) \<in> P \<Longrightarrow> \<not> is_var s \<and> \<not> is_var t \<and> \<not> defined R (the (root t), num_args t)"
  and chain: "min_ichain (P,R) s t \<sigma>"
  shows "min_ichain (P,R) s t (\<lambda> i. cs (\<sigma> i))"
proof -
  from chain have ichain: "ichain (P,R) s t \<sigma>"
    and min: "\<And> i. SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)
  from left_var_imp_not_SN[where R = R and t = "t 0 \<cdot> \<sigma> 0"] min[of 0]
  have nvar: "\<forall> lr \<in> R. \<forall> x. fst lr \<noteq> Var x" by force 
  from ichain_imp_clean_ichain_fcc[OF subset non_var nvar ichain]
  have ichain: "ichain (P,R) s t (\<lambda> i. cs (\<sigma> i))" unfolding cs_def .
  from subset have subset_P: "funas_below_trs P \<subseteq> F" unfolding funas_dpp.simps by auto
  from subset have subset_R: "funas_trs R \<subseteq> F" by simp
  {
    fix i
    from ichain have "(s i, t i) \<in> P" by simp
    with subset_P non_var[OF this] obtain f ts where ti: "t i = Fun f ts" and tiP: "funas_below_term (Fun f ts) \<subseteq> F" and ndef: "(f, length (map (\<lambda> t. t \<cdot> cs (\<sigma> i)) ts)) \<notin> defs R" 
      unfolding funas_below_trs_def funas_below_rule_def by (cases "t i", simp, force simp: defs_def)
    have "SN_elt (rstep R) (t i  \<cdot> (cs(\<sigma> i)))" 
    proof (simp add: ti, rule  SN_args_imp_SN[OF _ nvar ndef]) 
      fix u
      assume "u \<in> set (map (\<lambda> t. t \<cdot> cs (\<sigma> i)) ts)" (is "_ \<in> set ?ts")
      then obtain uu where uu: "uu \<in> set ts" and u: "u = uu \<cdot> cs (\<sigma> i)" by auto      
      from uu tiP have fuu: "funas_term uu \<subseteq> F" unfolding funas_below_term_def by auto
      hence wuu: "uu \<in> wf_terms F" by (rule funas_term_wf_terms)
      from SN_imp_SN_arg[OF min[of i, simplified ti, simplified]] uu have "SN_elt (rstep R) (uu \<cdot> \<sigma> i)" by auto
      from left_linear_SN_elt_imp_clean_SN_elt[OF left_linear this, of F] show "SN_elt (rstep R) u"
        unfolding clean_subst_apply_term[OF wuu] u cs_def .
    qed
  }
  with ichain show ?thesis by (simp add: minimal_cond_def)
qed


lemma min_ichain_imp_ichain:
  assumes "min_ichain DPP s t \<sigma>" shows "ichain DPP s t \<sigma>"
using assms by (cases DPP) simp_all

fun
  shift :: "'a iseq \<Rightarrow> nat \<Rightarrow> 'a iseq"
where
  "shift f j = (\<lambda> i. f (i+j))"

definition ci_subset :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool" 
where "ci_subset R S \<equiv> \<forall> l r. (l,r) \<in> R \<longrightarrow> (\<exists> l' r' C. (l',r') \<in> S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>)"

notation (xsymbols)
  ci_subset ("(_/ \<subseteq>ci _)" [50,51] 50)

lemma ci_subsetI: assumes "R \<subseteq> S"
  shows "ci_subset R S"
unfolding ci_subset_def
proof (intro allI impI)
  fix l r
  assume "(l,r) \<in> R"
  with assms have S: "(l,r) \<in> S" by auto
  show "\<exists> l' r' C. (l',r') \<in> S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>"
    by (rule exI[of _ l], rule exI[of _ r], rule exI[of _ \<box>], simp add: S)
qed

lemma ci_subset_refl: "R \<subseteq>ci R"
proof (unfold ci_subset_def, intro allI impI)
  fix l r
  assume rule: "(l,r) \<in> R"
  show "\<exists> l' r' C.  (l', r') \<in> R \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>"
    by (rule exI[of _ l], rule exI[of _ r], rule exI[of _ \<box>], auto simp: rule)
qed
  
lemma rstep_ci_mono: assumes "R \<subseteq>ci S"
  shows "rstep R \<subseteq> rstep S"
proof (simp only: rstep_subset_characterization, intro allI impI)
  fix l r
  assume "(l,r) \<in> R"
  with assms obtain l' r' C where cond: "(l',r') \<in> S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>" unfolding ci_subset_def by blast
  show "\<exists> l' r' C \<sigma>. (l',r') \<in> S \<and> l = C\<langle>l' \<cdot> \<sigma>\<rangle> \<and> r = C\<langle>r' \<cdot> \<sigma>\<rangle>"
    by (rule exI[of _ l'], rule exI[of _ r'], rule exI[of _ C], rule exI[of _ Var], simp add: cond)
qed

lemma minimal_cond_mono: 
  assumes subset: "R \<subseteq>ci R'" and cond: "minimal_cond R' s t \<sigma>" 
  shows "minimal_cond R s t \<sigma>"
unfolding minimal_cond_def
proof 
  fix i
  from cond have SN: "SN_elt (rstep R') (t i \<cdot> \<sigma> i)"
    unfolding minimal_cond_def by auto
  have "rstep R \<subseteq> rstep R'" by (rule rstep_ci_mono, rule subset)
  with SN show "SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by (rule SN_elt_subset)
qed

lemma min_ichainI[intro]: 
  assumes "R \<subseteq>ci R'" and "minimal_cond R' s t \<sigma>" and "ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)" 
  shows "min_ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  from assms have "minimal_cond R s t \<sigma>" using minimal_cond_mono by auto
  hence "minimal_cond R (shift s i) (shift t i) (shift \<sigma> i)" 
    unfolding minimal_cond_def by auto
  with assms show ?thesis by auto
qed

definition finite_dpp :: "('f,'v)dpp \<Rightarrow> bool" where "finite_dpp DPP = (\<not>(\<exists>s t \<sigma>. min_ichain DPP s t \<sigma>))"

fun next4 :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow>
 (('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)subst)"
where "next4 shp R s = (SOME (l,r,u,\<sigma>).
 (l,r) \<in> DP shp R \<and> r\<cdot>\<sigma> = sharp_term shp u \<and> u \<in> Tinf R \<and> (sharp_term shp s,l\<cdot>\<sigma>) \<in> (rstep R)^* )"

fun concrete_seq :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> nat \<Rightarrow>
 (('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)subst)"
where "concrete_seq shp R s 0       = (s,next4 shp R s)"
    | "concrete_seq shp R r (Suc n) = (let (_,_,_,s,_) = concrete_seq shp R r n in (s,next4 shp R s))"

lemma someI_ex4: "\<exists>w x y z. P(w,x,y,z) \<Longrightarrow> P(SOME(w,x,y,z). P(w,x,y,z))"
using some_eq_ex[of P] unfolding split_def by force


theorem not_SN_imp_min_ichain_both:
 fixes R :: "('f, 'v) trs" and shp :: "'f \<Rightarrow> 'f"
 assumes wf: "wf_trs R" and nSN: "\<not> SN (rstep R)" 
 shows "\<exists>s t \<sigma>. ichain (DP shp R, R) s t \<sigma> \<and> ((\<forall> f n. defined R (f,n) \<longrightarrow> \<not> defined R (shp f,n)) \<longrightarrow> min_ichain (DP shp R, R) s t \<sigma>)"
proof -
  from `\<not> (SN (rstep R))` obtain s' :: "('f, 'v) term" where "\<not>(SN_elt (rstep R) s')"
    unfolding SN_def by auto
  then obtain s where "s' \<unrhd>  s" and "s \<in> Tinf R" using not_SN_imp_subt_Tinf by auto
  let ?I = "\<lambda>i. fst(concrete_seq shp R s i)"
  let ?S = "\<lambda>i. fst(snd(concrete_seq shp R s i))"
  let ?T = "\<lambda>i. fst(snd(snd(concrete_seq shp R s i)))"
  let ?\<sigma> = "\<lambda>i. snd(snd(snd(snd(concrete_seq shp R s i))))"
  let ?P = "\<lambda>s l r u \<sigma>. (l,r) \<in> DP shp R \<and> r\<cdot>\<sigma> = sharp_term shp u \<and> u \<in> Tinf R \<and> (sharp_term shp s,l\<cdot>\<sigma>) \<in> (rstep R)^*"
  let ?P' =  "\<lambda>(l,r,u,\<sigma>). ?P s l r u \<sigma>"
  have main: "\<forall>i. ?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
  proof
    fix i show "?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
    proof (induct i)
      case 0 
      have "\<exists>l r u \<sigma>. ?P'(l,r,u,\<sigma>)"
        using Tinf_lemma_DP_explicit_subst[OF `wf_trs R` `s \<in> Tinf R`] by auto
      hence "?P'(next4 shp R s)" unfolding next4.simps using someI_ex4[where P = "?P'"] by best
      have ini: "?I 0 = s" unfolding concrete_seq.simps by simp
      have s: "?S 0 = fst(next4 shp R s)" unfolding concrete_seq.simps by simp
      have t: "?T 0 = fst(snd(next4 shp R s))" unfolding concrete_seq.simps by simp
      have sub: "?\<sigma> 0 = snd(snd(snd(next4 shp R s)))" unfolding concrete_seq.simps Let_def split_def by auto
      have nxt: "?I(Suc 0) = fst(snd(snd(next4 shp R s)))" unfolding concrete_seq.simps Let_def split_def by auto
      show ?case unfolding ini s t sub nxt using `?P'(next4 shp R s)` by auto
    next
      case (Suc i)
      let ?P'' = "\<lambda>(w,x,y,z). ?P (?I(Suc i)) w x y z"
      from Suc have "?I(Suc i) \<in> Tinf R" by simp
      have "\<exists>l r u \<sigma>. ?P''(l,r,u,\<sigma>)"
        using Tinf_lemma_DP_explicit_subst[OF `wf_trs R` `?I(Suc i) \<in> Tinf R`] by auto
      hence IH: "?P''(next4 shp R (?I(Suc i)))" unfolding next4.simps using someI_ex4[where P = "?P''"] by simp
      have s: "?S(Suc i) = fst(next4 shp R (?I(Suc i)))" unfolding concrete_seq.simps Let_def split_def by simp
      have t: "?T(Suc i) = fst(snd(next4 shp R (?I(Suc i))))" unfolding concrete_seq.simps Let_def split_def by simp
      have sub: "?\<sigma>(Suc i) = snd(snd(snd(next4 shp R (?I(Suc i)))))" unfolding concrete_seq.simps Let_def split_def by simp
      have nxt: "?I(Suc(Suc i)) = fst(snd(snd(next4 shp R (?I(Suc i)))))" unfolding concrete_seq.simps Let_def split_def by simp
      show ?case unfolding s t sub nxt using IH by auto
    qed
  qed
  from main have P: "\<forall>i. (?S i,?T i) \<in> DP shp R" by simp
  from main have eq: "\<And>i. sharp_term shp (?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by auto
  have R: "\<forall>i. (?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> (rstep R)^*"
  proof
    fix i show "(?T i \<cdot> ?\<sigma> i,?S(Suc i)\<cdot>?\<sigma>(Suc i)) \<in> (rstep R)^*"
    proof (cases i)
      case 0
      from main have first: "?T 0 \<cdot> ?\<sigma> 0 = sharp_term shp (?I(Suc 0))" by best
      from main have "(sharp_term shp (?I(Suc 0)),?S(Suc 0)\<cdot>?\<sigma>(Suc 0)) \<in> (rstep R)^*" by best
      thus ?thesis unfolding 0 unfolding first by simp
    next
      case (Suc i) show ?thesis unfolding eq[symmetric] using main by best
    qed
  qed
  from P and R have ichain: "ichain (DP shp R, R) ?S ?T ?\<sigma>" by auto
  show ?thesis
  proof (intro exI, rule conjI[OF ichain], intro impI)
    assume ndef: "\<forall> f n. defined R (f,n) \<longrightarrow> \<not> defined R (shp f,n)"
    have Tinf: "\<forall>i. ?I i \<in> Tinf R"
    proof
      fix i show "?I i \<in> Tinf R" using main by (induct i) (auto simp: `s \<in> Tinf R`)
    qed
    have SN_t: "\<forall>i. SN_elt (rstep R) (?T i \<cdot> ?\<sigma> i)"
    proof
      fix i
      obtain t where t: "t = ?I(Suc i)" by simp
      with Tinf have tinf: "t \<in> Tinf R" by best
      from Tinf_imp_def_root[OF wf this] obtain f ss where teq: "t = Fun f ss" and f: "defined R (f, length ss)" by auto
      have SN: "SN_elt (rstep R) (sharp_term shp t)"
        unfolding teq sharp_term.simps
      proof (rule SN_args_imp_SN)
        from ndef[THEN spec, THEN spec, THEN mp[OF _ f]] show "(shp f, length ss) \<notin> defs R" unfolding defs_def by simp
      next
        from wf[unfolded wf_trs_def] show "\<forall> lr \<in> R. \<forall> x. fst lr \<noteq> Var x" by force
      next
        fix s 
        assume "s \<in> set ss" 
        hence "t \<rhd> s" unfolding teq ..
        with tinf[unfolded Tinf_def] show "SN_elt (rstep R) s" by simp
      qed
      from eq have eq': "?T i \<cdot> ?\<sigma> i = sharp_term shp t" unfolding t by simp
      show "SN_elt (rstep R) (?T i \<cdot> ?\<sigma> i)" unfolding eq' using SN by simp
    qed
    from ichain and SN_t
    show "min_ichain (DP shp R, R) ?S ?T ?\<sigma>" 
      unfolding min_ichain.simps minimal_cond_def by blast
  qed
qed
 
hide_const concrete_seq next4

corollary not_SN_imp_min_ichain:
 fixes R :: "('f, 'v) trs" and shp :: "'f \<Rightarrow> 'f"
 assumes wf: "wf_trs R" and nSN: "\<not> SN (rstep R)" and ndef: "\<And> f n. defined R (f,n) \<Longrightarrow> \<not> defined R (shp f,n)"
 shows "\<exists>s t \<sigma>. min_ichain (DP shp R, R) s t \<sigma>"
using not_SN_imp_min_ichain_both[OF wf nSN, of shp] ndef by blast

corollary not_SN_imp_ichain:
 fixes R :: "('f, 'v) trs" and shp :: "'f \<Rightarrow> 'f"
 assumes wf: "wf_trs R" and nSN: "\<not> SN (rstep R)" 
 shows "\<exists>s t \<sigma>. ichain (DP shp R, R) s t \<sigma>"
using not_SN_imp_min_ichain_both[OF wf nSN, of shp] by blast

lemma finite_initial_dpp_imp_SN:
  assumes "wf_trs R" and ndef: "\<And> f n. defined R (f,n) \<Longrightarrow> \<not> defined R (shp f,n)" and "finite_dpp (DP shp R, R)" shows "SN(rstep R)"
proof (rule ccontr)
 assume "\<not>SN(rstep R)"
 from not_SN_imp_min_ichain[OF `wf_trs R` `\<not>(SN(rstep R))` ndef]
 obtain s t \<sigma> where "min_ichain (DP shp R, R) s t \<sigma>" by best
 with `finite_dpp (DP shp R, R)` show False unfolding finite_dpp_def by best
qed

lemma funas_dpp_initial_subset_sharp_sig:
  "funas_dpp (DP shp R, R) \<subseteq> sharp_sig shp (funas_trs R)"
using funas_DP_subset unfolding sharp_sig_def by simp

lemma SN_wfrstep_imp_SN_rstep:
  assumes "SN (wfrstep R)" shows "SN (rstep R)"
proof (cases "wf_trs R")
  case False
  with SN_wfrstep_imp_wf_trs[OF assms] show ?thesis by blast
next
  case True
  show ?thesis
  proof (rule ccontr)
    let ?F = "sharp_sig id (funas_trs R)"
    let ?c = "clean_term ?F"
    let ?cs = "clean_subst ?F"
    assume "\<not> SN (rstep R)"
    from not_SN_imp_ichain[OF True this]
      obtain s t \<sigma> where ichain: "ichain (DP id R, R) s t \<sigma>"
      unfolding finite_dpp_def by simp blast
    from ichain_imp_clean_ichain[OF funas_dpp_initial_subset_sharp_sig this]
      have clean_ichain: "ichain (DP id R,R) s t (\<lambda>i. ?cs(\<sigma> i))" .
    hence "\<not> SN (wfrstep R)"
      using clean_ichain_imp_not_SN_wfrstep[of id R s t]
      unfolding o_def inj_on_def by auto
    with assms show False by blast
  qed
qed

lemma SN_wfrstep_SN_rstep_conv: "SN (wfrstep R) = SN (rstep R)"
proof -
  {
    assume "\<not> SN (wfrstep R)"
    then obtain S where "\<forall>i. (S i, S (Suc i)) \<in> wfrstep R" unfolding SN_defs by auto
    hence "\<not> SN (rstep R)" unfolding SN_defs wfrstep_def by blast
  }
  with SN_wfrstep_imp_SN_rstep show ?thesis by blast
qed


lemma SN_wf_ground: assumes c: "(c,0) \<in> F" and F: "funas_trs R \<subseteq> F"
  and SN: "\<And> t. t \<in> wf_terms F \<Longrightarrow> ground t \<Longrightarrow> SN_elt (rstep R) t"
  shows "SN (rstep R)"
proof (rule SN_wfrstep_imp_SN_rstep, rule)
  fix f 
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> wfrstep R"
  hence f0: "f 0 \<in> wf_terms (funas_trs R)" and steps: "\<And> i. (f i, f (Suc i)) \<in> rstep R" unfolding wfrstep_def by auto
  from f0 wf_terms_subset[OF F] have f0: "f 0 \<in> wf_terms F" by auto
  let ?\<sigma> = "(\<lambda> _. Fun c [])"
  from f0 c have wf: "f 0 \<cdot> ?\<sigma> \<in> wf_terms F" (is "?t \<in> _")
    by (simp add: wf_terms.simps[of "Fun c []"])
  have "ground ?t" by auto
  from SN[OF wf this] have "SN_elt (rstep R) ?t" by auto
  from SNinstance_imp_SN[OF this] steps
  show False unfolding SN_defs by auto
qed


(* TODO: remove shp = id *)
lemma SN_imp_finite_initial_dpp:
  assumes wf: "wf_trs R" and SN: "SN(rstep R)" and id: "shp = id" shows "finite_dpp (DP shp R, R)"
proof (rule ccontr)
  assume "\<not> finite_dpp (DP shp R, R)"
  then obtain s t \<sigma> where ichain: "ichain (DP shp R,R) s t \<sigma>" unfolding finite_dpp_def by auto
  have "\<not> SN(wfrstep R)" using
    ichain_imp_clean_ichain[OF funas_dpp_initial_subset_sharp_sig ichain]
    clean_ichain_imp_not_SN_wfrstep[of shp R s t] unfolding id inj_on_def by auto
  thus False using SN by (simp add: SN_wfrstep_SN_rstep_conv)
qed

text {*
Well-formedness is important since a variable lhs implies that @{text R} is not terminating.
Still, there is no \emph{minimal} infinite chain, since there is no term that is strongly normalizing w.r.t.\ to @{text R}.
*}
(*
lemma SN_iff_finite_initial_dpp:
  assumes "wf_trs R" shows "SN(rstep R) = finite_dpp (DP R, ID R)"
using SN_imp_finite_initial_dpp[OF assms] finite_initial_dpp_imp_SN[OF assms] by blast
*)

subsection {* Soundness and Chain-Identifyingness of Termination Techniques and DP Processors *}

definition
  sound_tt :: "(('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool) \<Rightarrow> bool"
where
  "sound_tt tt \<equiv> \<forall>R R'. tt R R' \<and> SN (rstep R') \<longrightarrow> SN (rstep R)"

definition
  weak_ci_tt :: "(('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool) \<Rightarrow> bool"
where
  "weak_ci_tt tt \<equiv> \<forall>R R'. tt R R' \<longrightarrow> (R' \<subseteq>ci R \<and> (SN (rstep R') \<longrightarrow> SN (rstep R)))"

definition
  ci_tt :: "(('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool) \<Rightarrow> bool"
where
  "ci_tt tt \<equiv> \<forall>R R'. tt R R' \<longrightarrow> (R' \<subseteq> R \<and> (\<forall>t. ideriv R t \<longrightarrow> (\<exists>i. ideriv R' (shift t i))))"


definition
  sound_tt_impl :: "(('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> bool) \<Rightarrow> bool"
where
  "sound_tt_impl tt \<equiv> \<forall>R R'. tt R R' \<and> SN (rstep (set R')) \<longrightarrow> SN (rstep (set R))"

definition
  ci_tt_impl :: "(('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> bool) \<Rightarrow> bool"
where
  "ci_tt_impl tt \<equiv> \<forall>R R'. tt R R' \<longrightarrow>
     (set R' \<subseteq> set R \<and> (\<forall>t. ideriv (set R) t \<longrightarrow> (\<exists>i. ideriv (set R') (shift t i))))"

definition
  weak_ci_tt_impl :: "(('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> bool) \<Rightarrow> bool"
where
  "weak_ci_tt_impl tt \<equiv> \<forall>R R'. tt R R' \<longrightarrow>
     (set R' \<subseteq>ci set R \<and> (SN (rstep (set R')) \<longrightarrow> SN (rstep (set R))))"

definition
  ci_proc :: "(('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool) \<Rightarrow> bool"
where
  "ci_proc proc \<equiv> \<forall>P R P' R'. proc P R P' R' \<longrightarrow>
     (R' \<subseteq>ci R \<and> (\<forall>s t \<sigma>. min_ichain (P, R) s t \<sigma> \<longrightarrow>
       (\<exists>i. ichain (P', R') (shift s i) (shift t i) (shift \<sigma> i))))"

definition
  sound_proc :: "(('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs  \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool) \<Rightarrow> bool"
where
  "sound_proc proc \<equiv> \<forall>P R P' R'. proc P R P' R' \<and> finite_dpp (P', R') \<longrightarrow> finite_dpp (P, R)"

lemma ci_tt_to_weak_ci_tt: 
  assumes ci_tt: "ci_tt tt"
  shows "weak_ci_tt tt"
unfolding weak_ci_tt_def 
proof (intro allI impI)
  fix R rR
  assume tt: "tt R rR" 
  from ci_tt[unfolded ci_tt_def, THEN spec, THEN spec, THEN mp, OF tt] have one: "rR \<subseteq>ci R" and two: "\<And> t. ideriv R t \<Longrightarrow> \<exists> i. ideriv rR (shift t i)"  
    using ci_subsetI by auto
  {
    assume "SN (rstep rR)"
    hence "\<not> (\<exists> t. ideriv rR t)" unfolding ideriv_def SN_defs  by blast
    with two have "\<not> (\<exists> t. ideriv R t)" by blast
    hence "SN (rstep R)" unfolding ideriv_def SN_defs by blast
  }
  with one show "rR \<subseteq>ci R \<and> (SN(rstep rR) \<longrightarrow> SN(rstep R))" by simp
qed

lemma weak_ci_tt_to_sound_tt: 
  assumes ci_tt: "weak_ci_tt tt"
  shows "sound_tt tt"
using assms unfolding weak_ci_tt_def sound_tt_def by blast

lemma ci_proc_to_sound_proc[simp]: 
  assumes ci_proc: "ci_proc proc"
  shows "sound_proc proc"
unfolding sound_proc_def
proof (intro allI impI, elim conjE)
  fix P R rP rR
  assume cond: "proc P R rP rR" and finite: "finite_dpp (rP, rR)"
  show "finite_dpp (P,R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (P,R)"
    from this obtain s t \<sigma> where min: "min_ichain (P,R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where subset: "rR \<subseteq>ci R" and ichain: "ichain (rP,rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_proc_def by force
    have "min_ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = R and R = rR and P = rP] using ichain min subset by auto
    hence "\<not> finite_dpp (rP, rR)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

definition
  ci_proc_impl :: "(('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> bool) \<Rightarrow> bool"
where
  "ci_proc_impl proc \<equiv> \<forall> P R rP rR. (proc P R (computeRuleMap R) rP rR \<longrightarrow> (set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))))"

definition sound_proc_impl :: "(('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_proc_impl proc \<equiv> \<forall> P R rP rR. (proc P R (computeRuleMap R) rP rR \<longrightarrow> (finite_dpp (set rP, set rR) \<longrightarrow> finite_dpp (set P, set R)))"

definition reverse_trs :: "('f,'v)rules \<Rightarrow> ('f,'v)rules"
where "reverse_trs r \<equiv> map (\<lambda> lr. (snd lr, fst lr)) r"

definition
  ci_only_pairs_proc_impl :: "(('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rules \<Rightarrow> bool) \<Rightarrow> bool"
where
  "ci_only_pairs_proc_impl proc \<equiv> \<forall> P R rP. (proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP \<longrightarrow> ((\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)))))"

definition
  sound_only_pairs_proc_impl :: "(('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rules \<Rightarrow> bool) \<Rightarrow> bool"
where
  "sound_only_pairs_proc_impl proc \<equiv> \<forall> P R rP. (proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP \<longrightarrow> (finite_dpp (set rP, set R) \<longrightarrow> finite_dpp (set P, set R)))"

lemma ci_only_pairs_imp_sound_only_pairs[simp]: 
  assumes ci_proc: "ci_only_pairs_proc_impl proc"
  shows "sound_only_pairs_proc_impl proc" 
unfolding sound_only_pairs_proc_impl_def
proof ((rule allI)+, (rule impI)+)
  fix P R rP 
  assume cond: "proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP" and finite: "finite_dpp (set rP, set R)"
  show "finite_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (set P, set R)"
    from this obtain s t \<sigma> where min: "min_ichain (set P, set R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where ichain: "ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_only_pairs_proc_impl_def by blast
    have "min_ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = "set R" and R = "set R" and P = "set rP"] using ichain min ci_subset_refl[of "set R"] by auto
    hence "\<not> finite_dpp (set rP, set R)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

lemma ci_tt_to_weak_ci_tt_impl: 
  assumes ci_tt: "ci_tt_impl tt"
  shows "weak_ci_tt_impl tt"
unfolding weak_ci_tt_impl_def 
proof (intro allI impI)
  fix R rR
  assume tt: "tt R rR" 
  from ci_tt[unfolded ci_tt_impl_def, THEN spec, THEN spec, THEN mp, OF tt] have one: "set rR \<subseteq>ci set R" and two: "\<And> t. ideriv (set R) t \<Longrightarrow> \<exists> i. ideriv (set rR) (shift t i)"  
    using ci_subsetI by auto
  {
    assume "SN (rstep (set rR))"
    hence "\<not> (\<exists> t. ideriv (set rR) t)" unfolding ideriv_def SN_defs  by blast
    with two have "\<not> (\<exists> t. ideriv (set R) t)" by blast
    hence "SN (rstep (set R))" unfolding ideriv_def SN_defs by blast
  }
  with one show "set rR \<subseteq>ci set R \<and> (SN(rstep (set rR)) \<longrightarrow> SN(rstep (set R)))" by simp
qed

lemma weak_ci_tt_to_sound_tt_impl:
  assumes ci_tt: "weak_ci_tt_impl tt"
  shows "sound_tt_impl tt"
  using assms unfolding sound_tt_impl_def weak_ci_tt_impl_def by blast

lemma ci_imp_sound[simp]: 
  assumes ci_proc: "ci_proc_impl proc"
  shows "sound_proc_impl proc" 
unfolding sound_proc_impl_def
proof ((rule allI)+, (rule impI)+)
  fix P R rP rR
  assume cond: "proc P R (computeRuleMap R) rP rR" and finite: "finite_dpp (set rP, set rR)"
  show "finite_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (set P, set R)"
    from this obtain s t \<sigma> where min: "min_ichain (set P, set R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where ichain: "set rR \<subseteq>ci set R \<and> ichain (set rP,set rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_proc_impl_def by blast
    have "min_ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = "set R" and R = "set rR" and P = "set rP"] using ichain min by auto
    hence "\<not> finite_dpp (set rP, set rR)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

theorem finite_dpp_iff_rel_SN: "finite_dpp (P,R) = rel_SN ((CS P \<inter> {(s,t) | s t. SN_elt (rstep R) t}), rstep R)" (is "_ = rel_SN (?R,?S)")
proof
  assume SN: "rel_SN (?R,?S)"
  show "finite_dpp (P,R)"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain s t \<sigma> where chain: "min_ichain (P,R) s t \<sigma>" unfolding finite_dpp_def by auto
    obtain f where f: "\<And> i. f i = s i \<cdot> \<sigma> i" by auto
    from chain have "\<And> i. (s i \<cdot> \<sigma> i, t i \<cdot> \<sigma> i) \<in> ?R" and "\<And> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?S^*"  
      by (auto simp: minimal_cond_def)
    hence "\<And> i. (f i,f (Suc i)) \<in> ?S^* O ?R O ?S^*" unfolding f by auto
    with SN[unfolded rel_SN_def rel_step.simps] 
    show False unfolding SN_defs by blast
  qed
next
  assume finite: "finite_dpp (P,R)"
  show "rel_SN (?R,?S)" 
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain t where steps: "\<And> i. (t i, t (Suc i)) \<in> ?S^* O ?R O ?S^*"
      unfolding rel_SN_def rel_step.simps SN_defs by auto
    let ?prop = "\<lambda> s u i.  (t i, s) \<in> ?S^* \<and> (s,u) \<in> ?R \<and> (u, t (Suc i)) \<in> ?S^*"
    have "\<forall> i. \<exists> s u. ?prop s u i"
    proof
      fix i
      from steps[of i] show "\<exists> s u. ?prop s u i" by auto
    qed
    from choice[OF this] obtain s where "\<forall> i. \<exists> u. ?prop (s i) u i" ..
    from choice[OF this] obtain u where p: "\<And> i. ?prop (s i) (u i) i" by auto
    let ?prop2 = "\<lambda> l r \<sigma> i. (l, r) \<in> P \<and> SN_elt (rstep R) (r \<cdot> \<sigma>) \<and> r \<cdot> \<sigma> = u i \<and> l \<cdot> \<sigma> = s i"
    have "\<forall> i. \<exists> l r \<sigma>. ?prop2 l r \<sigma> i"
    proof
      fix i
      from p[of i] have "(s i, u i) \<in> CS P" by auto
      with subst_closed_decompose[OF this] obtain l r \<sigma> where  "s i = l \<cdot> \<sigma> \<and> u i = r \<cdot> \<sigma> \<and> (l,r) \<in> P" by auto
      with p[of i] have "?prop2 l r \<sigma> i" by auto
      thus "\<exists> l r \<sigma>. ?prop2 l r \<sigma> i" by blast
    qed
    from choice[OF this] obtain l where "\<forall> i. \<exists> r \<sigma> . ?prop2 (l i) r \<sigma> i" ..
    from choice[OF this] obtain r where "\<forall> i. \<exists> \<sigma> . ?prop2 (l i) (r i) \<sigma> i" ..
    from choice[OF this] obtain \<sigma> where p2: "\<And> i. ?prop2 (l i) (r i) (\<sigma> i) i" by auto
    {
      fix i
      from p2[of i] p[of i] p[of "Suc i"] p2[of "Suc i"]
      have "(r i \<cdot> \<sigma> i, l (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?S^* " by auto
    }
    with p2 have "min_ichain (P,R) l r \<sigma>" 
      by (auto simp: minimal_cond_def p2)
    with finite show False unfolding finite_dpp_def by blast
  qed
qed

definition internal_error where "internal_error \<equiv> (
  ''internal error'' +#+ shows_nl +@+
  ''please contact one of the CeTA members'' +#+ shows_nl +@+
  ''http://cl-informatik.uibk.ac.at/software/ceta'' +#+ shows_nl +@+
  ''ceta@informatik.uibk.ac.at'' +#+ shows_nl
)"

end
