theory Relations
imports Multiset

begin

(** At the moment these are ad-hoc definitions to deal with multisets.
  One could extend it quite easily by adding a membership predicate as
  a parameter. Is it worthy? *)

type_synonym 'a relation = "('a \<times> 'a) set"

text{* Common predicates on relations *}

definition reflexive :: "'a relation \<Rightarrow> bool" where
  "reflexive R \<equiv> (\<forall>t. (t,t) \<in> R)"

definition transitive :: "'a relation \<Rightarrow> bool" where
  "transitive R \<equiv> (\<forall>t u v. (t,u) \<in> R \<longrightarrow> (u,v) \<in> R \<longrightarrow> (t,v) \<in> R)"

definition included :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> bool" where
  "included R1 R2 \<equiv> (\<forall>t u. (t,u) \<in> R1 \<longrightarrow> (t,u) \<in> R2)"

definition compatible_l :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> bool" where
  "compatible_l R1 R2 \<equiv> (\<forall>t u v. (t,u) \<in> R1 \<longrightarrow> (u,v) \<in> R2 \<longrightarrow> (t,v) \<in> R2)"

definition compatible_r :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> bool" where
  "compatible_r R1 R2 \<equiv> (\<forall>t u v. (t,u) \<in> R2 \<longrightarrow> (u,v) \<in> R1 \<longrightarrow> (t,v) \<in> R2)"

text{* Local reflexivity *}

definition locally_reflexive :: "'a relation \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_reflexive R A \<equiv> (\<forall>t. t \<in># A \<longrightarrow> (t,t) \<in> R)"

definition locally_irreflexive :: "'a relation \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_irreflexive R A \<equiv> (\<forall>t. t \<in># A \<longrightarrow> (t,t) \<notin> R)"

text{* Local symmetry *}

definition locally_symmetric :: "'a relation \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_symmetric R A \<equiv> (\<forall>t u. t \<in># A \<longrightarrow> u \<in># A \<longrightarrow>
               (t,u) \<in> R \<longrightarrow> (u,t) \<in> R)"

definition locally_antisymmetric :: "'a relation \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_antisymmetric R A \<equiv> (\<forall>t u. t \<in># A \<longrightarrow> u \<in># A \<longrightarrow>
               (t,u) \<in> R \<longrightarrow> (u,t) \<in> R \<longrightarrow> t = u)"

text{* Local transitivity *}

definition locally_transitive :: "'a relation \<Rightarrow> 'a multiset \<Rightarrow> 'a multiset \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_transitive R A B C \<equiv> (\<forall>t u v.
               t \<in># A \<longrightarrow> u \<in># B \<longrightarrow> v \<in># C \<longrightarrow>
               (t,u) \<in> R \<longrightarrow> (u,v) \<in> R \<longrightarrow> (t,v) \<in> R)"

text{* Local inclusion *}

definition locally_included :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> 'a multiset \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_included R1 R2 A B \<equiv> (\<forall>t u. t \<in># A \<longrightarrow> u \<in># B \<longrightarrow>
               (t,u) \<in> R1 \<longrightarrow>  (t,u) \<in> R2)"

text{* Local transitivity compatibility *}

definition locally_compatible_l :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> 'a multiset \<Rightarrow> 'a multiset \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_compatible_l R1 R2 A B C \<equiv> (\<forall>t u v. t \<in># A \<longrightarrow> u \<in># B \<longrightarrow> v \<in># C \<longrightarrow>
               (t,u) \<in> R1 \<longrightarrow> (u,v) \<in> R2 \<longrightarrow> (t,v) \<in> R2)"

definition locally_compatible_r :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> 'a multiset \<Rightarrow> 'a multiset \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_compatible_r R1 R2 A B C \<equiv> (\<forall>t u v. t \<in># A \<longrightarrow> u \<in># B \<longrightarrow> v \<in># C \<longrightarrow>
               (t,u) \<in> R2 \<longrightarrow> (u,v) \<in> R1 \<longrightarrow> (t,v) \<in> R2)"

text{* Local splitting *}

definition locally_splittable :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> 'a relation \<Rightarrow> 'a multiset \<Rightarrow> 'a multiset \<Rightarrow> bool" where
  "locally_splittable R1 R2 R3 A B \<equiv> (\<forall>t u. t \<in># A \<longrightarrow> u \<in># B \<longrightarrow>
               (t,u) \<in> R1 \<longrightarrow> (t,u) \<in> R2 \<or> (t,u) \<in> R3)"

text{* included + compatible \<Rightarrow> transitive *}

lemma in_cl_tr:
  assumes "included R1 R2"
  and "compatible_l R2 R1"
  shows "transitive R1"
proof-
{
  fix x y z
  assume s_x_y: "(x,y) \<in> R1" and s_y_z: "(y,z) \<in> R1"
  from assms(2)[unfolded compatible_l_def, rule_format,
   OF assms(1)[unfolded included_def, rule_format, OF s_x_y] s_y_z]
  have "(x,z) \<in> R1" by assumption
}
thus ?thesis unfolding transitive_def by fast
qed

lemma in_cr_tr:
  assumes "included R1 R2"
  and "compatible_r R2 R1"
  shows "transitive R1"
proof-
{
  fix x y z
  assume s_x_y: "(x,y) \<in> R1" and s_y_z: "(y,z) \<in> R1"
  from assms(2)[unfolded compatible_r_def, rule_format,
   OF s_x_y assms(1)[unfolded included_def, rule_format, OF s_y_z]]
  have "(x,z) \<in> R1" by assumption
}
thus ?thesis unfolding transitive_def by fast
qed

text{* If a property holds globally, it also holds locally. Obviously.*}

lemma r_lr:
  assumes "reflexive R"
  shows "locally_reflexive R A"
using assms unfolding reflexive_def and locally_reflexive_def by simp

lemma tr_ltr:
  assumes "transitive R"
  shows "locally_transitive R A B C"
using assms unfolding transitive_def and locally_transitive_def by fast

lemma in_lin:
  assumes "included R1 R2"
  shows "locally_included R1 R2 A B"
using assms unfolding included_def and locally_included_def by simp

lemma cl_lcl:
  assumes "compatible_l R1 R2"
  shows "locally_compatible_l R1 R2 A B C"
using assms unfolding compatible_l_def and locally_compatible_l_def by simp

lemma cr_lcr:
  assumes "compatible_r R1 R2"
  shows "locally_compatible_r R1 R2 A B C"
using assms unfolding compatible_r_def and locally_compatible_r_def by simp

text{* If a predicate holds on a set then it holds on
  all the subsets: *}

lemma lr_trans_l:
  assumes "locally_reflexive R (A + B)"
  shows "locally_reflexive R A"
using assms unfolding locally_reflexive_def
by auto

lemma li_trans_l:
  assumes "locally_irreflexive R (A + B)"
  shows "locally_irreflexive R A"
using assms unfolding locally_irreflexive_def
by auto

lemma ls_trans_l:
  assumes "locally_symmetric R (A + B)"
  shows "locally_symmetric R A"
using assms unfolding locally_symmetric_def
by auto

lemma las_trans_l:
  assumes "locally_antisymmetric R (A + B)"
  shows "locally_antisymmetric R A"
using assms unfolding locally_antisymmetric_def
by auto

lemma lt_trans_l:
  assumes "locally_transitive R (A + B) (C + D) (E + F)"
  shows "locally_transitive R A C E"
using assms[unfolded locally_transitive_def, rule_format]
unfolding locally_transitive_def by auto

lemma lin_trans_l: 
  assumes "locally_included R1 R2 (A + B) (C + D)"
  shows "locally_included R1 R2 A C"
using assms unfolding locally_included_def by auto

lemma lcl_trans_l: 
  assumes "locally_compatible_l R1 R2 (A + B) (C + D) (E + F)"
  shows "locally_compatible_l R1 R2 A C E"
using assms[unfolded locally_compatible_l_def, rule_format]
unfolding locally_compatible_l_def by auto

lemma lcr_trans_l: 
  assumes "locally_compatible_r R1 R2 (A + B) (C + D) (E + F)"
  shows "locally_compatible_r R1 R2 A C E"
using assms[unfolded locally_compatible_r_def, rule_format]
unfolding locally_compatible_r_def by auto

lemma lsp_trans_l: 
  assumes "locally_splittable R1 R2 R3 (A + B) (C + D)"
  shows "locally_splittable R1 R2 R3 A C"
using assms unfolding locally_splittable_def by auto

lemma lr_trans_r:
  assumes "locally_reflexive R (A + B)"
  shows "locally_reflexive R B"
using assms unfolding locally_reflexive_def
by auto

lemma li_trans_r:
  assumes "locally_irreflexive R (A + B)"
  shows "locally_irreflexive R B"
using assms unfolding locally_irreflexive_def
by auto

lemma ls_trans_r:
  assumes "locally_symmetric R (A + B)"
  shows "locally_symmetric R B"
using assms unfolding locally_symmetric_def
by auto

lemma las_trans_r:
  assumes "locally_antisymmetric R (A + B)"
  shows "locally_antisymmetric R B"
using assms unfolding locally_antisymmetric_def
by auto

lemma lt_trans_r:
  assumes "locally_transitive R (A + B) (C + D) (E + F)"
  shows "locally_transitive R B D F"
using assms[unfolded locally_transitive_def, rule_format]
unfolding locally_transitive_def
by auto

lemma lin_trans_r: 
  assumes "locally_included R1 R2 (A + B) (C + D)"
  shows "locally_included R1 R2 B D"
using assms unfolding locally_included_def by auto

lemma lcl_trans_r:
  assumes "locally_compatible_l R1 R2 (A + B) (C + D) (E + F)"
  shows "locally_compatible_l R1 R2 B D F"
using assms[unfolded locally_compatible_l_def, rule_format]
unfolding locally_compatible_l_def by auto

lemma lcr_trans_r: 
  assumes "locally_compatible_r R1 R2 (A + B) (C + D) (E + F)"
  shows "locally_compatible_r R1 R2 B D F"
using assms[unfolded locally_compatible_r_def, rule_format]
unfolding locally_compatible_r_def by auto

lemma lsp_trans_r:
  assumes "locally_splittable R1 R2 R3 (A + B) (C + D)"
  shows "locally_splittable R1 R2 R3 B D"
using assms unfolding locally_splittable_def by auto

lemma lr_minus:
  assumes "locally_reflexive R A"
  shows "locally_reflexive R (A - B)"
using assms unfolding locally_reflexive_def
by auto

lemma li_minus:
  assumes "locally_irreflexive R A"
  shows "locally_irreflexive R (A - B)"
using assms unfolding locally_irreflexive_def
by auto

lemma ls_minus:
  assumes "locally_symmetric R A"
  shows "locally_symmetric R (A - B)"
using assms unfolding locally_symmetric_def
by auto

lemma las_minus:
  assumes "locally_antisymmetric R A"
  shows "locally_antisymmetric R (A - B)"
using assms unfolding locally_antisymmetric_def
by auto

lemma lt_minus:
  assumes "locally_transitive R A C E"
  shows "locally_transitive R (A - B) (C - D) (E - F)"
using assms[unfolded locally_transitive_def, rule_format]
unfolding locally_transitive_def
by (metis Multiset.diff_le_self mset_leD) (* thanks sledgehammer! *)

lemma lin_minus: 
  assumes "locally_included R1 R2 A C"
  shows "locally_included R1 R2 (A - B) (C - D)"
using assms unfolding locally_included_def by auto

lemma lcl_minus:
  assumes "locally_compatible_l R1 R2 A C E"
  shows "locally_compatible_l R1 R2 (A - B) (C - D) (E - F)"
using assms[unfolded locally_compatible_l_def, rule_format]
unfolding locally_compatible_l_def
by (metis Multiset.diff_le_self mset_leD)

lemma lcr_minus: 
  assumes "locally_compatible_r R1 R2 A C E"
  shows "locally_compatible_r R1 R2 (A - B) (C - D) (E - F)"
using assms[unfolded locally_compatible_r_def, rule_format]
unfolding locally_compatible_r_def
by (metis Multiset.diff_le_self mset_leD)

lemma lsp_minus:
  assumes "locally_splittable R1 R2 R3 A C"
  shows "locally_splittable R1 R2 R3 (A - B) (C - D)"
using assms unfolding locally_splittable_def by auto

end