(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* Term Rewrite Systems *}
theory Trs
imports Term "~~/src/HOL/Library/Infinite_Set"
begin

text {* A theory on first-order term rewrite systems (TRSs). *}

text {* A rewrite rule is a pair of terms. A TRS is a set of
rewrite rules. *}
type_synonym ('f, 'v) rule = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) trs  = "('f, 'v) rule set"

subsection {* Variables of Rules *}

definition
  vars_rule :: "('f, 'v) rule \<Rightarrow> 'v set"
where
  "vars_rule r = vars_term (fst r) \<union> vars_term (snd r)"

subsection {* Variables of TRSs *}

definition
  vars_trs :: "('f, 'v) trs \<Rightarrow> 'v set"
where
  "vars_trs R \<equiv> \<Union>r\<in>R. vars_rule r"

lemma finite_trs_has_finite_vars:
  assumes "finite R" shows "finite (vars_trs R)"
using assms unfolding vars_trs_def vars_rule_def by simp

lemmas vars_defs = vars_trs_def vars_rule_def

subsection {* Function Symbols of Rules *}

definition
  funs_rule :: "('f, 'v) rule \<Rightarrow> 'f set"
where
  "funs_rule r \<equiv> funs_term (fst r) \<union> funs_term (snd r)"


text {* The same including arities. *}
definition
  funas_rule :: "('f, 'v) rule \<Rightarrow> 'f sig"
where
  "funas_rule r \<equiv> funas_term (fst r) \<union> funas_term (snd r)"


subsection {* Function Symbols of TRSs *}

definition
  funs_trs :: "('f, 'v) trs \<Rightarrow> 'f set"
where
  "funs_trs R \<equiv> \<Union>r\<in>R. funs_rule r"

definition
  funas_trs :: "('f, 'v) trs \<Rightarrow> 'f sig"
where
  "funas_trs R \<equiv> \<Union>r\<in>R. funas_rule r"

lemma funas_empty[simp]: "funas_trs {} = {}" unfolding funas_trs_def by simp

lemmas funs_defs = funs_trs_def funs_rule_def
lemmas funas_defs = funas_trs_def funas_rule_def

text {* A function symbol is said to be \emph{defined} (w.r.t.\ to a given
TRS) if it occurs as root of some left-hand side. *}
definition
  defined :: "('f, 'v) trs \<Rightarrow> ('f \<times> nat) \<Rightarrow> bool"
where
  "defined R fn \<equiv> \<exists>l r. (l, r) \<in> R \<and> root l = Some (fst fn) \<and> num_args l = (snd fn)"

text {* The set of all defined symbols. *}
definition
  "defs" :: "('f, 'v) trs \<Rightarrow> ('f \<times> nat) set"
where
  "defs R \<equiv> {(f, n). defined R (f, n)}"


text {* The remaining symbols are called \emph{constructor symbols}, since
they are only used to construct data. *}
definition cons :: "('f, 'v) trs \<Rightarrow> ('f \<times> nat) set" where "cons R \<equiv> funas_trs R - defs R"

text {* Closure under contexts of a binary relation on terms. *}
inductive_set
  ctxt_closure :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs" ("CC")
  for R :: "('f, 'v) trs"
where
  ctxt_closureI[intro]: "(s, t) \<in> R \<Longrightarrow> (C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> CC R"

lemma subset_ctxt_closure: "R \<subseteq> CC R"
proof
  fix s t assume "(s, t) \<in> R"
  from ctxt_closureI[OF this, of \<box>] show "(s, t) \<in> CC R" by simp
qed

definition
  ctxt_closed :: "('f, 'v) trs \<Rightarrow> bool"
where
  "ctxt_closed R = (CC R \<subseteq> R)"

lemma ctxt_closed_correctness: "ctxt_closed r \<Longrightarrow> r = CC r"
proof -
  assume "ctxt_closed r"
  hence a: "CC r \<subseteq> r" by (unfold ctxt_closed_def)
  from subset_ctxt_closure have b: "r \<subseteq> CC r" .
  from a b show "r = CC r" by simp
qed

lemma ctxt_closed_imp_ctxt:
 assumes "ctxt_closed R" and "(s,t) \<in> R" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> R"
proof -
 from `ctxt_closed R` have "R = CC R" by (rule ctxt_closed_correctness)
 from `(s,t) \<in> R` have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> CC R" by (rule ctxt_closureI)
 with `R = CC R` show ?thesis by simp
qed

lemma ctxt_closed_R_imp_supt_R_distr:
  assumes "ctxt_closed R" and "s \<rhd> t" and "(t,u) \<in> R" shows "\<exists>t. (s,t) \<in> R \<and> t \<rhd> u"
proof -
  from `s \<rhd> t` obtain C where "C \<noteq> \<box>" and "C\<langle>t\<rangle> = s" by auto
  from `ctxt_closed R` and `(t,u) \<in> R`
    have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
  from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" by auto
  from RCtCu have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` .
  from this and `C\<langle>u\<rangle> \<rhd> u` show ?thesis by auto
qed

lemma ctxt_closed_supt_R_imp_R_supt:
   assumes "ctxt_closed R" and "s \<rhd> t" and "(t,u) \<in> R" shows "\<exists>t. (s,t) \<in> R \<and> t \<rhd> u"
proof -
  from `s \<rhd> t` obtain "C" where "C \<noteq> \<box>" and "C\<langle>t\<rangle> = s" by auto
  from `ctxt_closed R` and `(t,u) \<in> R`
    have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
  from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" by (rule nectxt_imp_supt_ctxt)
  from `(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R` have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` by simp
  from this and `C\<langle>u\<rangle> \<rhd> u` show ?thesis by auto
qed

lemma ctxt_closed_R_imp_R_qc_supt:
  assumes "ctxt_closed R" shows "quasi_commute R (op \<rhd>)" (is "quasi_commute ?r ?s")
proof -
  have "?s O ?r \<subseteq> ?r O (?r \<union> ?s)^*"
  proof (rule subsetI2)
    fix x y assume "(x,y) \<in> ?s O ?r"
    then obtain z where "x \<rhd> z" and "(z,y) \<in> R" by best
    with `ctxt_closed R` obtain w where "(x,w) \<in> R" and "w \<rhd> y"
      using ctxt_closed_supt_R_imp_R_supt[of R x z y] by auto
    hence "(w,y) \<in> ?s" by simp
    hence "(w,y) \<in> ?r \<union> ?s" by auto
    hence "(w,y) \<in> (?r \<union> ?s)^*" by auto
    with `(x,w) \<in> R` show "(x,y) \<in> ?r O (?r \<union> ?s)^*" by auto
  qed
  thus ?thesis unfolding quasi_commute_def by force
qed

lemma ctxt_closure_idemp[simp]: "CC (CC R) = (CC R)"
proof
 show "CC(CC R) \<subseteq> CC R"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CC(CC R)"
  thus "(s,t) \<in> CC R"
  proof
   fix s' t' C assume s:"s = C\<langle>s'\<rangle>" and t: "t = C\<langle>t'\<rangle>" and "(s',t') \<in> CC R"
   from `(s',t') \<in> CC R` show "(s,t) \<in> CC R"
   proof
    fix s'' t'' D assume s': "s' = D\<langle>s''\<rangle>" and t': "t' = D\<langle>t''\<rangle>" and "(s'',t'') \<in> R"
    hence Rst'': "(s'',t'') \<in> R" by auto
    with s t s' t' have st: "s = (C \<circ>c D)\<langle>s''\<rangle>" "t = (C \<circ>c D)\<langle>t''\<rangle>" by auto
    from Rst'' show ?thesis unfolding st by (rule ctxt_closureI)
   qed
  qed
 qed
next
 show "CC R \<subseteq> CC(CC R)"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CC R"
  hence "(\<box>\<langle>s\<rangle>,\<box>\<langle>t\<rangle>) \<in> CC(CC R)" by (rule ctxt_closureI)
  thus "(s,t) \<in> CC(CC R)" by auto
 qed
qed

text {* Let~$R$ be a relation on terms that is closed under contexts. If~$R$
is well-founded then $R \cup \rhd$ is well-founed. *}
lemma SN_imp_SN_union_supt:
  assumes "SN R" and "ctxt_closed R"
  shows "SN (R \<union> op \<rhd>)"
proof -
  from `ctxt_closed R` have "quasi_commute R (op \<rhd>)" by (rule ctxt_closed_R_imp_R_qc_supt)
  have "SN (op \<rhd>)" by (rule SN_supt)
  from `SN R` and `SN (op \<rhd>)` and `quasi_commute R (op \<rhd>)`
  show ?thesis by (rule quasi_commute_imp_SN)
qed

inductive_set
  subst_closure :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs" ("CS")
  for R
where
 subst_closureI[intro]: "(s, t) \<in> R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> CS R"


lemma subst_closureI2: "(l,r) \<in> R \<Longrightarrow> s = l \<cdot> \<sigma> \<Longrightarrow> t = r \<cdot> \<sigma> \<Longrightarrow> (s,t) \<in> CS R"
  using subst_closure.subst_closureI by blast

lemma subset_subst_closure: "R \<subseteq> CS R"
proof
  fix s t assume "(s, t) \<in> R"
  from subst_closureI[OF this, of Var] show "(s, t) \<in> CS R" by simp
qed

definition
  subst_closed :: "('f, 'v) trs \<Rightarrow> bool"
where 
  "subst_closed R = (CS R \<subseteq> R)"

lemma subst_closed_correctness: assumes "subst_closed R" shows "R = CS R"
using assms subset_subst_closure unfolding subst_closed_def by auto

lemma subst_closed_union: 
  assumes r: "subst_closed r"
  and s: "subst_closed s"
  shows "subst_closed (r \<union> s)"
unfolding subst_closed_def
proof (rule subsetI2)
  fix x y	
  assume "(x,y) \<in> CS (r \<union> s)"
  thus "(x,y) \<in> r \<union> s"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r \<union> s"
    from mem show ?thesis
    proof
      assume "(xx,yy) \<in> r"
      hence "(x,y) \<in> CS r" by (simp add: x y, rule subst_closureI)
      thus ?thesis using r unfolding subst_closed_def by auto
    next
      assume "(xx,yy) \<in> s"
      hence "(x,y) \<in> CS s" by (simp add: x y, rule subst_closureI)
      thus ?thesis using s unfolding subst_closed_def by auto
    qed
  qed
qed

lemma subst_closed_comp: 
  assumes r: "subst_closed r"
  and s: "subst_closed s"
  shows "subst_closed (r O s)"
unfolding subst_closed_def
proof (rule subsetI2)
  fix x y	
  assume "(x,y) \<in> CS (r O s)"
  thus "(x,y) \<in> r O s"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r O s"    
    from mem obtain zz where xx: "(xx,zz) \<in> r" and yy: "(zz,yy) \<in> s" by auto
    let ?z = "zz \<cdot> sig"
    from xx have "(x,?z) \<in> CS r" by (simp add: x, rule subst_closureI)
    hence x: "(x,?z) \<in> r" using r unfolding subst_closed_def by auto
    from yy have "(?z,y) \<in> CS s" 
      by (simp add: y, rule subst_closureI)
    hence y: "(?z,y) \<in> s" using s unfolding subst_closed_def by auto
    from x y show "(x,y) \<in> r O s" by auto
  qed
qed

lemma subst_closed_rtrancl: 
  assumes r: "subst_closed r"
  shows "subst_closed (r^*)"
unfolding subst_closed_def
proof (rule subsetI2)
  fix x y	
  assume "(x,y) \<in> CS (r^*)"
  thus "(x,y) \<in> r^*"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r^*"
    from mem have "(xx \<cdot> sig, yy \<cdot> sig) \<in> r^*"
    proof (induct, simp)
      case (step yy zz)
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> CS r" by auto
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> r" using r unfolding subst_closed_def by auto
      with step show ?case by auto
    qed
    with x y show "(x,y) \<in> r^*" by simp
  qed
qed



lemma subst_closed_trancl: 
  assumes r: "subst_closed r"
  shows "subst_closed (r^+)"
unfolding subst_closed_def
proof (rule subsetI2)
  fix x y	
  assume "(x,y) \<in> CS (r^+)"
  thus "(x,y) \<in> r^+"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r^+"
    from mem have "(xx \<cdot> sig, yy \<cdot> sig) \<in> r^+"
    proof (induct)
      case (step yy zz)
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> CS r" by auto
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> r" using r unfolding subst_closed_def by auto
      with step show ?case by auto
    next
      case (base yy)
      hence "(xx \<cdot> sig, yy \<cdot> sig) \<in> CS r" by auto
      hence "(xx \<cdot> sig, yy \<cdot> sig) \<in> r" using r unfolding subst_closed_def by auto
      thus ?case by auto
    qed
    with x y show "(x,y) \<in> r^+" by simp
  qed
qed

lemma subst_closed_decompose:
  assumes "(s,t) \<in> (CS R)" shows "\<exists>\<sigma> u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and> (u,v) \<in> R"
using assms by (cases rule: subst_closure.cases) force


lemma stable_loop_imp_not_SN: assumes stable: "subst_closed r" and steps: "(s,s \<cdot> sig) \<in> r^+"
  shows "\<not> SN_elt r s"
proof -
  let ?f =  "\<lambda> i. s \<cdot> (power.power Var op \<circ>s sig i)"
  have main: "\<And> i. (?f i, ?f (Suc i)) \<in> r^+"
  proof -
    fix i
    show "(?f i, ?f (Suc i)) \<in> r^+"
    proof (induct i, simp add: steps)
      case (Suc i)
      from Suc subst_closed_trancl[OF stable] have step: "(?f i \<cdot> sig, ?f (Suc i) \<cdot> sig) \<in> r^+" unfolding subst_closed_def
        by auto
      let ?sigg = "power.power Var op \<circ>s sig i"
      let ?siggs = "power.power Var op \<circ>s sig (Suc i)"
      have idi: "?sigg \<circ>s sig = sig \<circ>s ?sigg" by (rule subst_monoid_mult.power_commutes)
      have idsi: "?siggs \<circ>s sig = sig \<circ>s ?siggs" by (rule subst_monoid_mult.power_commutes)
      have "?f i \<cdot> sig = s \<cdot> ?sigg  \<circ>s sig" by simp
      also have "\<dots> = ?f (Suc i)" unfolding idi by simp
      finally have one: "?f i \<cdot> sig = ?f (Suc i)" .
      have "?f (Suc i) \<cdot> sig = s \<cdot> ?siggs \<circ>s sig" by simp
      also have "\<dots> = ?f (Suc (Suc i))" unfolding idsi by simp
      finally have two: "?f (Suc i) \<cdot> sig = ?f (Suc (Suc i))" by simp
      show ?case using one two step by simp
    qed
  qed
  hence "\<not> SN_elt (r^+) (?f 0)" unfolding SN_elt_def by best
  thus ?thesis using SN_elt_imp_SN_elt_trancl by force
qed

lemma subst_subst_closed_subseteq_subst_closed:
  fixes R::"('f, 'v) trs" shows "CS (CS R) \<subseteq> CS R"
proof (rule subsetI2)
  fix s t :: "('f, 'v) term" assume a: "(s,t) \<in> CS (CS R)"
  hence "\<exists>(\<sigma>::('f, 'v) subst) u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and> (u,v) \<in> CS R"
    using subst_closed_decompose by auto
  hence "\<exists>(\<sigma>::('f, 'v) subst) u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and>
    (\<exists>(\<tau>::('f, 'v) subst) l r. (u = l\<cdot>\<tau>) \<and> (v = r\<cdot>\<tau>) \<and> (l,r) \<in> R)"
    using subst_closed_decompose[of _ _ R] by best
  hence "\<exists>(\<sigma>::('f, 'v) subst) (\<tau>::('f, 'v) subst) l r.
    (s = l \<cdot> (\<tau> \<circ>s \<sigma>)) \<and> (t = r \<cdot> (\<tau> \<circ>s \<sigma>)) \<and> (l, r) \<in> R" by auto
  thus "(s,t) \<in> (CS R)" by blast
qed

lemma subst_closed_idemp[simp]: "CS (CS R) = CS R"
using subst_subst_closed_subseteq_subst_closed subset_subst_closure by auto

lemma subst_closed_closure_subst[simp]: "subst_closed (CS R)"
unfolding subst_closed_def by simp

lemma supteq_subst_closed: "subst_closed (op \<unrhd>)"
unfolding subst_closed_def
proof (rule subsetI2)
  fix s t :: "('f, 'v) term" assume "(s, t) \<in> CS (op \<unrhd>)"
  then obtain u v and \<sigma> :: "('f, 'v) subst" where "u \<unrhd> v" and "u \<cdot> \<sigma> = s" and "v \<cdot> \<sigma> = t"
    by (cases rule: subst_closure.cases) auto
  with supteq_stable[of u v \<sigma>] show "s \<unrhd> t" by simp
qed

lemma supt_subst_closed: "subst_closed (op \<rhd>)"
unfolding subst_closed_def
proof (rule subsetI2)
  fix s t :: "('f, 'v) term"  assume "(s, t) \<in> CS (op \<rhd>)"
  then obtain u v and \<sigma> :: "('f, 'v) subst" where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "u \<rhd> v"
    by (cases rule: subst_closure.cases) auto
  thus "s \<rhd> t" unfolding s t using supt_stable by simp
qed

text {*
The single-step relation induced by a TRS~@{text R} is the relation @{text R} closed under
substitutions and contexts.
*}
inductive_set
  rstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term ars"
  for R
where
  id[intro]: "(s, t) \<in> R \<Longrightarrow> (s, t) \<in> rstep R" |
  subst[intro]: "(s, t) \<in> rstep R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep R" |
  ctxt[intro]: "(s, t) \<in> rstep R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> rstep R"

lemma rstep_CC_CS: "rstep R = CC (CS R)"
proof
  show "rstep R \<subseteq> CC (CS R)"
  proof
    fix s t assume "(s, t) \<in> rstep R" thus "(s, t) \<in> CC (CS R)"
    proof (induct)
      case (id s t)
      from ctxt_closureI[OF subst_closureI[OF `(s, t) \<in> R`], of \<box> Var] show ?case using id by simp
    next
      case (subst s t \<sigma>)
      from `(s, t) \<in> CC (CS R)` obtain C u v where "(u, v) \<in> CS R"
        and s: "s = C\<langle>u\<rangle>" and t: "t = C\<langle>v\<rangle>" by cases auto
      from subst_closureI[OF `(u, v) \<in> CS R`, of \<sigma>] have "(u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> CS R" by simp
      from ctxt_closureI[OF this, of "C \<cdot>c \<sigma>"] show ?case unfolding s t by simp
    next
      case (ctxt s t C)
      from ctxt_closureI[OF `(s, t) \<in> CC (CS R)`, of C] show ?case by simp
    qed
  qed
next
  show "CC (CS R) \<subseteq> rstep R"
  proof
    fix s t assume "(s, t) \<in> CC (CS R)"
    then obtain C u v where s: "s = C\<langle>u\<rangle>" and t: "t = C\<langle>v\<rangle>" and "(u, v) \<in> CS R" by cases auto
    from `(u, v) \<in> CS R` obtain \<sigma> u' v' where u: "u = u' \<cdot> \<sigma>" and v: "v = v' \<cdot> \<sigma>"
      and "(u', v') \<in> R" by (cases) auto
    from `(u', v') \<in> R` have "(u', v') \<in> rstep R" ..
    hence "(u, v) \<in> rstep R" unfolding u v ..
    thus "(s, t) \<in> rstep R" unfolding s t ..
  qed
qed

definition
  wfrstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs"
where
  "wfrstep R \<equiv>
     {(s,t). (s,t) \<in> rstep R \<and> s \<in> wf_terms (funas_trs R) \<and> t \<in> wf_terms (funas_trs R)}"

lemma ctxt_closed_rstep: "ctxt_closed (rstep R)"
unfolding ctxt_closed_def rstep_CC_CS ctxt_closure_idemp ..

lemma ctxt_closed_rstep_seq: "ctxt_closed((rstep R)^*)" unfolding ctxt_closed_def
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> CC((rstep R)^*)" thus "(s,t) \<in> (rstep R)^*"
 proof (induct)
  case (ctxt_closureI s t C) thus ?case
  proof (induct)
   case base thus ?case by simp
  next
   case (step y z)
   hence "(C\<langle>y\<rangle>,C\<langle>z\<rangle>) \<in> rstep R" using ctxt_closed_rstep unfolding ctxt_closed_def by blast
   with `(C\<langle>s\<rangle>,C\<langle>y\<rangle>) \<in> (rstep R)^*` show ?case by simp
  qed
 qed
qed

lemma subset_rstep: "R \<subseteq> rstep R" by auto

lemma subst_closure_rstep_subset: "CS (rstep R) \<subseteq> rstep R"
proof
  fix s t assume "(s, t) \<in> CS (rstep R)"
  then obtain \<sigma> u v
    where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "(u, v) \<in> rstep R" by cases auto
  thus "(s, t) \<in> rstep R" unfolding s t by auto
qed

lemma subst_closed_rstep: "subst_closed (rstep R)"
using subst_closure_rstep_subset unfolding subst_closed_def by auto

lemma subst_closed_rstep_seq: "subst_closed ((rstep R)^*)" unfolding subst_closed_def
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> CS((rstep R)^*)" thus "(s,t) \<in> (rstep R)^*"
 proof (induct)
  case (subst_closureI s t \<sigma>) thus ?case
  proof (induct)
   case base thus ?case by simp
  next
   case (step y z)
   hence "(y \<cdot> \<sigma>,z \<cdot> \<sigma>) \<in> rstep R" using subst_closed_rstep unfolding subst_closed_def by blast
   with `(s \<cdot> \<sigma>,y \<cdot> \<sigma>) \<in> (rstep R)^*` show ?case by simp
  qed
 qed
qed

lemma comp_stable:
  assumes R: "\<And>x y \<sigma>. (x, y) \<in> R \<Longrightarrow> (x \<cdot> \<sigma>, y \<cdot> \<sigma>) \<in> R"
    and S: "\<And>x y \<sigma>. (x, y) \<in> S \<Longrightarrow> (x \<cdot> \<sigma>, y \<cdot> \<sigma>) \<in> S"
  shows "\<And>x y \<sigma>. (x, y) \<in> R O S \<Longrightarrow> (x \<cdot> \<sigma>, y \<cdot> \<sigma>) \<in> R O S"
proof -
 fix s t \<sigma> assume "(s,t) \<in> R O S"
 then obtain u where "(s,u) \<in> R" and "(u,t) \<in> S" by auto
 from R[OF `(s,u) \<in> R`] have "(s \<cdot> \<sigma>,u \<cdot> \<sigma>) \<in> R" by simp
 from S[OF `(u,t) \<in> S`] have "(u \<cdot> \<sigma>,t \<cdot> \<sigma>) \<in> S" by simp
 with `(s \<cdot> \<sigma>,u \<cdot> \<sigma>) \<in> R`
 show "(s \<cdot> \<sigma>,t \<cdot> \<sigma>) \<in> R O S" by auto
qed

lemma comp_subst_closed: assumes "subst_closed R" and "subst_closed S" shows "subst_closed(R O S)" unfolding subst_closed_def
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> CS(R O S)"
 then obtain u v \<sigma> where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "(u,v) \<in> R O S" by (cases rule: subst_closure.cases) auto
 from `(u,v) \<in> R O S` obtain w where "(u,w) \<in> R" and "(w,v) \<in> S" by auto
 from `subst_closed R` and `(u,w) \<in> R` have "(s,w \<cdot> \<sigma>) \<in> R" unfolding subst_closed_def s by auto
 from `subst_closed S` and `(w,v) \<in> S` have "(w \<cdot> \<sigma>,t) \<in> S" unfolding subst_closed_def t by auto
 with `(s,w \<cdot> \<sigma>) \<in> R` show "(s,t) \<in> R O S" by auto
qed

lemma ctxt_closed_supt_subset:
  assumes "ctxt_closed R" shows "op \<rhd> O R \<subseteq> R O op \<rhd>"
proof
  fix s t assume "(s, t) \<in> op \<rhd> O R"
  then obtain u where "s \<rhd> u" and "(u, t) \<in> R" by auto
  obtain v where "(s, v) \<in> R" and "v \<rhd> t"
    using ctxt_closed_R_imp_supt_R_distr[OF assms `s \<rhd> u` `(u, t) \<in> R`] by auto
  from `v \<rhd> t` and `(s, v) \<in> R` show "(s, t) \<in> R O op \<rhd>" by auto
qed

lemmas supt_rsteps_subset = ctxt_closed_supt_subset[OF ctxt_closed_rstep_seq]

lemma supteq_rsteps_subset: "op \<unrhd> O (rstep R)^* \<subseteq> (rstep R)^* O op \<unrhd>" (is "?S \<subseteq> ?T")
proof
  fix s t assume "(s, t) \<in> ?S"
  hence "(s, t) \<in> op \<rhd> O (rstep R)^* \<union> (rstep R)^*" by (auto simp: supt_supteq_conv)
  thus "(s, t) \<in> ?T"
  proof
    assume "(s, t) \<in> op \<rhd> O (rstep R)^*" thus ?thesis
      using supt_rsteps_subset unfolding supt_supteq_set_conv by auto
  next
    assume "(s, t) \<in> (rstep R)^*" thus ?thesis by auto
  qed
qed

lemma quasi_commute_rsteps_supt: "quasi_commute ((rstep R)^*) (op \<rhd>)"
unfolding quasi_commute_def using supt_rsteps_subset[of R] by auto

text {*
The set of minimally non-terminating Terms, i.e., non-terminating terms whose
proper subterms are terminating w.r.t.\ a given TRS.
*}
definition
  Tinf :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set"
where
  "Tinf R \<equiv> {t. \<not> SN_elt (rstep R) t \<and> (\<forall>s\<lhd>t. SN_elt (rstep R) s)}"

lemma not_SN_imp_subt_Tinf:
  assumes "\<not> SN_elt (rstep R) t" shows "\<exists>s\<unlhd>t. s \<in> Tinf R"
using assms proof (induct rule: subterm_induct)
  case (subterm t) show ?case
  proof (cases "\<forall>s\<lhd>t. SN_elt (rstep R) s")
    case True with subterm show ?thesis by (auto simp: Tinf_def)
  next
    case False
    then obtain s where "t \<rhd> s" and "\<not> SN_elt (rstep R) s" by best
    with subterm obtain u where "u \<unlhd> s" and "u \<in> Tinf R" by auto
    from `t \<rhd> s` and `s \<unrhd> u` have "t \<rhd> u" by (rule supt_supteq_trans)
    with `u \<in> Tinf R` show ?thesis by auto
  qed
qed

lemma not_SN_imp_Tinf:
  assumes "\<not> SN (rstep R)" shows "\<exists>t. t \<in> Tinf R"
using assms not_SN_imp_subt_Tinf unfolding SN_def by auto

(* @{term "Tinf"} characterizes termination. *)
lemma Tinf_SN_conv: "(Tinf R = {}) = (SN (rstep R))"
proof
  assume "Tinf R = {}" thus "SN (rstep R)" using not_SN_imp_Tinf by blast
next
  assume "SN (rstep R)" thus "Tinf R = {}" unfolding Tinf_def SN_defs by blast
qed

definition
  rstep_r_p_s :: "('f, 'v) trs \<Rightarrow> ('f, 'v) rule \<Rightarrow> pos \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) trs"
where
 "rstep_r_p_s R r p \<sigma> \<equiv> {(s,t).
 let C = ctxt_of_pos_term p s in p \<in> poss s \<and> r \<in> R \<and> (C\<langle>fst r\<cdot>\<sigma>\<rangle> = s) \<and> (C\<langle>snd r\<cdot>\<sigma>\<rangle> = t)}"

lemma "p \<in> poss t \<Longrightarrow> hole_pos(ctxt_of_pos_term p t) = p" by (induct p t rule: ctxt_of_pos_term.induct) auto

lemma ctxt_of_pos_of_ctxt: "ctxt_of_pos_term (hole_pos C) (C\<langle>s\<rangle>) = C" by (induct C) auto

lemma ep_pos_term: "\<epsilon> \<in> poss t" by (induct rule: term_induct) auto

(* The position of the hole in a context C is contained in the set
of positions of any term obtained by applying C. *)
lemma hole_pos_in_poss: "hole_pos C \<in> poss (C\<langle>t\<rangle>)"
by (induct C arbitrary: t) (auto simp: ep_pos_term)

lemma rstepE[elim]:
  assumes "(s, t) \<in> rstep R"
    and imp: "\<And>C \<sigma> l r. \<lbrakk>(l, r) \<in> R; s = C\<langle>l \<cdot> \<sigma>\<rangle>; t = C\<langle>r \<cdot> \<sigma>\<rangle>\<rbrakk> \<Longrightarrow> P"
  shows "P"
proof -
  from assms[unfolded rstep_CC_CS]
    obtain C u v where s: "s = C\<langle>u\<rangle>" and t: "t = C\<langle>v\<rangle>" and "(u, v) \<in> CS R" by cases auto
  from `(u, v) \<in> CS R`
    obtain \<sigma> u' v' where u: "u = u' \<cdot> \<sigma>" and v: "v = v' \<cdot> \<sigma>" and "(u', v') \<in> R" by cases auto
  from imp[OF `(u', v') \<in> R`, unfolded s t u v] show "P" by simp
qed

lemma rstepI[intro]:
  assumes "(l, r) \<in> R" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "(s, t) \<in> rstep R"
using assms by auto

lemma rstep_iff_rstep_r_p_s:
  "(s,t) \<in> rstep R = (\<exists>l r p \<sigma>. (s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>)" (is "?lhs = ?rhs")
proof
  assume "(s, t) \<in> rstep R"
  then obtain C \<sigma> l r where s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and "(l, r) \<in> R" by auto
  let ?p = "hole_pos C"
  let ?C = "ctxt_of_pos_term ?p s"
  have C: "ctxt_of_pos_term ?p s = C" unfolding s by (induct C) simp_all
  have "?p \<in> poss s" unfolding s by simp
  moreover have "(l, r) \<in> R" by fact
  moreover have "?C\<langle>l \<cdot> \<sigma>\<rangle> = s" unfolding C by (simp add: s)
  moreover have "?C\<langle>r \<cdot> \<sigma>\<rangle> = t" unfolding C by (simp add: t)
  ultimately show "?rhs" unfolding rstep_r_p_s_def Let_def by auto
next
  assume "?rhs"
  then obtain l r p \<sigma> where "p \<in> poss s"
    and "(l, r) \<in> R"
    and s[symmetric]: "(ctxt_of_pos_term p s)\<langle>l \<cdot> \<sigma>\<rangle> = s"
    and t[symmetric]: "(ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle> = t"
    unfolding rstep_r_p_s_def Let_def by auto
  thus "?lhs" by auto
qed


text {* Rewriting steps below the root position. *}
definition
  nrrstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs"
where
  "nrrstep R \<equiv> {(s,t). \<exists>r i ps \<sigma>. (s,t) \<in> rstep_r_p_s R r (i<#ps) \<sigma>}"

text {* An alternative characterisation of non-root rewrite steps. *}
lemma nrrstep_def':
  "nrrstep R = {(s, t). \<exists>l r C \<sigma>. (l, r) \<in> R \<and> C \<noteq> \<box> \<and> s = C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = C\<langle>r\<cdot>\<sigma>\<rangle>}"
  (is "?lhs = ?rhs")
proof
  show "?lhs \<subseteq> ?rhs"
  proof (rule subsetI2)
    fix s t assume "(s, t) \<in> nrrstep R"
    then obtain l r i ps \<sigma> where step: "(s, t) \<in> rstep_r_p_s R (l, r) (i <# ps) \<sigma>"
      unfolding nrrstep_def by best
    let ?C = "ctxt_of_pos_term (i <# ps) s"
    from step have"i <# ps \<in> poss s" and "(l, r) \<in> R" and "s = ?C\<langle>l\<cdot>\<sigma>\<rangle>" and "t = ?C\<langle>r\<cdot>\<sigma>\<rangle>"
      unfolding rstep_r_p_s_def Let_def by auto
    moreover from `i <# ps \<in> poss s` have "?C \<noteq> \<box>" by (induct s) auto
    ultimately show "(s, t) \<in> ?rhs" by auto
  qed
next
  show "?rhs \<subseteq> ?lhs"
  proof (rule subsetI2)
    fix s t assume "(s, t) \<in> ?rhs"
    then obtain l r C \<sigma> where in_R: "(l, r) \<in> R" and "C \<noteq> \<box>"
      and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    from `C \<noteq> \<box>` obtain i p where ip: "hole_pos C = i <# p" by (induct C) auto
    have "i <# p \<in> poss s" using hole_pos_poss[of C] unfolding s ip by simp
    hence C: "C = ctxt_of_pos_term (i <# p) s" unfolding s ip[symmetric]
      by (simp add: ctxt_of_pos_of_ctxt)
    from `i <# p \<in> poss s` in_R s t have "(s, t) \<in> rstep_r_p_s R (l, r) (i <# p) \<sigma>"
      unfolding rstep_r_p_s_def C by simp
    thus "(s, t) \<in> nrrstep R" unfolding nrrstep_def by best
  qed
qed

text {* Rewriting step at the root position. *}
definition
  rrstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs"
where
  "rrstep R \<equiv> {(s,t). \<exists>r \<sigma>. (s,t) \<in> rstep_r_p_s R r \<epsilon> \<sigma>}"

text {* An alternative characterisation of root rewrite steps. *}
lemma rrstep_def': "rrstep R = {(s, t). \<exists>l r \<sigma>. (l, r) \<in> R \<and> s = l\<cdot>\<sigma> \<and> t = r\<cdot>\<sigma>}"
  (is "_ = ?rhs")
by (auto simp: rrstep_def rstep_r_p_s_def)


lemma subst_closed_rrstep: "subst_closed (rrstep R)"
unfolding subst_closed_def
proof
  fix ss ts
  assume "(ss,ts) \<in> CS (rrstep R)"
  thus "(ss,ts) \<in> rrstep R"
  proof
    fix s t \<sigma>
    assume ss: "ss = s \<cdot> \<sigma>" and ts: "ts = t \<cdot> \<sigma>" and step: "(s,t) \<in> rrstep R"
    from step obtain l r \<delta> where lr: "(l,r) \<in> R" and s: "s = l \<cdot> \<delta>" and t: "t = r \<cdot> \<delta>" unfolding rrstep_def' by auto
    obtain sig where "sig = \<delta> \<circ>s \<sigma>" by auto
    with ss s ts t have "ss = l \<cdot> sig" and "ts = r \<cdot> sig" by simp+
    with lr show "(ss,ts) \<in> rrstep R" unfolding rrstep_def' by (auto simp: Let_def)
  qed
qed

lemma rstep_iff_rrstep_or_nrrstep: "rstep R = (rrstep R \<union> nrrstep R)"
proof
 show "rstep R \<subseteq> rrstep R \<union> nrrstep R"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> rstep R"
  then obtain l r p \<sigma> where rstep_rps: "(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
   by (auto simp: rstep_iff_rstep_r_p_s)
  thus "(s,t) \<in> rrstep R \<union> nrrstep R" unfolding rrstep_def nrrstep_def by (cases p) auto
 qed
next
 show "rrstep R \<union> nrrstep R \<subseteq> rstep R"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> rrstep R \<union> nrrstep R"
  thus "(s,t) \<in> rstep R" by (auto simp: rrstep_def nrrstep_def rstep_iff_rstep_r_p_s)
 qed
qed

lemma rstep_i_pos_imp_rstep_arg_i_pos:
 assumes nrrstep: "(Fun f ss,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>"
 shows "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) ps \<sigma>"
proof -
 from nrrstep obtain C where C:"C = ctxt_of_pos_term (i<#ps) (Fun f ss)"
  and pos: "(i<#ps) \<in> poss (Fun f ss)"
  and Rlr: "(l,r) \<in> R"
  and Fun: "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss"
  and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
 then obtain D where C':"C = More f (take i ss) D (drop (Suc i) ss)" by auto
 hence CFun: "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f (take i ss @ (D\<langle>l\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from pos have len: "i < length ss" by auto
 from len
 have "(take i ss @ (D\<langle>l\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>l\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with C Fun CFun have ssi: "ss!i = D\<langle>l\<cdot>\<sigma>\<rangle>" by auto
 from C' t have t': "t = Fun f (take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from len
 have "(take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with t' have "t|_<i> = (D\<langle>r\<cdot>\<sigma>\<rangle>)|_\<epsilon>" by auto
 hence subt_at: "t|_<i> = D\<langle>r\<cdot>\<sigma>\<rangle>" by simp
 from C C' have "D = ctxt_of_pos_term ps (ss!i)" by auto
 with pos Rlr ssi[symmetric] subt_at[symmetric]
 show ?thesis unfolding rstep_r_p_s_def Let_def by auto
qed

lemma ctxt_closure_is_closed_under_ctxts: "ctxt_closed (CC R)"
unfolding ctxt_closed_def by auto

lemma ctxt_closed_rstep_id[simp]: "CC (rstep R) = rstep R" by (simp add: rstep_CC_CS)

lemma subst_closed_rstep_equal_rstep: "CS (rstep R) = rstep R" 
using subst_closure_rstep_subset subst_closed_correctness unfolding subst_closed_def by blast

lemma supt_rstep_subset: "op \<rhd> O rstep R \<subseteq> rstep R O op \<rhd>"
proof (rule subsetI2)
  fix s t assume "(s,t) \<in> op \<rhd> O rstep R" thus "(s,t) \<in> rstep R O op \<rhd>"
  proof (induct s)
    case (Var x)
    hence "\<exists>u. Var x \<rhd> u \<and> (u,t) \<in> rstep R" by auto
    then obtain u where "Var x \<rhd> u" and "(u,t) \<in> rstep R" by auto
    from `Var x \<rhd> u` show ?case by (cases rule: supt.cases)
  next
    case (Fun f ss)
    then obtain u where "Fun f ss \<rhd> u" and "(u,t) \<in> rstep R" by auto
    from `Fun f ss \<rhd> u` obtain C
      where "C \<noteq> \<box>" and "C\<langle>u\<rangle> = Fun f ss" by auto
    from `C \<noteq> \<box>` have "C\<langle>t\<rangle> \<rhd> t" by (rule nectxt_imp_supt_ctxt)
    from `(u,t) \<in> rstep R` have "(C\<langle>u\<rangle>,C\<langle>t\<rangle>) \<in> rstep R" ..
    hence "(Fun f ss,C\<langle>t\<rangle>) \<in> rstep R" unfolding `C\<langle>u\<rangle> = Fun f ss` .
    with `C\<langle>t\<rangle> \<rhd> t` show ?case by auto
  qed
qed

lemma ne_rstep_seq_imp_list_of_terms:
  assumes "(s,t) \<in> (rstep R)^+"
  shows "\<exists>ts. length ts > 1 \<and> ts!0 = s \<and> ts!(length ts - 1) = t \<and>
         (\<forall>i<length ts - 1. (ts!i,ts!(Suc i)) \<in> (rstep R))" (is "\<exists>ts. _ \<and> _ \<and> _ \<and> ?P ts")
using assms proof (induct rule: trancl.induct)
  case (r_into_trancl x y)
  let ?ts = "[x,y]"
  have "length ?ts > 1" by simp
  moreover have "?ts!0 = x" by simp
  moreover have "?ts!(length ?ts - 1) = y" by simp
  moreover from r_into_rtrancl r_into_trancl have "?P ?ts" by auto
  ultimately show ?case by fast
next
  case (trancl_into_trancl x y z)
  then obtain ts where "length ts > 1" and "ts!0 = x"
    and last1: "ts!(length ts - 1) = y" and "?P ts" by auto
  let ?ts = "ts@[z]"
  have len: "length ?ts = length ts + 1" by simp
  from `length ts > 1` have "length ?ts > 1" by auto
  moreover with `ts!0 = x` have "?ts!0 = x" by (induct ts) auto
  moreover have last2: "?ts!(length ?ts - 1) = z" by simp
  moreover have "?P ?ts"
  proof (intro allI impI)
    fix i assume A: "i < length ?ts - 1"
    show "(?ts!i,?ts!(Suc i)) \<in> rstep R"
    proof (cases "i < length ts - 1")
      case True with `?P ts` A show ?thesis unfolding len unfolding nth_append by auto
    next
      case False with A have "i = length ts - 1" by simp
      with last1 `length ts > 1` have "?ts!i = y" unfolding nth_append by auto
      have "Suc i = length ?ts - 1" using `i = length ts - 1` using `length ts > 1` by auto
      with last2 have "?ts!(Suc i) = z" by auto
      from `(y,z) \<in> rstep R` show ?thesis unfolding `?ts!(Suc i) = z` `?ts!i = y` .
    qed
  qed
  ultimately show ?case by fast
qed

definition SN_rel_supt :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "SN_rel_supt R \<equiv> (SN_rel R R) \<union> (SN_rel (op \<rhd>) R)"

lemma SN_rel_imp_SN_subt:
  assumes "ctxt_closed R" and "SN_elt R s" and "s \<unrhd> t"
  shows "SN_elt R t"
proof (rule ccontr)
  assume "\<not> SN_elt R t"
  then obtain A where "A 0 = t" and "\<forall>i. (A i,A (Suc i)) \<in> R"
    unfolding SN_elt_def[where A = "R" and a = "t"] by best
  from `s \<unrhd> t` obtain C where "s = C\<langle>t\<rangle>" by auto
  let ?B = "\<lambda>i. C\<langle>A i\<rangle>"
  have "\<forall>i. (?B i,?B(Suc i)) \<in> R"
  proof
    fix i
    from `\<forall>i. (A i,A(Suc i)) \<in> R` have "(?B i,?B(Suc i)) \<in> CC(R)" using ctxt_closureI by fast
    thus "(?B i,?B(Suc i)) \<in> R" using `ctxt_closed R` unfolding ctxt_closed_def by auto
  qed
  with `A 0 = t` have "?B 0 = s \<and> (\<forall>i. (?B i,?B(Suc i)) \<in> R)" unfolding `s = C\<langle>t\<rangle>` by auto
  hence "\<not> SN_elt R s" unfolding SN_elt_def by auto
  with assms show "False" by simp
qed

lemma ctxt_closed_R_imp_supt_SN_rel_distr:
  assumes "ctxt_closed R"
    and "(s,t) \<in> (SN_rel (op \<rhd>) R)"
    and "(t,u) \<in> SN_rel R R"
  shows "(\<exists>t. (s,t) \<in> SN_rel R R \<and> (t,u) \<in> SN_rel (op \<rhd>) R)" (is "\<exists> t. _ \<and> (t,u) \<in> ?snSub")
proof -
  from `(s,t) \<in> ?snSub` have "SN_elt R s" and "s \<rhd> t" unfolding SN_rel_def supt_def by auto
  from `s \<rhd> t` obtain C where "C \<noteq> Hole" and "C\<langle>t\<rangle> = s" by best
  from `(t,u) \<in> SN_rel R R` have "(t,u) \<in> R" unfolding SN_rel_def by auto
  from `ctxt_closed R` and this have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
  with `C\<langle>t\<rangle> = s` `SN_elt R s` have one: "SN_elt R C\<langle>u\<rangle>"
    using step_preserves_SN_elt[of "C\<langle>t\<rangle>" "C\<langle>u\<rangle>" R] by blast
  from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" by auto
  with one have "(C\<langle>u\<rangle>,u) \<in> ?snSub" unfolding SN_rel_def supt_def by auto
  from RCtCu have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` .
  from this and `SN_elt R s` and `(C\<langle>u\<rangle>,u) \<in> ?snSub` show ?thesis unfolding SN_rel_def by auto
qed

lemma ctxt_closed_R_imp_SN_rel_qc_supt:
  assumes "ctxt_closed R"
  shows "quasi_commute (SN_rel R R) (SN_rel supt R)" (is "quasi_commute ?r ?s")
proof -
  have "?s O ?r \<subseteq> ?r O (?r \<union> ?s)^*" 
  proof (rule subsetI2)
    fix x y 
    assume "(x,y) \<in> ?s O ?r"
    from this obtain z where "(x,z) \<in> ?s" and "(z,y) \<in> ?r" by best
    hence "x \<rhd> z" and "(z,y) \<in> R" and "SN_elt R x" unfolding SN_rel_def supt_def by auto
    with `ctxt_closed R` obtain w where "(x,w) \<in> R" and "w \<rhd> y" using ctxt_closed_supt_R_imp_R_supt[where R = R and s = x and t = z and u = y] by auto
    from this and `SN_elt R x` have "SN_elt R w" using step_preserves_SN_elt
      unfolding supt_supteq_conv by auto
    with `w \<rhd> y` have "(w,y) \<in> ?s" unfolding SN_rel_def supt_def by auto
    with `(x,w) \<in> R` `SN_elt R x` show "(x,y) \<in> ?r O (?r \<union> ?s)^*" unfolding SN_rel_def by auto
  qed
  thus ?thesis unfolding quasi_commute_def .
qed

lemma ctxt_closed_imp_SN_SN_rel_supt:
 assumes "ctxt_closed R"
 shows "SN (SN_rel_supt R)"
proof -
  let ?r = "SN_rel R R"
  let ?s = "SN_rel (op \<rhd>) R"
  from `ctxt_closed R` have "quasi_commute ?r ?s"
    by (rule ctxt_closed_R_imp_SN_rel_qc_supt)
  have "SN (op \<rhd>)" by (rule SN_supt)
  hence "SN ?s" unfolding SN_rel_def SN_def SN_elt_def by auto
  have "SN ?r" by (rule SN_SN_rel_idemp)
  from `SN ?r` and `SN ?s` and `quasi_commute ?r ?s`
  show ?thesis unfolding SN_rel_supt_def by  (rule quasi_commute_imp_SN)
qed

lemma SN_SN_rel_supt_rstep: 
  shows "SN (SN_rel_supt (rstep R))"
proof -
  have "ctxt_closed (rstep R)" by (rule ctxt_closed_rstep)
  thus ?thesis by (rule ctxt_closed_imp_SN_SN_rel_supt)
qed

lemma nrrstep_imp_pos_term:
  "(Fun f ss,t) \<in> nrrstep R \<Longrightarrow>
    \<exists>i s. t = Fun f (ss[i:=s]) \<and> (ss!i,s) \<in> rstep R \<and> i < length ss"
proof -
 assume "(Fun f ss,t) \<in> nrrstep R"
 then obtain l r i ps \<sigma> where rstep_rps:"(Fun f ss,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>"
  unfolding nrrstep_def by auto
 then obtain C
 where "(l,r) \<in> R"
  and pos: "(i<#ps) \<in> poss (Fun f ss)"
  and C: "C = ctxt_of_pos_term (i<#ps) (Fun f ss)"
  and "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss"
  and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t"
 unfolding rstep_r_p_s_def Let_def by auto
 then obtain D where "C = More f (take i ss) D (drop (Suc i) ss)" by auto
 with t have t': "t = Fun f (take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from rstep_rps have "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) ps \<sigma>"
  by (rule rstep_i_pos_imp_rstep_arg_i_pos)
 hence rstep:"(ss!i,t|_<i>) \<in> rstep R" by (auto simp: rstep_iff_rstep_r_p_s)
 hence "(C\<langle>ss!i\<rangle>,C\<langle>t|_<i>\<rangle>) \<in> rstep R" ..
 from pos have len: "i < length ss" by auto
 from len
 have "(take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with C t' have "t|_<i> = D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 with t' have "t = Fun f (take i ss @ (t|_<i>) # drop (Suc i) ss)" by auto
 with len have "t = Fun f (ss[i:=t|_<i>])" by (auto simp: upd_conv_take_nth_drop)
 with rstep len show "\<exists>i s. t = Fun f (ss[i:=s]) \<and> (ss!i,s) \<in> rstep R \<and> i < length ss" by auto
qed

lemma rstep_cases[consumes 1, case_names root nonroot]:
 "\<lbrakk>(s,t) \<in> rstep R; (s,t) \<in> rrstep R \<Longrightarrow> P; (s,t) \<in> nrrstep R \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
by (auto simp: rstep_iff_rrstep_or_nrrstep)

lemma nrrstep_imp_rstep: "(s,t) \<in> nrrstep R \<Longrightarrow> (s,t) \<in> rstep R"
by (auto simp: rstep_iff_rrstep_or_nrrstep)

lemma nrrstep_imp_Fun: "(s,t) \<in> nrrstep R \<Longrightarrow> \<exists>f ss. s = Fun f ss"
proof -
 assume "(s,t) \<in> nrrstep R"
 then obtain i ps where "i<#ps \<in> poss s"
 unfolding nrrstep_def rstep_r_p_s_def Let_def by auto
 thus "\<exists>f ss. s = Fun f ss" by (cases s) auto
qed

lemma nrrstep_imp_subt_rstep:
 assumes "(s,t) \<in> nrrstep R"
 shows "\<exists>i. i < num_args s \<and> num_args s = num_args t \<and> (s|_<i>,t|_<i>) \<in> rstep R \<and> (\<forall>j. i \<noteq> j \<longrightarrow> s|_<j> = t|_<j>)"
proof -
 from `(s,t) \<in> nrrstep R` obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by blast
 with `(s,t) \<in> nrrstep R` have "(Fun f ss,t) \<in> nrrstep R" by simp
 then obtain i u where "t = Fun f (ss[i := u])" and "(ss!i,u) \<in> rstep R" and "i < length ss"
  using nrrstep_imp_pos_term by best
 from `s = Fun f ss` and `t = Fun f (ss[i := u])` have "num_args s = num_args t" by auto
 from `i < length ss` and `s = Fun f ss` have "i < num_args s" by auto
 from `s = Fun f ss` have "s|_<i> = (ss!i)" by auto
 from `t = Fun f (ss[i := u])` and `i < length ss` have "t|_<i> = u" by auto
 from `s = Fun f ss` and `t = Fun f (ss[i := u])`
 have  "\<forall>j. j \<noteq> i \<longrightarrow> s|_<j> = t|_<j>" by auto
 with `(ss!i,u) \<in> rstep R`
  and `i < num_args s`
  and `num_args s = num_args t`
  and `s|_<i> = (ss!i)`[symmetric] and `t|_<i> = u`[symmetric]
 show ?thesis by auto
qed

lemma finite_range:
  fixes f::"'a iseq"
  assumes "finite (range f)"
  shows "\<exists>x. \<exists>\<^sub>\<infinity>i. f i = x"
proof (rule ccontr)
  assume "\<not>(\<exists>x. \<exists>\<^sub>\<infinity>i. f i = x)"
  hence "\<forall>x. \<exists>j. \<forall>i>j. f i \<noteq> x" unfolding INFM_nat by blast
  from choice[OF this] obtain j where neq: "\<forall>x. \<forall>i>j x. f i \<noteq> x" ..
  from finite_range_imageI[OF assms] have "finite (range (j \<circ> f))" by (simp add: comp_def)
  from finite_nat_bounded[OF this] obtain m where "range (j \<circ> f) \<subseteq> {..<m}" ..
  hence "j (f m) < m" by (auto simp: comp_def)
  with neq show False by blast
qed

lemma nrrstep_subt: assumes "(s, t) \<in> nrrstep R" shows "\<exists>u\<lhd>s. \<exists>v\<lhd>t. (u, v) \<in> rstep R"
proof -
  from assms obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding nrrstep_def' by best
  from `C \<noteq> \<box>` s have "s \<rhd> l\<cdot>\<sigma>" by auto
  moreover from `C \<noteq> \<box>` t have "t \<rhd> r\<cdot>\<sigma>" by auto
  moreover from `(l, r) \<in> R` have "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rstep R" by auto
  ultimately show ?thesis by auto
qed

lemma nrrstep_args:
  assumes "(s, t) \<in> nrrstep R"
  shows "\<exists>f ss ts. s = Fun f ss \<and> t = Fun f ts \<and> length ss = length ts
    \<and> (\<exists>j<length ss. (ss!j, ts!j) \<in> rstep R \<and> (\<forall>i<length ss. i \<noteq> j \<longrightarrow> ss!i = ts!i))"
proof -
  from assms obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding nrrstep_def' by best
  from `C \<noteq> \<box>` obtain f ss1 D ss2 where C: "C = More f ss1 D ss2" by (induct C) auto
  have "s = Fun f (ss1 @ D\<langle>l\<cdot>\<sigma>\<rangle> # ss2)" (is "_ = Fun f ?ss") by (simp add: s C)
  moreover have "t = Fun f (ss1 @ D\<langle>r\<cdot>\<sigma>\<rangle> # ss2)" (is "_ = Fun f ?ts") by (simp add: t C)
  moreover have "length ?ss = length ?ts" by simp
  moreover have "\<exists>j<length ?ss.
    (?ss!j, ?ts!j) \<in> rstep R \<and> (\<forall>i<length ?ss. i \<noteq> j \<longrightarrow> ?ss!i = ?ts!i)"
  proof -
    let ?j = "length ss1"
    have "?j < length ?ss" by simp
    moreover have "(?ss!?j, ?ts!?j) \<in> rstep R"
    proof -
      from `(l, r) \<in> R` have "(D\<langle>l\<cdot>\<sigma>\<rangle>, D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
      thus ?thesis by auto
    qed
    moreover have "\<forall>i<length ?ss. i \<noteq> ?j \<longrightarrow> ?ss!i = ?ts!i" (is "\<forall>i<length ?ss. _ \<longrightarrow> ?P i")
    proof (intro allI impI)
      fix i assume "i < length ?ss" and "i \<noteq> ?j"
      hence "i < length ss1 \<or> i > length ss1" by auto
      thus "?P i"
      proof
        assume "i < length ss1" thus "?P i" by (auto simp: nth_append)
      next
        assume "i > length ss1" thus "?P i"
          using `i < length ?ss` by (auto simp: nth_Cons' nth_append)
      qed
    qed
    ultimately show ?thesis by best
  qed
  ultimately show ?thesis by auto
qed


lemma subterms_NF_imp_SN_elt_nrrstep:
  assumes "\<forall>s\<lhd>t. s \<in> NF (rstep R)" shows "SN_elt (nrrstep R) t"
proof
  fix S assume "S 0 = t" and "\<forall>i. (S i, S (Suc i)) \<in> nrrstep R"
  hence "(t, S (Suc 0)) \<in> nrrstep R" by auto
  then obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>" and t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>" unfolding nrrstep_def' by auto
  then obtain f ss1 D ss2 where C: "C = More f ss1 D ss2" by (induct C) auto
  have "t \<rhd> D\<langle>l\<cdot>\<sigma>\<rangle>" unfolding t C by auto
  moreover have "D\<langle>l\<cdot>\<sigma>\<rangle> \<notin> NF (rstep R)"
  proof -
    from `(l, r) \<in> R` have "(D\<langle>l\<cdot>\<sigma>\<rangle>, D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
    thus ?thesis by auto
  qed
  ultimately show False using assms by simp
qed

lemma args_NF_imp_SN_elt_nrrstep:
  assumes "\<forall>t\<in>set ts. t \<in> NF (rstep R)" shows "SN_elt (nrrstep R) (Fun f ts)"
proof
  fix S assume "S 0 = Fun f ts" and "\<forall>i. (S i, S (Suc i)) \<in> nrrstep R"
  hence "(Fun f ts, S (Suc 0)) \<in> nrrstep R" unfolding `S 0 = Fun f ts`[symmetric] by simp
  then obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>" and "Fun f ts = C\<langle>l\<cdot>\<sigma>\<rangle>"
    unfolding nrrstep_def' by auto
  then obtain ss1 D ss2 where C: "C = More f ss1 D ss2" by (induct C) auto
  with `Fun f ts = C\<langle>l\<cdot>\<sigma>\<rangle>` have "D\<langle>l\<cdot>\<sigma>\<rangle> \<in> set ts" by auto
  moreover have "D\<langle>l\<cdot>\<sigma>\<rangle> \<notin> NF (rstep R)"
  proof -
    from `(l, r) \<in> R` have "(D\<langle>l\<cdot>\<sigma>\<rangle>, D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
    thus ?thesis by auto
  qed
  ultimately show False using assms by simp
qed

fun
  idx :: "(nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> nat"
where
  "idx f 0 = 0" |
  "idx f (Suc i) = Suc (f (idx f i))"


lemma Tinf_imp_SN_elt_nrrstep:
  assumes "t \<in> Tinf R" shows "SN_elt (nrrstep R) t"
proof
  fix S assume "S 0 = t" and nrseq: "\<forall>i. (S i, S (Suc i)) \<in> nrrstep R"
  hence "\<forall>i. \<exists>f ss ts. S i = Fun f ss \<and> S (Suc i) = Fun f ts \<and> length ss = length ts
    \<and> (\<exists>j<length ss. (ss!j, ts!j) \<in> rstep R \<and> (\<forall>k<length ss. k \<noteq> j \<longrightarrow> ss!k = ts!k))"
    (is "\<forall>i. \<exists>f ss ts. ?P i f ss ts")
    using nrrstep_args by fast
  from choice[OF this] obtain f where "\<forall>i. \<exists>ss ts. ?P i (f i) ss ts" by best
  from choice[OF this] obtain ss where "\<forall>i. \<exists>ts. ?P i (f i) (ss i) ts" by best
  from choice[OF this] obtain ts where P: "\<forall>i. ?P i (f i) (ss i) (ts i)" by best
  moreover have "\<forall>i. f i = f (Suc i) \<and> ts i = ss (Suc i)"
  proof
    fix i
    from P have "S (Suc i) = Fun (f (Suc i)) (ss (Suc i))" by simp
    moreover from P have "S (Suc i) = Fun (f i) (ts i)" by blast+
    ultimately show "f i = f (Suc i) \<and> ts i = ss (Suc i)" by simp
  qed
  ultimately have "\<forall>i. ?P i (f i) (ss i) (ss (Suc i))" by auto
  hence "\<forall>i. length (ss i) = length (ss (Suc i))"
    and "\<forall>i. \<exists>j<length (ss i). (ss i ! j, ss (Suc i) ! j) \<in> rstep R
    \<and> (\<forall>k<length (ss i). k \<noteq> j \<longrightarrow> ss i ! k = ss (Suc i) ! k)" by blast+
  from choice[OF this(2)] obtain \<pi> where pi: "\<forall>i. \<pi> i < length (ss i)
    \<and> (ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R
    \<and> (\<forall>k<length (ss i). k \<noteq> \<pi> i \<longrightarrow> ss i ! k = ss (Suc i) ! k)" by blast+
  hence seq: "\<forall>i. (ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R" by simp
  have len: "\<forall>i. length (ss i) = length (ss 0)"
  proof
    fix i show "length (ss i) = length (ss 0)"
    proof (induct i)
      case 0 show ?case ..
    next
      case (Suc i) thus ?case using `\<forall>i. length (ss i) = length (ss (Suc i))` by simp
    qed
  qed
  have "\<forall>i. \<pi> i < length (ss 0)"
  proof
    fix i
    have "length (ss i) = length (ss 0)" using len by simp
    moreover from pi have "\<pi> i < length (ss i)" by simp
    ultimately show "\<pi> i < length (ss 0)" by simp
  qed
  hence "range \<pi> \<subseteq> {i. i < length (ss 0)}" by auto
  moreover have "finite {i. i < length (ss 0)}" by simp
  ultimately have "finite (range \<pi>)" by (rule finite_subset)
  from finite_range[OF this] obtain j where j: "\<forall>i. \<exists>k\<ge>i. \<pi> k = j" by (auto simp: INFM_nat_le)
  from choice[OF this] obtain q where q: "\<forall>i. q i \<ge> i \<and> \<pi> (q i) = j" by auto
  have j_len: "\<forall>i. j < length (ss i)"
  proof
    fix i
    from j obtain k where "k \<ge> i" and "\<pi> k = j" by best
    moreover from pi have "\<pi> k < length (ss k)" by simp
    ultimately have "j < length (ss k)" by simp
    moreover from len have "length (ss k) = length (ss 0)" by simp
    moreover from len have "length (ss i) = length (ss 0)" by simp
    ultimately show "j < length (ss i)" by simp
  qed
  from seq have step: "\<forall>i. (ss (q i) ! \<pi> (q i), ss (Suc (q i)) ! \<pi> (q i)) \<in> rstep R" by simp
  have refl_seq: "\<forall>i. (ss i ! j, ss (Suc i) ! j) \<in> (rstep R)^="
  proof
    fix i show "(ss i ! j, ss (Suc i) ! j) \<in> (rstep R)^="
    proof (cases "\<pi> i = j")
      assume "\<pi> i = j"
      moreover from pi have "(ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R" by simp
      ultimately show ?thesis by simp
    next
      assume "\<pi> i \<noteq> j"
      moreover from `\<forall>i. j < length (ss i)` have "j < length (ss i)" ..
      moreover from pi have "j < length (ss i) \<and> j \<noteq> \<pi> i \<longrightarrow> ss i ! j = ss (Suc i) ! j" by simp
      ultimately show ?thesis by simp
    qed
  qed
  let ?s = "\<lambda>i. ss (idx q i) ! j"
  have "\<forall>i. (ss i ! j, ss (Suc (q i)) ! j) \<in> (rstep R)^+"
  proof
    fix i
    from q have "q i \<ge> i" by simp
    from iseq_imp_steps[OF refl_seq this] have "(ss i ! j, ss (q i) ! j) \<in> (rstep R)^*" by simp
    moreover from step have "(ss (q i) ! j, ss (Suc (q i)) ! j) \<in> rstep R" by (simp add: q)
    ultimately show "(ss i ! j, ss (Suc (q i)) ! j) \<in> (rstep R)^+" by simp
  qed
  hence "\<forall>i. (?s i, ?s (Suc i)) \<in> (rstep R)^+" by simp
  hence "\<not> SN_elt (rstep R) (?s 0)" using SN_elt_imp_SN_elt_trancl by best
  moreover have "?s 0 \<lhd> t"
  proof -
    from P have t: "t = Fun (f 0) (ss 0)" by (simp add: `S 0 = t`[symmetric])
    have "j < length (ss 0)" using j_len by simp
    hence "?s 0 \<in> set (ss 0)" by simp
    thus ?thesis unfolding t ..
  qed
  moreover from assms have "\<forall>s\<lhd>t. SN_elt (rstep R) s" unfolding Tinf_def by simp
  ultimately show False by simp
qed

hide_const idx

lemma rrstep_imp_rule_subst:
  assumes "(s,t) \<in> rrstep R"
  shows "\<exists>l r \<sigma>. (l,r) \<in> R \<and> (l\<cdot>\<sigma>) = s \<and> (r\<cdot>\<sigma>) = t"
proof -
  have "ctxt_of_pos_term \<epsilon> s = Hole" by auto
  obtain l r \<sigma> where "(s,t) \<in> rstep_r_p_s R (l,r) \<epsilon> \<sigma>" using assms unfolding rrstep_def by best
  hence "let C = ctxt_of_pos_term \<epsilon> s in \<epsilon> \<in> poss s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def by simp
  with `ctxt_of_pos_term \<epsilon> s = Hole` have "(l,r) \<in> R" and "l\<cdot>\<sigma> = s" and "r\<cdot>\<sigma> = t" unfolding Let_def by auto
  thus ?thesis by auto
qed

lemma nrrstep_preserves_root:
  assumes "(Fun f ss,t) \<in> nrrstep R" (is "(?s,t) \<in> nrrstep R") shows "\<exists>ts. t = (Fun f ts)"
using assms unfolding nrrstep_def rstep_r_p_s_def Let_def by auto

lemma nrrstep_equiv_root: assumes "(s,t) \<in> nrrstep R" shows "\<exists>f ss ts. s = Fun f ss \<and> t = Fun f ts"
proof -
 from assms obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by best
 with assms obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 from `s = Fun f ss` and `t = Fun f ts` show ?thesis by best
qed

lemma nrrstep_reflects_root:
 assumes "(s,Fun g ts) \<in> nrrstep R" (is "(s,?t) \<in> nrrstep R")
 shows "\<exists>ss. s = (Fun g ss)"
proof -
 from assms obtain f ss ts' where "s = Fun f ss" and "Fun g ts = Fun f ts'" using nrrstep_equiv_root by best
 hence "f = g" by simp
 with `s = Fun f ss` have "s = Fun g ss" by simp
 thus ?thesis by auto
qed

lemma nrrsteps_preserve_root:
 assumes "(Fun f ss,t) \<in> (nrrstep R)^*"
 shows "\<exists>ts. t = (Fun f ts)"
using assms by induct (auto simp: nrrstep_preserves_root)

lemma nrrstep_Fun_imp_arg_rsteps:
   assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R") and "i < length ss"
  shows "(ss!i,ts!i) \<in> (rstep R)^*"
proof -
 from assms have "<i> \<in> poss ?s" using ep_pos_term by simp
 from `(?s,?t) \<in> nrrstep R`
 obtain l r j ps \<sigma>
 where A: "let C = ctxt_of_pos_term (j<#ps) ?s in (j<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding nrrstep_def rstep_r_p_s_def by force
 let ?C = "ctxt_of_pos_term (j<#ps) ?s"
 from A have "(j<#ps) \<in> poss ?s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" using Let_def by auto
 have C: "?C = More f (take j ss) (ctxt_of_pos_term ps (ss!j)) (drop (Suc j) ss)" (is "?C = More f ?ss1 ?D ?ss2") by auto
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = (ss!j)" by (auto simp: take_drop_imp_nth)
 from `(l,r) \<in> R` have "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> (CS R)" by auto
 hence "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" by (auto simp only: ctxt_closure.ctxt_closureI)
 hence D_rstep: "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R"
   unfolding rstep_CC_CS by auto
 from `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` and C have "?t = Fun f (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 hence ts: "ts = (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 have "j = i \<or> j \<noteq> i" by simp
 from `j<#ps \<in> poss ?s` have "j < length ss" by simp
 hence "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append_take)
 with ts have "ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 have "j = i \<or> j \<noteq> i" by simp
 thus "(ss!i,ts!i) \<in> (rstep R)^*"
 proof
  assume "j = i"
  with `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` and `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` and D_rstep show ?thesis by auto
 next
  assume "j \<noteq> i"
  with `i < length ss` and `j < length ss` have "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! i = ss!i" by (auto simp: nth_append_take_drop_is_nth_conv)
  with ts have "ts!i = ss!i" by auto
  thus ?thesis by auto
 qed
qed

lemma nrrstep_imp_arg_rsteps:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args s" shows "(args s!i,args t!i) \<in> (rstep R)^*"
proof (cases s)
 fix x assume "s = Var x" thus ?thesis using assms by auto
next
 fix f ss assume "s = Fun f ss"
 hence "(Fun f ss,t) \<in> nrrstep R" using assms by simp
 then obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 with `(s,t) \<in> nrrstep R` and `s = Fun f ss` have "(Fun f ss,Fun f ts) \<in> nrrstep R" by simp
 from  `s = Fun f ss` and `i < num_args s` have "i < length ss" by simp
 with `(Fun f ss,Fun f ts) \<in> nrrstep R`
   have "(ss!i,ts!i) \<in> (rstep R)^*" by (rule nrrstep_Fun_imp_arg_rsteps)
 with `s = Fun f ss` and `t = Fun f ts` show ?thesis by simp
qed

lemma nrrsteps_imp_rsteps: "(s,t) \<in> (nrrstep R)^* \<Longrightarrow> (s,t) \<in> (rstep R)^*"
proof (induct rule: rtrancl.induct)
 case (rtrancl_refl a) thus ?case by simp
next
 case (rtrancl_into_rtrancl a b c)
 hence IH: "(a,b) \<in> (rstep R)^*" and nrrstep: "(b,c) \<in> nrrstep R" by auto
 from nrrstep have "(b,c) \<in> rstep R" using nrrstep_imp_rstep by auto
 with IH show ?case by auto
qed

lemma nrrstep_Fun_preserves_num_args:
  assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R")
  shows "length ss = length ts"
proof -
 from assms obtain l r i ps \<sigma>
   where "let C = ctxt_of_pos_term (i<#ps) ?s in (i<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" (is "let C = ?C in _")
 unfolding nrrstep_def rstep_r_p_s_def by force
 hence "(l,r) \<in> R" and Cl: "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and Cr: "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding Let_def by auto
 have C: "?C = More f (take i ss) (ctxt_of_pos_term ps (ss!i)) (drop (Suc i) ss)" (is "?C = More f ?ss1 ?D ?ss2") by simp
 from C and Cl have s: "?s = Fun f (take i ss @ ?D\<langle>l\<cdot>\<sigma>\<rangle> # drop (Suc i) ss)" (is "?s = Fun f ?ss") by simp
 from C and Cr have t: "?t = Fun f (take i ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc i) ss)" (is "?t = Fun f ?ts") by simp
 from s and t have ss: "ss = ?ss" and ts: "ts = ?ts" by auto
 have "length ?ss = length ?ts" by auto
 with ss and ts show ?thesis by simp
qed

lemma nrrstep_equiv_num_args:
 assumes "(s,t) \<in> nrrstep R" shows "num_args s = num_args t"
proof -
 from assms obtain f ss ts where s:"s = Fun f ss" and t:"t = Fun f ts" using nrrstep_equiv_root by best
 with assms have "(Fun f ss,Fun f ts) \<in> nrrstep R" by simp
 hence "length ss = length ts" by (rule nrrstep_Fun_preserves_num_args)
 with s and t show ?thesis by auto
qed

lemma nrrsteps_equiv_num_args:
 assumes "(s,t) \<in> (nrrstep R)^*" shows "num_args s = num_args t"
using assms by (induct, auto simp: nrrstep_equiv_num_args)
 
lemma nrrstep_preserves_num_args:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args s" shows "i < num_args t"
proof (cases s)
 fix x assume "s = Var x" thus ?thesis using assms by auto
next
 fix f ss assume "s = Fun f ss"
 with assms obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 from `(s,t) \<in> nrrstep R` have "length ss = length ts" unfolding `s = Fun f ss` and `t = Fun f ts` by (rule nrrstep_Fun_preserves_num_args)
 with assms and `s = Fun f ss` and `t = Fun f ts` show ?thesis by auto
qed

lemma nrrstep_reflects_num_args:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args t" shows "i < num_args s"
proof (cases t)
 fix x assume "t = Var x" thus ?thesis using assms by auto
next
 fix g ts assume "t = Fun g ts"
 with assms obtain ss where "s = Fun g ss" using nrrstep_reflects_root by best
 from `(s,t) \<in> nrrstep R` have "length ss = length ts" unfolding `s = Fun g ss` and `t = Fun g ts` by (rule nrrstep_Fun_preserves_num_args)
 with assms and `s = Fun g ss` and `t = Fun g ts` show ?thesis by auto
qed

lemma nrrsteps_imp_arg_rsteps: 
 assumes "(s,t) \<in> (nrrstep R)^*" and "i < num_args s"
 shows "(args s!i,args t!i) \<in> (rstep R)^*"
using assms
proof (induct rule: rtrancl.induct)
 case (rtrancl_refl a) thus ?case by auto
next
 case (rtrancl_into_rtrancl a b c)
 hence IH: "(args a!i,args b!i) \<in> (rstep R)^*" by auto
 from `(a,b) \<in> (nrrstep R)^*` and `i < num_args a` have "i < num_args b" by induct (auto simp: nrrstep_preserves_num_args)
 with `(b,c) \<in> nrrstep R`
   have "(args b!i,args c!i) \<in> (rstep R)^*" by (rule nrrstep_imp_arg_rsteps)
 with IH show ?case by simp
qed

definition
  wf_trs :: "('f, 'v) trs \<Rightarrow> bool"
where
  "wf_trs R = (\<forall>l r. (l,r) \<in> R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l)"

lemma wf_trs_imp_lhs_Fun:
  "wf_trs R \<Longrightarrow> (l,r) \<in> R \<Longrightarrow> \<exists>f ts. l = Fun f ts"
unfolding wf_trs_def by blast

lemma rstep_imp_Fun:
 assumes "wf_trs R"
 shows "(s, t) \<in> rstep R \<Longrightarrow> \<exists>f ss. s = Fun f ss"
proof -
  assume "(s, t) \<in> rstep R"
  then obtain l r p \<sigma> where "(s, t) \<in> rstep_r_p_s R (l, r) p \<sigma>"
    unfolding rstep_iff_rstep_r_p_s by best
  hence "(ctxt_of_pos_term p s)\<langle>l \<cdot> \<sigma>\<rangle> = s" (is "?C\<langle>l \<cdot> \<sigma>\<rangle> = s") and "(l, r) \<in> R"
    unfolding rstep_r_p_s_def Let_def by auto
  with `wf_trs R` obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
  from `?C\<langle>l \<cdot> \<sigma>\<rangle> = s` obtain C where "C\<langle>l \<cdot> \<sigma>\<rangle> = s" by best
  hence "C\<langle>Fun f (map (\<lambda>x. x \<cdot> \<sigma>) ss)\<rangle> = s" unfolding `l = Fun f ss` by simp
  thus "\<exists>f ss. s = Fun f ss" by (induct C) auto
qed


lemma SN_elt_imp_SN_elt_subt:
  assumes "SN_elt (rstep R) t" shows "\<forall>s\<unlhd>t. SN_elt (rstep R) s"
proof (rule ccontr)
  assume "\<not>(\<forall>s\<unlhd>t. SN_elt (rstep R) s)"
  then obtain s where "t \<unrhd> s" and "\<not> SN_elt (rstep R) s" by auto
  then obtain S where "S 0 = s" and seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R" by auto
  from `t \<unrhd> s` obtain C where t: "t = C\<langle>s\<rangle>" by auto
  let ?S = "\<lambda>i. C\<langle>S i\<rangle>"
  from `S 0 = s` have "?S 0 = t" by (simp add: t)
  moreover from seq have "\<forall>i. (?S i, ?S (Suc i)) \<in> rstep R" by blast
  ultimately have "\<not> SN_elt (rstep R) t" by best
  with assms show False by simp
qed

lemma not_SN_elt_subt_imp_not_SN_elt:
  assumes "\<not> SN_elt (rstep R) t" and "s \<unrhd> t"
  shows "\<not> SN_elt (rstep R) s"
using assms SN_elt_imp_SN_elt_subt by blast

lemma SN_elt_instance_imp_SN_elt_var: 
  assumes "SN_elt (rstep R) (t\<cdot>\<sigma>)" and "x \<in> vars_term t"
  shows "SN_elt (rstep R) (Var x\<cdot>\<sigma>)"
proof -
  from assms have "t \<unrhd> Var x" by auto
  hence "t\<cdot>\<sigma> \<unrhd> (Var x)\<cdot>\<sigma>" by (rule supteq_stable)
  with SN_elt_imp_SN_elt_subt and assms show ?thesis by best
qed

lemma var_imp_var_of_arg:
  assumes "x \<in> vars_term (Fun f ss)" (is "x \<in> vars_term ?s")
  shows "\<exists>i < num_args (Fun f ss). x \<in> vars_term (ss!i)"
proof -
  from assms have "x \<in> \<Union>set (map vars_term ss)" by simp
  hence "x \<in> (\<Union>i<length ss. vars_term(ss!i))" unfolding set_conv_nth by auto
  hence "\<exists>i<length ss. x \<in> vars_term(ss!i)" using UN_iff by best
  then obtain i where "i < length ss" and "x \<in> vars_term(ss!i)" by auto
  hence "i < num_args ?s" by simp
  with `x \<in> vars_term(ss!i)` show ?thesis by auto
qed

lemma subt_instance_and_not_subst_imp_subt:
  "s\<cdot>\<sigma> \<unrhd> t \<Longrightarrow> \<forall>x \<in> vars_term s. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t) \<Longrightarrow> \<exists>u. s \<unrhd> u \<and> t = u\<cdot>\<sigma>"
proof (induct s arbitrary: t rule: term_induct)
 case (Var x) thus ?case by auto
next
 case (Fun f ss)
 from `Fun f ss\<cdot>\<sigma> \<unrhd> t` have "(Fun f ss\<cdot>\<sigma> = t) \<or> (Fun f ss\<cdot>\<sigma> \<rhd> t)" by auto
 thus ?case
 proof
  assume "Fun f ss\<cdot>\<sigma> = t" with Fun show ?thesis by auto
 next
  assume "Fun f ss\<cdot>\<sigma> \<rhd> t"
  hence "Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ss) \<rhd> t" by simp
  hence "\<exists>s \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss). s \<unrhd> t" (is "\<exists>s \<in> set ?ss'. s \<unrhd> t") by (rule supt_Fun_imp_arg_supteq)
  then obtain s' where "s' \<in> set ?ss'" and "s' \<unrhd> t" by best
  hence "\<exists>i < length ?ss'. ?ss'!i = s'" using in_set_conv_nth[where ?x = "s'"] by best
  then obtain i where "i < length ?ss'" and "?ss'!i = s'" by best
  hence "?ss'!i = (ss!i)\<cdot>\<sigma>" by auto
  from `?ss'!i = s'` have "s' = (ss!i)\<cdot>\<sigma>" unfolding `?ss'!i = (ss!i)\<cdot>\<sigma>` by simp
  from `s' \<unrhd> t` have "(ss!i)\<cdot>\<sigma> \<unrhd> t" unfolding `s' = (ss!i)\<cdot>\<sigma>` by simp
  with `i < length ?ss'` have "(ss!i) \<in> set ss" by auto
  with `(ss!i)\<cdot>\<sigma> \<unrhd> t` have "\<exists>s \<in> set ss. s\<cdot>\<sigma> \<unrhd> t" by best
  then obtain s where "s \<in> set ss" and "s\<cdot>\<sigma> \<unrhd> t" by best
  with Fun have "\<forall>x \<in> vars_term s. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t)" by force
  from Fun 
  have IH: "s \<in> set ss \<longrightarrow> (\<forall>v. s\<cdot>\<sigma> \<unrhd> v \<longrightarrow> (\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> v) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> v = u\<cdot>\<sigma>))"
  by auto
  with `s \<in> set ss` have "!!v. s\<cdot>\<sigma> \<unrhd> v \<Longrightarrow> (\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> v) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> v = u\<cdot>\<sigma>)"
  by simp
  with `s\<cdot>\<sigma> \<unrhd> t` have "(\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> t) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> t = u\<cdot>\<sigma>)" by simp
  with `\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> t` obtain u where "s \<unrhd> u" and "t = u\<cdot>\<sigma>" by best
  with `s \<in> set ss` have "Fun f ss \<unrhd> u" by auto
  with `t = u\<cdot>\<sigma>` show ?thesis by best
 qed
qed

lemma SN_imp_SN_subt:
  assumes "SN_elt (rstep R) s" and "s \<unrhd> t"
  shows "SN_elt (rstep R) t"
proof
  fix S assume  "S 0 = t" and seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R"
  from `s \<unrhd> t` obtain C where "s = C\<langle>t\<rangle>" by auto
  let ?T = "\<lambda>i. C\<langle>S i\<rangle>"
  have "\<forall>i. (?T i, ?T (Suc i)) \<in> rstep R"
  proof
    fix i
    from seq have "(?T i, ?T (Suc i)) \<in> CC(rstep R)" using ctxt_closureI by auto
    thus "(?T i, ?T (Suc i)) \<in> rstep R" by simp
  qed
  with `S 0 = t` have "\<not> SN_elt (rstep R) s" unfolding `s = C\<langle>t\<rangle>` by auto
  with assms show "False" by simp
qed

lemma subterm_preserves_SN:
  assumes SN: "SN_elt (rstep R) t" and supt: "t \<rhd> s"
  shows "SN_elt (rstep R) s"
proof -
  from supt have "t \<unrhd> s" by auto
  thus ?thesis using SN_imp_SN_subt[OF SN, of s] by simp
qed

lemma SN_elt_rstep_imp_SN_elt_supt_union_rstep:
  assumes "SN_elt (rstep R) t"
  shows "SN_elt (supt \<union> rstep R) t" (is "SN_elt ?S t")
proof (rule ccontr)
  assume "\<not> SN_elt ?S t"
  then obtain s where ini: "s 0 = t" and seq: "\<forall>i. (s i, s (Suc i)) \<in> ?S"
    unfolding SN_elt_def by auto
  have SN: "\<forall>i. SN_elt (rstep R) (s i)"
  proof
    fix i show "SN_elt (rstep R) (s i)"
    proof (induct i)
      case 0 show ?case using assms unfolding `s 0 = t` .
    next
      case (Suc i)
      from seq have "(s i, s (Suc i)) \<in> ?S" by simp
      thus ?case
      proof
        assume "(s i, s (Suc i)) \<in> supt"
        from subterm_preserves_SN[OF Suc this] show ?thesis .
      next
        assume "(s i, s (Suc i)) \<in> rstep R"
        from step_preserves_SN_elt[OF this Suc] show ?thesis .
      qed
    qed
  qed
  have "\<not> (\<exists>S. \<forall>i. (S i, S (Suc i)) \<in> SN_rel_supt (rstep R))"
    using SN_SN_rel_supt_rstep unfolding SN_defs by auto
  moreover have "\<forall>i. (s i, s (Suc i)) \<in> SN_rel_supt (rstep R)"
  proof
    fix i
    from SN have SN: "SN_elt (rstep R) (s i)" by simp
    from seq have "(s i, s (Suc i)) \<in> supt \<union> rstep R" by simp
    thus "(s i, s (Suc i)) \<in> SN_rel_supt (rstep R)"
      unfolding SN_rel_supt_def SN_rel_def using SN by auto
  qed
  ultimately show False by auto
qed

lemma SN_supt_rstep_trancl:
  assumes a: "SN (rstep R)"
  shows "SN ((supt \<union> rstep R)^+)"
proof -
  have "SN (supt \<union> rstep R)"
    unfolding SN_def
    by (rule allI, rule SN_elt_rstep_imp_SN_elt_supt_union_rstep, rule a[unfolded SN_def, THEN spec])
  thus "SN ((supt \<union> rstep R)^+)" by (rule SN_imp_SN_trancl)
qed

lemma var_subt_imp_subt_in_vars: "r \<unrhd> Var x \<Longrightarrow> x \<in> vars_term r"
proof (induct r rule: term_induct)
 case (Var y) hence "x = y" by (cases rule: supteq.cases) auto
 thus ?case by simp
next
 case (Fun f ss)
 from `Fun f ss \<unrhd> Var x` have "(Fun f ss = Var x) \<or> (Fun f ss \<rhd> Var x)" by auto
 thus ?case
 proof
  assume "Fun f ss = Var x" thus ?thesis by auto
 next
  assume "Fun f ss \<rhd> Var x"
  then obtain s where "s \<in> set ss" and "s \<unrhd> Var x" using supt_Fun_imp_arg_supteq by best
  with Fun have "s \<unrhd> Var x \<Longrightarrow> x \<in> vars_term s" by best
  with `s \<unrhd> Var x` have "x \<in> vars_term s" by simp
  with `s \<in> set ss` show ?thesis by auto
 qed
qed

lemma SN_args_imp_SN_subt:
  assumes "\<And>i. i < num_args (Fun f ss) \<Longrightarrow> SN_elt (rstep R) (ss!i)"
  shows "\<forall>s. (Fun f ss) \<rhd> s \<longrightarrow> SN_elt (rstep R) s"
proof (intro allI impI)
  fix s assume "Fun f ss \<rhd> s" (is "?t \<rhd> s")
  hence "\<exists>u \<in> set ss. u \<unrhd> s" using supt_Fun_imp_arg_supteq by best
  then obtain u where "u \<in> set ss" and "u \<unrhd> s" by best
  hence "\<exists>i < length ss. ss!i = u" unfolding in_set_conv_nth by simp
  then obtain i where "i < length ss" and "ss!i = u" by best
  hence "i < num_args ?t" by simp
  with assms have "SN_elt (rstep R) (ss!i)" by simp
  with `u \<unrhd> s` show "SN_elt (rstep R) s" unfolding `ss!i = u`[symmetric]
    using SN_imp_SN_subt[where R = R and s = "ss!i"] by auto
qed

lemma SN_imp_SN_arg:
  assumes sn: "SN_elt (rstep R) (Fun f ts)" and arg: "t \<in> set ts" shows "SN_elt (rstep R) t"
proof -
  from arg have "Fun f ts \<unrhd> t" by auto
  with sn show ?thesis by (rule SN_imp_SN_subt)
qed

lemma SNinstance_imp_SN:
  assumes "SN_elt (rstep R) (t \<cdot> \<sigma>)" 
  shows "SN_elt (rstep R) t"
proof(rule ccontr)
  assume "\<not> SN_elt (rstep R) t"
  hence "\<not> SN_elt (rstep R) (t \<cdot> \<sigma>)" 
  proof (simp add: SN_elt_def)
    assume "\<exists>f. f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)"
    from this obtain f where prem: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" ..
    let ?g = "\<lambda> i. (f i) \<cdot> \<sigma>"
    from prem have "?g 0 = t \<cdot> \<sigma> \<and> (\<forall> i. (?g i, ?g (Suc i)) \<in> rstep R)" using subst_closed_rstep 
      unfolding subst_closed_def by auto 
    thus "\<exists> f. f 0 = t \<cdot> \<sigma> \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" by auto
  qed
  with assms show "False" by simp
qed

lemma rstep_imp_C_s_r:
  assumes "(s,t) \<in> rstep R"
  shows "\<exists>C \<sigma> l r. (l,r) \<in> R \<and> s = C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = C\<langle>r\<cdot>\<sigma>\<rangle>"
proof -
 from assms obtain l r p \<sigma> where step:"(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
  unfolding rstep_iff_rstep_r_p_s by best
 let ?C = "ctxt_of_pos_term p s"
 from step have "p \<in> poss s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = t"
  unfolding rstep_r_p_s_def Let_def by auto
 hence "(l,r) \<in> R \<and> s = ?C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = ?C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 thus ?thesis by force
qed

lemma rstepD[elim]:
  assumes "(s, t) \<in> rstep R"
  obtains l r C \<sigma>
    where "(l, r) \<in> R" and "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and "t = C\<langle>r\<cdot>\<sigma>\<rangle>"
using rstep_imp_C_s_r[OF assms] by auto

lemma SN_Var:
 assumes "wf_trs R" shows "SN_elt (rstep R) (Var x)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain S where [symmetric]: "S 0 = Var x"
    and seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R" by auto
  hence "(Var x, S (Suc 0)) \<in> rstep R" by simp
  then obtain C l r \<sigma> where "(l, r) \<in> R" and x: "Var x = C\<langle>l\<cdot>\<sigma>\<rangle>" by best
  hence "Var x = l\<cdot>\<sigma>" by (induct C) simp_all
  then obtain y where "l = Var y" by (induct l) simp_all
  with assms and `(l, r) \<in> R` show False unfolding wf_trs_def by blast
qed

lemma Tinf_lemma:
  assumes "wf_trs R" and "t \<in> Tinf R"
  shows "\<exists>l r u \<sigma>.
    (l, r) \<in> R \<and> 
    r \<unrhd> u \<and> 
    \<not> is_var u \<and> 
    (t, l\<cdot>\<sigma>) \<in> (nrrstep R)^* \<and> 
    (l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rrstep R \<and> 
    r\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma> \<and>
    l\<cdot>\<sigma> \<in> Tinf R \<and>
    u\<cdot>\<sigma> \<in> Tinf R \<and>
    \<not>(l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>)"
proof (cases t)
  case (Var x)
  from `t \<in> Tinf R` have "\<not> SN_elt (rstep R) (Var x)" unfolding Var Tinf_def by simp
  with SN_Var[OF `wf_trs R`] show ?thesis by simp
next
  case (Fun f ts)
  let ?t = "Fun f ts"
  from `t \<in> Tinf R` have "\<forall>s\<lhd>?t. SN_elt (rstep R) s"
    and "\<not> SN_elt (rstep R) ?t" by (auto simp: Tinf_def Fun)
  then obtain S where "S 0 = ?t" and seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R" by best
  have "\<exists>i. (S 0, S i) \<in> (nrrstep R)^* \<and> (S i, S (Suc i)) \<in> rrstep R"
  proof -
    from `t \<in> Tinf R` have "SN_elt (nrrstep R) (S 0)"
      by (simp add: Tinf_imp_SN_elt_nrrstep `S 0 = ?t` Fun)
    with seq union_iseq_SN_elt_imp_first_step[of S "rrstep R" "nrrstep R"]
      obtain j where first_step: "(S j, S (Suc j)) \<in> rrstep R"
      and nrsteps: "\<forall>i<j. (S i, S (Suc i)) \<in> nrrstep R"
      by (auto simp: rstep_iff_rrstep_or_nrrstep)
    from nrsteps have "(S 0, S j) \<in> (nrrstep R)^*" by (induct j) auto
    with first_step show ?thesis by auto
  qed
  then obtain j where nrseq: "(S 0, S j) \<in> (nrrstep R)^*"
    and rrstep: "(S j, S (Suc j)) \<in> rrstep R" by auto

  from rrstep_imp_rule_subst[OF rrstep]
    obtain l r \<sigma> where "(l, r) \<in> R" and "l\<cdot>\<sigma> = S j" and "r\<cdot>\<sigma> = S (Suc j)" by auto

  have "\<not> SN_elt (rstep R) (r\<cdot>\<sigma>)"
  proof -
    let ?T = "\<lambda>i. S (Suc j + i)"
    from `r\<cdot>\<sigma> = S (Suc j)` have "?T 0 = r\<cdot>\<sigma>" by simp
    moreover from `\<forall>i. (S i, S (Suc i)) \<in> rstep R` have "\<forall>i. (?T i, ?T (Suc i)) \<in> rstep R"
      by simp
    ultimately show ?thesis unfolding SN_defs by best
  qed
  then obtain t' where "r\<cdot>\<sigma> \<unrhd> t'" and "t' \<in> Tinf R" using not_SN_imp_subt_Tinf by auto

  from nrseq have "(?t, S j) \<in> (nrrstep R)^*" by (simp add: `S 0 = ?t`)
  moreover from nrrsteps_preserve_root[OF this] obtain ls'
    where "S j = Fun f ls'" by auto
  ultimately have "l\<cdot>\<sigma> = Fun f ls'" and "(?t, Fun f ls') \<in> (nrrstep R)^*"
    by (auto simp: `l\<cdot>\<sigma> = S j`)
  hence "(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*" by simp

  from wf_trs_imp_lhs_Fun[OF `wf_trs R` `(l, r) \<in> R`] obtain g ls
    where "l = Fun g ls" by best
  with `l\<cdot>\<sigma> = Fun f ls'` have "\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>" by auto

  have "\<forall>i<length ts. (ts!i, (ls!i)\<cdot>\<sigma>) \<in> (rstep R)^*"
  proof (intro allI impI)
    fix i assume "i < length ts"
    hence "i < num_args ?t" by simp
    with `(?t, S j) \<in> (nrrstep R)^*` have "(args ?t!i, args (Fun f ls')!i) \<in> (rstep R)^*"
      unfolding `S j = Fun f ls'` by (rule nrrsteps_imp_arg_rsteps)
    hence "(ts!i, ls'!i) \<in> (rstep R)^*" by simp
    from `(?t, Fun f ls') \<in> (nrrstep R)^*` and `i < num_args ?t`
      have "i < num_args (Fun f ls')" by induct (auto simp: nrrstep_preserves_num_args)
    hence "i < length ls'" by simp
    with `\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>` and `(ts!i, ls'!i) \<in> (rstep R)^*`
    show "(ts!i, (ls!i)\<cdot>\<sigma>) \<in> (rstep R)^*" by simp
  qed

  from `(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*` have "length ts = num_args (l\<cdot>\<sigma>)"
    by induct (auto simp: nrrstep_equiv_num_args)
  with `l\<cdot>\<sigma> = Fun f ls'` and `\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>`
    have "\<forall>i<length ts. (ls!i)\<cdot>\<sigma> = ls'!i" by simp
  moreover with `\<forall>s\<lhd>?t. SN_elt (rstep R) s` have "\<forall>i<length ts. SN_elt (rstep R) (ts!i)"
    unfolding supt_supteq_conv by auto
  ultimately have "\<forall>i<length ts. SN_elt (rstep R) ((ls!i)\<cdot>\<sigma>)"
    using `\<forall>i<length ts. (ts!i, (ls!i)\<cdot>\<sigma>) \<in> (rstep R)^*`
    using steps_preserve_SN_elt[of _ _ "rstep R"] by blast
  with `\<forall>i<length ts. (ls!i)\<cdot>\<sigma> = ls'!i` and `length ts = num_args (l\<cdot>\<sigma>)`
    and `l\<cdot>\<sigma> = Fun f ls'` and `\<forall>i<length ts. SN_elt (rstep R) ((ls!i)\<cdot>\<sigma>)`
    have "\<forall>i<num_args (Fun f ls'). SN_elt (rstep R) (ls'!i)" by auto
  hence "\<forall>s\<lhd>Fun f ls'. SN_elt (rstep R) s" using SN_args_imp_SN_subt[of f ls' R] by blast
  hence "\<forall>s\<lhd>l\<cdot>\<sigma>. SN_elt (rstep R) s" by (simp add: `l\<cdot>\<sigma> = Fun f ls'`)

  from `(l, r) \<in> R` have "(l, r) \<in> R" .
  
  moreover have "l\<cdot>\<sigma> \<in> Tinf R"
  proof -
    let ?T = "\<lambda>i. S (j + i)"
    have "?T 0 = l\<cdot>\<sigma>" by (simp add: `l\<cdot>\<sigma> = S j`)
    moreover from `\<forall>i. (S i, S (Suc i)) \<in> rstep R`
      have "\<forall>i. (?T i, ?T (Suc i)) \<in> rstep R" by simp
    ultimately have "\<not> SN_elt (rstep R) (l\<cdot>\<sigma>)" unfolding SN_defs by best
    with `\<forall>s\<lhd>l\<cdot>\<sigma>. SN_elt (rstep R) s` show ?thesis by (simp add: Tinf_def)
  qed

  moreover have "\<exists>u. r \<unrhd> u \<and> \<not> is_var u \<and> r\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma> \<and> u\<cdot>\<sigma> \<in> Tinf R \<and> \<not>(l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>)"
  proof -
    have "\<forall>x\<in>vars_term r. SN_elt (rstep R) (Var x \<cdot> \<sigma>)"
    proof
      fix x
      assume "x \<in> vars_term r"
      hence "x \<in> vars_term l" using `(l, r) \<in> R` and `wf_trs R` by (auto simp: wf_trs_def)
      then obtain j where "j < length ls" and "x \<in> vars_term (ls!j)"
        using var_imp_var_of_arg unfolding `l = Fun g ls` by force
      hence "j < num_args (l\<cdot>\<sigma>)" unfolding `l = Fun g ls` by simp
      with `(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*` have "j < num_args ?t"
        by (induct rule: converse_rtrancl_induct) (auto simp: nrrstep_reflects_num_args)
      with `\<forall>i<length ts. SN_elt (rstep R) ((ls!i)\<cdot>\<sigma>)`
        have "SN_elt (rstep R) ((ls!j)\<cdot>\<sigma>)" by simp
      with `x \<in> vars_term (ls!j)` show "SN_elt (rstep R) (Var x \<cdot> \<sigma>)"
        using SN_elt_instance_imp_SN_elt_var[of R "ls!j"] by simp
    qed

    have "\<forall>x\<in>vars_term r. \<not>(Var x\<cdot>\<sigma> \<unrhd> t')"
    proof
      fix x
      assume "x \<in> vars_term r"
      hence "SN_elt (rstep R) (Var x \<cdot> \<sigma>)"
        using `\<forall>x\<in>vars_term r. SN_elt (rstep R) (Var x \<cdot> \<sigma>)` by simp
      show "\<not> Var x \<cdot> \<sigma>  \<unrhd> t'"
      proof
        assume "Var x \<cdot> \<sigma> \<unrhd> t'"
        with `SN_elt (rstep R) (Var x \<cdot> \<sigma>)` have "SN_elt (rstep R) t'"
          using SN_imp_SN_subt by auto
        with `t' \<in> Tinf R` show False unfolding Tinf_def by auto
      qed
    qed
    from subt_instance_and_not_subst_imp_subt[OF `r\<cdot>\<sigma> \<unrhd> t'` this]
      obtain u where "r \<unrhd> u" and "t' = u\<cdot>\<sigma>" by best
    
    moreover have "\<not> is_var u"
    proof
      assume "is_var u"
      then obtain x where "u = Var x" by (cases u) auto
      with `r \<unrhd> u` have "r \<unrhd> Var x" by simp
      from var_subt_imp_subt_in_vars[OF this] have "x \<in> vars_term r" .
      with `\<forall>x\<in>vars_term r. SN_elt (rstep R) (Var x \<cdot> \<sigma>)` and `t' \<in> Tinf R`
      show False by (simp add: Tinf_def `u = Var x` `t' = u\<cdot>\<sigma>`)
    qed

    moreover from `r \<unrhd> u` have "r\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma>" by (rule supteq_stable)
    moreover from `t' \<in> Tinf R` have "u\<cdot>\<sigma> \<in> Tinf R" by (simp add: `t' = u\<cdot>\<sigma>`)
    moreover have "\<not>(l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>)"
    proof
      assume "l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>"
      with `l\<cdot>\<sigma> \<in> Tinf R` have "SN_elt (rstep R) (u\<cdot>\<sigma>)" unfolding Tinf_def by force
      with `u\<cdot>\<sigma> \<in> Tinf R` show False by (simp add: Tinf_def)
    qed

    ultimately show ?thesis by auto
  qed

  moreover from `(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*` have "(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*" .

  moreover from `(l, r) \<in> R` have "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rrstep R" unfolding rrstep_def' by best
  ultimately show ?thesis by (auto simp: Fun)
qed
 
lemma Tinf_imp_def_root:
  assumes "wf_trs R" and "t \<in> Tinf R"
  shows "\<exists>f ss. t = Fun f ss \<and> defined R (f, length ss)"
proof -
  from assms obtain l r \<sigma> where "(l, r) \<in> R"
    and "(t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*" using Tinf_lemma by blast
  with `wf_trs R` obtain f ls where l: "l = Fun f ls" unfolding wf_trs_def by best
  hence "l\<cdot>\<sigma> = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ls)" (is "_ = Fun f ?ls'") by simp
  with `(t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*` have "\<exists>ts. t = Fun f ts"
    by (induct rule: converse_rtrancl_induct) (auto simp: nrrstep_reflects_root)
  then obtain ts where t: "t = Fun f ts" by best
  have "root l = Some f" by (simp add: l)
  with `(l, r) \<in> R` nrrsteps_equiv_num_args[OF `(t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*`] t l
    have "defined R (f, length ts)" unfolding defined_def by auto
  thus ?thesis by (simp add: t)
qed

fun map_rule :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)rule \<Rightarrow> ('g,'v)rule"
where "map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr))"

fun map_trs :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('g,'v)trs"
where "map_trs fg R = (map_rule fg) ` R"

lemma wf_trs_map_trs[simp]: "wf_trs (map_trs f R) = wf_trs R"
unfolding wf_trs_def
proof (rule iffI, intro allI impI)
  fix l r
  assume "\<forall>l r. (l, r) \<in> map_trs f R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l" and "(l, r) \<in> R"
  thus "(\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l" by (cases l, force+)
next
  assume ass: " \<forall>l r. (l, r) \<in> R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l"
  show "\<forall>l r. (l, r) \<in> map_trs f R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l"
  proof (intro allI impI)
    fix l r
    assume "(l, r) \<in> map_trs f R"
    with ass 
    show "(\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l"
      by (cases l, force+)
  qed
qed


type_synonym
  ('f, 'v) rules = "('f,'v) rule list"

fun map_trs_list :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)rules \<Rightarrow> ('g,'v)rules"
  where "map_trs_list fg R = map (map_rule fg) R"

lemma map_trs_list[simp]: "set (map_trs_list fg R) = map_trs fg (set R)" by auto
declare map_trs_list.simps[simp del]

lemma map_trs_comp: "map_trs fg (map_trs gh R) = map_trs (fg o gh) R"
proof -  
  have mr: "map_rule (fg o gh) = map_rule fg o map_rule gh"
    by (rule ext, auto simp: map_term_comp)
  thus ?thesis
    by (auto simp: map_term_comp image_compose[symmetric] mr)
qed

lemma map_trs_mono: assumes "R \<subseteq> R'" shows "map_trs fg R \<subseteq> map_trs fg R'"
  using assms by auto

lemma map_trs_power_mono: 
  fixes R R' :: "('f,'v)trs" and fg :: "'f \<Rightarrow> 'f"
  assumes "R \<subseteq> R'" shows "((map_trs fg)^^n) R \<subseteq> ((map_trs fg)^^n) R'"
using assms by (induct n, simp, auto)

declare map_trs.simps[simp del]


lemma rstep_induct[case_names IH, induct set: "rstep"]:
  assumes "(s, t) \<in> rstep R"
    and IH: "\<And>C \<sigma> l r. (l, r) \<in> R \<Longrightarrow> P C\<langle>l \<cdot> \<sigma>\<rangle> C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "P s t"
proof -
  from `(s, t) \<in> rstep R` obtain C \<sigma> l r
    where "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  from IH[OF `(l, r) \<in> R`] show ?thesis unfolding s t .
qed
text {*
We use @{thm[source] rstep_induct} as the default induction rule for @{term rstep}.
*}

lemma rstep_imp_map_rstep:
  assumes "(s,t) \<in> rstep R" 
  shows "(map_term fg s, map_term fg t) \<in> rstep (map_trs fg R)"
using assms proof (induct)
  case (IH C \<sigma> l r)
  hence "(map_term fg l, map_term fg r) \<in> map_trs fg R" (is "(?l, ?r) \<in> ?R") 
    unfolding map_trs.simps by force
  hence "(?l, ?r) \<in> rstep ?R" ..
  hence "(?l \<cdot> map_subst fg \<sigma>, ?r \<cdot> map_subst fg \<sigma>) \<in> rstep ?R" ..
  thus ?case by auto
qed

lemma rsteps_imp_map_rsteps: assumes "(s,t) \<in> (rstep R)^*"
  shows "(map_term fg s, map_term fg t) \<in> (rstep (map_trs fg R))^*"
using assms
proof (induct, clarify)
  case (step y z)
  hence "(map_term fg y, map_term fg z) \<in> rstep (map_trs fg R)" using rstep_imp_map_rstep 
    by (auto simp: map_trs.simps)
  with step show ?case by auto
qed

lemma SN_map_imp_SN: 
  fixes R:: "('g,'v)trs" and t :: "('g,'v)term"
  assumes SN: "SN_elt (rstep (map_trs fg R)) (map_term fg t)" 
  shows "SN_elt (rstep R) t"
proof (rule ccontr)
  assume "\<not> SN_elt (rstep R) t"
  from this obtain f where cond: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" 
  unfolding SN_elt_def by auto
  obtain g where g: "g = (\<lambda> i. map_term fg (f i))" by auto
  with cond have cond2: "g 0 = map_term fg t \<and> (\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R))"
    using rstep_imp_map_rstep[where fg = fg] by blast
  from SN
    have "\<not> (\<exists> g. (g 0 = map_term fg t \<and> (\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R))))" 
    unfolding SN_elt_def by auto
  with cond2 show False  by auto
qed

lemma rstep_iff_map_rstep:
  assumes "inj fg"
  shows "(s, t) \<in> rstep R \<longleftrightarrow> (map_term fg s, map_term fg t) \<in> rstep (map_trs fg R)"
proof
  assume "(s,t) \<in> rstep R" 
  thus "(map_term fg s, map_term fg t) \<in> rstep(map_trs fg R)"
    by (rule rstep_imp_map_rstep)
next
  assume "(map_term fg s, map_term fg t) \<in> rstep (map_trs fg R)" 
  hence "(map_term fg s, map_term fg t) \<in> CC(CS(map_trs fg R))"
    by (simp add: rstep_CC_CS)
  then obtain C u v where "(u,v) \<in> CS(map_trs fg R)" and "C\<langle>u\<rangle> = map_term fg s"
    and "C\<langle>v\<rangle> = map_term fg t"
    by (cases rule: ctxt_closure.cases) force
  then obtain \<sigma> w x where "(w,x) \<in> map_trs fg R" and "w\<cdot>\<sigma> = u" and "x\<cdot>\<sigma> = v"
    by (cases rule: subst_closure.cases) force
  then obtain l r where "w = map_term fg l" and "x = map_term fg r"
    and "(l,r) \<in> R" by (auto simp: map_trs.simps)
  have ps: "C\<langle>map_term fg l \<cdot> \<sigma>\<rangle> = map_term fg s" and pt: "C\<langle>map_term fg r \<cdot> \<sigma>\<rangle> = map_term fg t"
    unfolding `w = map_term fg l`[symmetric] `x = map_term fg r`[symmetric] `w\<cdot>\<sigma> = u` `x\<cdot>\<sigma> = v`
  `C\<langle>u\<rangle> = map_term fg s` `C\<langle>v\<rangle> = map_term fg t` by auto
  let ?gf = "the_inv fg"
  let ?C = "map_ctxt ?gf C"
  let ?\<sigma> = "map_subst ?gf \<sigma>"
  have gffg: "?gf \<circ> fg = id" using the_inv_f_f[OF assms] by (intro ext, auto)
  from ps and pt have "s = map_term ?gf (C\<langle>map_term fg l \<cdot> \<sigma>\<rangle>)"
    and "t = map_term ?gf (C\<langle>map_term fg r \<cdot> \<sigma>\<rangle>)" by (auto simp: map_term_comp gffg)
  hence s: "s = ?C\<langle>map_term ?gf (map_term fg l \<cdot> \<sigma>)\<rangle>"
    and t: "t = ?C\<langle>map_term ?gf (map_term fg r \<cdot> \<sigma>)\<rangle>" using map_ctxt_distrib by auto
  from s have "s = ?C\<langle>l\<cdot>?\<sigma>\<rangle>" by (simp add: map_subst_distrib map_term_comp gffg)
  from t have "t = ?C\<langle>r\<cdot>?\<sigma>\<rangle>" by (simp add: map_subst_distrib map_term_comp gffg)
  from `(l, r) \<in> R` have "(l\<cdot>?\<sigma>, r\<cdot>?\<sigma>) \<in> CS R" by (rule subst_closureI)
  hence "(?C\<langle>l\<cdot>?\<sigma>\<rangle>,?C\<langle>r\<cdot>?\<sigma>\<rangle>) \<in> CC(CS R)" by (rule ctxt_closureI)
  thus "(s,t) \<in> rstep R" unfolding `s = ?C\<langle>l\<cdot>?\<sigma>\<rangle>` `t = ?C\<langle>r\<cdot>?\<sigma>\<rangle>` by (simp add: rstep_CC_CS)
qed


lemma instance_no_subt_imp_no_subt:
 assumes "\<not> s\<cdot>\<sigma> \<rhd> (t\<cdot>\<sigma>)"
 shows "\<not> s \<rhd> t"
proof (rule ccontr)
 assume "\<not> \<not> s \<rhd> t"
 have "s \<rhd> t \<Longrightarrow> s\<cdot>\<sigma> \<rhd> (t\<cdot>\<sigma>)" by (rule supt_stable)
 with `\<not> \<not> s \<rhd> t` have "s\<cdot>\<sigma> \<rhd> (t\<cdot>\<sigma>)" by simp
 with assms show "False" by simp
qed

lemma Tinf_imp_Fun: 
 assumes "wf_trs R" and "t \<in> Tinf R"
 shows "\<exists>f ts. t = Fun f ts"
proof (cases t)
 fix x assume "t = Var x"
 from `wf_trs R` have "SN_elt (rstep R) t" unfolding `t = Var x` using SN_Var by best
 with `t \<in> Tinf R` show ?thesis unfolding Tinf_def by auto
next
 fix f ts assume "t = Fun f ts" thus ?thesis by auto
qed

lemma cs_subseteq_cs_ext1:
  assumes "(s, t) \<in> CS R"
  shows "(s, t) \<in> CS (R \<union> S)"
proof -
 from assms obtain \<sigma> u v 
 where R:"(u,v) \<in> R" and s: "s = u\<cdot>\<sigma>" and t:"t = v\<cdot>\<sigma>" by (cases rule: subst_closure.cases) auto
 from R have "(u,v) \<in> (R \<union> S)" ..
 thus ?thesis unfolding s t by (rule subst_closureI)
qed

lemma cs_subseteq_cs_ext2:
 assumes "(s,t) \<in> CS S"
 shows "(s,t) \<in> CS(R \<union> S)"
proof -
 have comm:"(S \<union> R) = (R \<union> S)" by auto
 from assms have "(s,t) \<in> CS(S \<union> R)" by (rule cs_subseteq_cs_ext1)
 thus ?thesis unfolding comm .
qed

lemma cs_subseteq_cs_ext1': "CS R \<subseteq> CS (R \<union> S)"
by (intro subsetI2) (rule cs_subseteq_cs_ext1)

lemma cs_union_distr: "CS(R \<union> S) = (CS R) \<union> (CS S)"
proof
 show "CS (R \<union> S) \<subseteq> ((CS R) \<union> (CS S))"
 proof (intro subsetI2)
  fix s t assume "(s,t) \<in> CS(R \<union> S)"
  then obtain \<sigma> u v where RS:"(u,v) \<in> (R \<union> S)" and s:"s = u\<cdot>\<sigma>" and t:"t = v\<cdot>\<sigma>"
   by (cases rule: subst_closure.cases) auto
  from RS have "(u,v) \<in> R \<or> (u,v) \<in> S" by simp
  hence "(s,t) \<in> CS R \<or> (s,t) \<in> CS S"
  proof
   assume "(u,v) \<in> R" thus ?thesis unfolding s t using subst_closureI by auto
  next
   assume "(u,v) \<in> S" thus ?thesis unfolding s t using subst_closureI by auto
  qed
  thus "(s,t) \<in> CS R \<union> CS S" by simp
 qed
next
 show "(CS R) \<union> (CS S) \<subseteq> CS (R \<union> S)"
 proof (intro subsetI2)
  fix s t assume "(s,t) \<in> CS R \<union> CS S"
  hence "(s,t) \<in> CS R \<or> (s,t) \<in> CS S" by simp
  thus "(s,t) \<in> CS(R \<union> S)"
  proof (rule disjE)
   assume "(s,t) \<in> CS R" thus ?thesis by (rule cs_subseteq_cs_ext1)
  next
   assume "(s,t) \<in> CS S" thus ?thesis by (rule cs_subseteq_cs_ext2)
  qed
 qed
qed

lemma cc_subseteq_cc_ext1:
 assumes "(s,t) \<in> CC R"
 shows "(s,t) \<in> CC(R \<union> S)"
proof -
 from assms obtain C u v
  where R:"(u,v) \<in> R" and s:"s = C\<langle>u\<rangle>" and t:"t = C\<langle>v\<rangle>" by (cases rule: ctxt_closure.cases) auto
 from R have "(u,v) \<in> (R \<union> S)" ..
 thus ?thesis unfolding s t by (rule ctxt_closureI)
qed

lemma cc_subseteq_cc_ext2:
 assumes "(s,t) \<in> CC S"
 shows "(s,t) \<in> CC(R \<union> S)"
proof -
 have comm:"(S \<union> R) = (R \<union> S)" by auto
 from assms have "(s,t) \<in> CC(S \<union> R)" by (rule cc_subseteq_cc_ext1)
 thus ?thesis unfolding comm .
qed

lemma cc_union_distr: "CC(R \<union> S) = (CC R) \<union> (CC S)"
proof
 show "CC (R \<union> S) \<subseteq> (CC R \<union> CC S)"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CC(R \<union> S)"
  then obtain C u v where RS:"(u,v) \<in> (R \<union> S)" and s:"s = C\<langle>u\<rangle>" and t:"t = C\<langle>v\<rangle>"
   by (cases rule: ctxt_closure.cases) auto
  from RS have "(u,v) \<in> R \<or> (u,v) \<in> S" by simp
  thus "(s,t) \<in> CC R \<union> CC S"
  proof
   assume "(u,v) \<in> R" thus ?thesis unfolding s t using ctxt_closureI by auto
  next
   assume "(u,v) \<in> S" thus ?thesis unfolding s t using ctxt_closureI by auto
  qed
 qed
next
 show "(CC R) \<union> (CC S) \<subseteq> CC (R \<union> S)"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CC R \<union> CC S"
  hence "(s,t) \<in> CC R \<or> (s,t) \<in> CC S" by simp
  thus "(s,t) \<in> CC(R \<union> S)"
  proof (rule disjE)
   assume "(s,t) \<in> CC R" thus ?thesis by (rule cc_subseteq_cc_ext1)
  next
   assume "(s,t) \<in> CC S" thus ?thesis by (rule cc_subseteq_cc_ext2)
  qed
 qed
qed

lemma CC_monotone: assumes "r \<subseteq> s" shows "CC r \<subseteq> CC s"
using assms
proof auto
  fix u v
  assume "(u,v) \<in> CC r"
  thus "(u,v) \<in> CC s"
  proof
    fix uu vv C
    assume "u = C\<langle>uu\<rangle>" and "v = C\<langle>vv\<rangle>" and "(uu,vv) \<in> r"
    with assms show ?thesis by (auto simp: ctxt_closureI)
  qed
qed

lemma CS_monotone: assumes "r \<subseteq> s" shows "CS r \<subseteq> CS s"
using assms
proof auto
  fix u v
  assume "(u,v) \<in> CS r"
  thus "(u,v) \<in> CS s"
  proof
    fix uu vv \<sigma>
    assume "u = uu \<cdot> \<sigma>" and "v = vv \<cdot> \<sigma>" and "(uu,vv) \<in> r"
    with assms show ?thesis by (auto simp: subst_closureI)
  qed
qed

lemma rstep_mono:
 assumes "rR \<subseteq> R"
 shows "rstep rR \<subseteq> rstep R"
proof-
 from assms have "CS rR \<subseteq> CS R" using CS_monotone by auto
 hence "CC (CS rR) \<subseteq> CC (CS R)" using CC_monotone by auto
 thus ?thesis by (simp add: rstep_CC_CS)
qed

lemma rstep_map_trs_power_mono: 
  fixes R R' :: "('f,'v)trs" and fg :: "'f \<Rightarrow> 'f"
  assumes subset: "R \<subseteq> R'" shows "rstep (((map_trs fg)^^n) R) \<subseteq> rstep (((map_trs fg)^^n) R')"
  by (rule rstep_mono, rule map_trs_power_mono, rule subset)

lemma cs_diff_distr[simp]: "CS A - CS B  \<subseteq> CS (A - B)" (* SW *)
proof-
 have "A = (A - B) \<union> (B \<inter> A)" by auto
 hence "CS A = CS ((A - B) \<union> (B \<inter> A))" by auto
 hence "CS A = CS (A - B) \<union> CS (B \<inter> A)" using cs_union_distr by auto
 hence "CS A - CS (B \<inter> A) = CS (A - B) \<union> CS (B \<inter> A) - CS (B \<inter> A)" by auto
 hence one:"CS A - CS (B \<inter> A) \<subseteq> CS (A - B)" by auto
 have "B \<inter> A \<subseteq> B" by auto
 from CS_monotone[OF this] have "CS (B \<inter> A) \<subseteq>  CS B" .
 hence "CS A - CS B \<subseteq> CS A - CS (B \<inter> A)" by auto
 with one show "CS A - CS B \<subseteq> CS (A - B)" by auto
qed

lemma cc_diff_distr[simp]: "CC A - CC B \<subseteq> CC (A - B)"  (* SW *)
proof-
 have "A = (A - B) \<union> (B \<inter> A)" by auto
 hence "CC A = CC ((A - B) \<union> (B \<inter> A))" by auto
 hence "CC A = CC (A - B) \<union> CC (B \<inter> A)" using cc_union_distr by auto
 hence one:"CC A - CC (B \<inter> A) = CC (A - B) \<union> CC (B \<inter> A) - CC (B \<inter> A)" by auto
 hence one:"CC A - CC (B \<inter> A) \<subseteq> CC (A - B)" by auto
 have "B \<inter> A \<subseteq> B" by auto
 from CC_monotone[OF this] have "CC (B \<inter> A) \<subseteq>  CC B" .
 hence "CC A - CC B \<subseteq> CC A - CC (B \<inter> A)" by auto
 with one show "CC A - CC B \<subseteq> CC (A - B)" by auto
qed

lemma rstep_diff:  (* SW *)
 shows "rstep R - rstep R' \<subseteq> rstep (R - R')"
proof-
 have "CC (CS R) - CC (CS R') \<subseteq> CC (CS R - CS R')" using cc_diff_distr by auto
 also have "\<dots> \<subseteq> CC (CS (R - R'))" using cs_diff_distr CC_monotone[where ?r="CS R - CS R'"] by auto
 finally show ?thesis unfolding rstep_CC_CS by auto
qed

   
lemma someI_ex2: "\<exists>x y. P (x,y) \<Longrightarrow> P (SOME (x,y). P (x,y))" using some_eq_ex by auto

lemma subsetI3:"(\<And>x y z. (x,y,z) \<in> A \<Longrightarrow> (x,y,z) \<in> B) \<Longrightarrow> A \<subseteq> B" by auto

lemma aux: "(\<Union>a\<in>P. {(x,y,z). x = fst a \<and> y = snd a \<and> Q a z}) = {(x,y,z). (x,y) \<in> P \<and> Q (x,y) z}" (is "?P = ?Q")
proof
 show "?P \<subseteq> ?Q"
 proof
  fix x assume "x \<in> ?P"
  then obtain a where "a \<in> P" and "x \<in> {(x,y,z). x = fst a \<and> y = snd a \<and> Q a z}" by auto
  then obtain b where "Q (fst x, fst (snd x)) (snd (snd x))" and "(fst x, fst (snd x)) = a" and "snd (snd x) = b" by force
  from `a \<in> P` have "(fst a,snd a) \<in> P" unfolding split_def by simp
  from `Q (fst x, fst (snd x)) (snd (snd x))` have "Q a b" unfolding `(fst x, fst (snd x)) = a` `snd (snd x) = b` .
  from `(fst a,snd a) \<in> P` and `Q a b` show "x \<in> ?Q" unfolding split_def `(fst x, fst (snd x)) = a`[symmetric] `snd (snd x) = b`[symmetric] by simp
 qed
next
 show "?Q \<subseteq> ?P"
 proof (rule subsetI3)
  fix x y z assume "(x,y,z) \<in> ?Q"
  hence "(x,y) \<in> P" and "Q (x,y) z" by auto
  hence "x = fst(x,y) \<and> y = snd(x,y) \<and> Q (x,y) z" by auto
  hence "(x,y,z) \<in> {(x,y,z). x = fst(x,y) \<and> y = snd(x,y) \<and> Q (x,y) z}" by auto
  hence "\<exists>p\<in>P. (x,y,z) \<in> {(x,y,z). x = fst p \<and> y = snd p \<and> Q p z}" using `(x,y) \<in> P` by blast
  thus "(x,y,z) \<in> ?P" unfolding UN_iff[symmetric] by simp
 qed
qed
 
lemma finiteR_imp_finiteDP': 
 assumes "finite R"
 shows "finite{(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R (h, length us) \<and> \<not>(l \<rhd> u))}"
proof -
 have "\<And>l r. (l,r) \<in> R \<Longrightarrow> finite{u. r \<unrhd> u}"
 proof -
  fix l r
  assume "(l,r) \<in> R"
  show "finite{u. r \<unrhd> u}" by (rule finite_subterms)
 qed
 with `finite R` have "finite(\<Union>(l,r) \<in> R. {u. r \<unrhd> u})" by auto
 have "finite(\<Union>lr\<in>R. {(l,r,u). l = fst lr \<and> r = snd lr \<and> snd lr \<unrhd> u})"
 proof (rule finite_UN_I)
  show "finite R" by (rule `finite R`)
 next
  fix lr
  assume "lr \<in> R"
  have "finite {u. snd lr \<unrhd> u}" by (rule finite_subterms)
  thus "finite {(l,r,u). l = fst lr \<and> r = snd lr \<and> snd lr \<unrhd> u}" by auto
 qed
 hence "finite {(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u}" unfolding aux by auto
 have "{(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u} \<supseteq> {(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R (h, length us) \<and> \<not>(l \<rhd> u))}" by auto
 with `finite {(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u}` show ?thesis using finite_subset by fast
qed

lemma card_image_le':
 assumes "finite S"
 shows "card (\<Union>y\<in>S.{x. x = f y}) \<le> card S"
proof -
 have A:"(\<Union>y\<in>S. {x. x = f y}) = f ` S" by auto
 from assms show ?thesis unfolding A using card_image_le by auto
qed

lemma subteq_of_map_imp_map: "map_term g s \<unrhd> t \<Longrightarrow> \<exists>u. t = map_term g u"
proof (induct s arbitrary: t)
 case (Var x)
 hence "map_term g (Var x) \<rhd> t \<or> map_term g (Var x) = t" by auto
 thus ?case
 proof
  assume "map_term g (Var x) \<rhd> t" thus ?thesis by (cases rule: supt.cases) auto
 next
  assume "map_term g (Var x) = t" thus ?thesis by best
 qed
next
 case (Fun f ss)
 hence "map_term g (Fun f ss) \<rhd> t \<or> map_term g (Fun f ss) = t" by auto
 thus ?case
 proof
  assume "map_term g (Fun f ss) \<rhd> t"
  thus ?case using Fun by (cases rule: supt.cases) (auto simp: supt_supteq_conv)
 next
  assume "map_term g (Fun f ss) = t" thus ?thesis by best
 qed
qed

lemma map_term_inj:
  assumes "inj fg"
  shows "inj (map_term fg)"
proof -
  {
    fix s t
    assume "map_term fg s = map_term fg t"
    hence "s = t"
    proof (induct s arbitrary: t)
      case (Var x) with assms show ?case by (induct t) auto
    next
      case (Fun f ss) thus ?case
      proof (induct t)
        case (Var y) thus ?case by auto
      next
        case (Fun g ts)
        hence A: "map (map_term fg) ss = map (map_term fg) ts" by simp
        hence len_eq:"length ss = length ts" by (rule map_eq_imp_length_eq)
        from A have "!!i. i < length ss \<Longrightarrow> (map (map_term fg) ss)!i = (map (map_term fg) ts)!i" by auto
        with len_eq have eq: "!!i. i < length ss \<Longrightarrow> map_term fg  (ss!i) = map_term fg (ts!i)" using nth_map by auto
        have in_set: "!!i. i < length ss \<Longrightarrow> (ss!i) \<in> set ss" by auto
        from Fun have "\<forall>i < length ss. (ss!i) = (ts!i)" using in_set eq by auto
        with len_eq have "ss = ts" using nth_equalityI[where xs = ss and ys = ts] by simp
        have "f = g" using Fun `inj fg` unfolding inj_on_def by auto
        with `ss = ts` show ?case by simp
      qed
    qed
  }
  thus ?thesis unfolding inj_on_def by auto
qed 

lemma rsteps_closed_ctxt: assumes "(s,t) \<in> (rstep R)^*" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> (rstep R)^*"
proof -
 from assms obtain n where "(s,t) \<in> (rstep R)^^n" using rtrancl_is_UN_rel_pow by auto
 thus ?thesis
 proof (induct n arbitrary: s)
  case 0 thus ?case by auto
 next
  case (Suc n)
  from rel_pow_Suc_D2[OF `(s,t) \<in> (rstep R)^^Suc n`] obtain u
    where "(s,u) \<in> (rstep R)" and "(u,t) \<in> (rstep R)^^n" by auto
  from `(s,u) \<in> (rstep R)` have "(C\<langle>s\<rangle>,C\<langle>u\<rangle>) \<in> (rstep R)" ..
  from Suc and `(u,t) \<in> (rstep R)^^n` have "(C\<langle>u\<rangle>,C\<langle>t\<rangle>) \<in> (rstep R)^*" by simp
  with `(C\<langle>s\<rangle>,C\<langle>u\<rangle>) \<in> (rstep R)` show ?case by auto
 qed
qed


lemma not_Nil_imp_last: "xs \<noteq> [] \<Longrightarrow> \<exists>ys y. xs = ys@[y]"
proof (induct xs)
 case Nil thus ?case by simp
next
 case (Cons x xs) show ?case
 proof (cases xs)
  assume "xs = []" with Cons show ?thesis by simp
 next
  fix x' xs' assume "xs = x'#xs'"
  hence "xs \<noteq> []" by simp
  with Cons obtain ys y where "xs = ys@[y]" by auto
  hence "x#xs = x#(ys@[y])" by simp
  hence "x#xs = (x#ys)@[y]" by simp
  thus ?thesis by auto
 qed
qed

lemma Nil_or_last: "xs = [] \<or> (\<exists>ys y. xs = ys@[y])"
proof (cases xs)
 assume "xs = []" thus ?thesis by simp
next
 fix x' xs' assume "xs = x'#xs'"
 hence "xs \<noteq> []" by simp
 thus ?thesis using not_Nil_imp_last by auto
qed

lemma args_rsteps_imp_rsteps: assumes "length ss = length ts" and args:"\<forall>i<length ss. (ss!i,ts!i) \<in> (rstep R)^*"
shows "(Fun f ss,Fun f ts) \<in> (rstep R)^*"
proof -
 let ?C = "\<lambda>i. More f (take i ts) \<box> (drop (Suc i) ss)"
 have a:"\<forall>i<length ss. (Fun f ss,(?C i)\<langle>ts!i\<rangle>) \<in> (rstep R)^*"
 proof (intro allI impI)
  fix i assume "i < length ss" thus "(Fun f ss,(?C i)\<langle>ts!i\<rangle>) \<in> (rstep R)^*"
  proof (induct i)
   case 0
   with args have "(ss!0,ts!0) \<in> (rstep R)^*" by simp
   hence "((?C 0)\<langle>ss!0\<rangle>,(?C 0)\<langle>ts!0\<rangle>) \<in> (rstep R)^*" by (rule rsteps_closed_ctxt)
   hence "((More f [] \<box> (drop (Suc 0) ss))\<langle>ss!0\<rangle>,(?C 0)\<langle>ts!0\<rangle>) \<in> (rstep R)^*" by simp
   hence "(Fun f ((ss!0)#drop (Suc 0) ss),(?C 0)\<langle>ts!0\<rangle>) \<in> (rstep R)^*" by simp
   thus "(Fun f ss,(?C 0)\<langle>ts!0\<rangle>) \<in> (rstep R)^*" using `0 < length ss`
     using nth_drop_0[where ss = ss] by simp
  next
   case (Suc i)
   hence IH: "(Fun f ss,(?C i)\<langle>ts!i\<rangle>) \<in> (rstep R)^*" by auto
   from Suc have "i < length ss" by simp
   with `length ss = length ts` have "i < length ts" by simp
   have "(?C(Suc i))\<langle>ss!(Suc i)\<rangle> = Fun f (take (Suc i) ts @ (ss!(Suc i)) # drop (Suc(Suc i)) ss)" by simp
   also have "\<dots> = Fun f ((take i ts @ [ts!i]) @ (ss!(Suc i)) # drop (Suc(Suc i)) ss)"
    using take_Suc_conv_app_nth[OF `i < length ts`] by simp
   also have "\<dots> = Fun f ((take i ts @ [ts!i]) @ drop (Suc i) ss)"
    using nth_drop'[OF `Suc i < length ss`] by simp
   also have "\<dots> = Fun f (take i ts @ (ts!i) # drop (Suc i) ss)" by simp
   also have "\<dots> = (More f (take i ts) \<box> (drop (Suc i) ss))\<langle>ts!i\<rangle>" by simp
   finally have b:"(?C(Suc i))\<langle>ss!(Suc i)\<rangle> = (?C i)\<langle>ts!i\<rangle>" by auto
   from IH have c: "(Fun f ss,(?C(Suc i))\<langle>ss!(Suc i)\<rangle>) \<in> (rstep R)^*" unfolding b[symmetric] by simp
   from Suc args have "(ss!(Suc i),ts!(Suc i)) \<in> (rstep R)^*" by simp
   hence "((?C(Suc i))\<langle>ss!(Suc i)\<rangle>,(?C(Suc i))\<langle>ts!(Suc i)\<rangle>) \<in> (rstep R)^*" by (rule rsteps_closed_ctxt)
   with c show ?case by auto
  qed
 qed
 have "ss = [] \<or> (\<exists>ss' s'. ss = ss'@[s'])" by (rule Nil_or_last)
 thus ?thesis
 proof
  assume "ss = []"
  hence "ts = []" using `length ss = length ts` by simp
  show ?thesis unfolding `ss = []` `ts = []` by auto
 next
  assume "\<exists>ss' s'. ss = ss' @ [s']"
  then obtain s' ss' where "ss = ss'@[s']" by auto
  let ?i = "length ss'"
  have "?i < length ss" using `ss = ss'@[s']` by simp
  have "ts = [] \<or> (\<exists>ts' t'. ts = ts'@[t'])" by (rule Nil_or_last)
  thus ?thesis
  proof
   assume "ts = []" with `ss = ss'@[s']` and `length ss = length ts` show ?thesis by simp
  next
   assume "\<exists>ts' t'. ts = ts'@[t']"
   then obtain t' ts' where "ts = ts'@[t']" by auto
   with a and `?i < length ss` have b:"(Fun f ss,(?C ?i)\<langle>(ts'@[t'])!?i\<rangle>) \<in> (rstep R)^*" by simp
   from `length ss = length ts` have "length ss' = length ts'"
    unfolding `ss = ss'@[s']` `ts = ts'@[t']` by simp
   hence "?i = length ts'" by simp
   have "(?C ?i)\<langle>(ts'@[t'])!?i\<rangle> = Fun f (take ?i ts @ ((ts'@[t'])!?i) # drop (Suc ?i) ss)" by simp
   also have "\<dots> = Fun f (take ?i (ts'@[t']) @ ((ts'@[t'])!?i) # drop (Suc ?i) (ss'@[s']))"
    unfolding `ss = ss'@[s']` and `ts = ts'@[t']` by simp
   also have "\<dots> = Fun f (ts' @ ((ts'@[t'])!?i) # drop (Suc ?i) (ss'@[s']))"
    unfolding `length ss' = length ts'` by auto
   also have "\<dots> = Fun f (ts' @ t' # drop (Suc ?i) (ss'@[s']))"
    unfolding `length ss' = length ts'` by auto
   also have "\<dots> = Fun f (ts' @ t' # [])" by auto
   also have "\<dots> = Fun f (ts @ [])" using `ts = ts'@[t']` by simp
   finally have "(?C ?i)\<langle>(ts'@[t'])!?i\<rangle> = Fun f ts" by simp
   with b show ?thesis by auto
  qed
 qed
qed

lemma nrrstep_Fun_imp_arg_rstep:
 fixes ss :: "('f,'v)term list"
 assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R")
 shows "\<exists>C i. i < length ss \<and> (ss!i,ts!i) \<in> rstep R \<and> C\<langle>ss!i\<rangle> = Fun f ss \<and> C\<langle>ts!i\<rangle> = Fun f ts"
proof -
 from `(?s,?t) \<in> nrrstep R`
 obtain l r j ps \<sigma> where A: "let C = ctxt_of_pos_term (j<#ps) ?s in (j<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding nrrstep_def rstep_r_p_s_def by force
 let ?C = "ctxt_of_pos_term (j<#ps) ?s"
 from A have "(j<#ps) \<in> poss ?s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" using Let_def by auto
 have C: "?C = More f (take j ss) (ctxt_of_pos_term ps (ss!j)) (drop (Suc j) ss)" (is "?C = More f ?ss1 ?D ?ss2") by auto
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = (ss!j)" by (auto simp: take_drop_imp_nth)
 from `(l,r) \<in> R` have "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> (CS R)" by auto
 hence "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" by (auto simp only: ctxt_closure.ctxt_closureI)
 hence D_rstep: "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R"
   by (auto simp: rstep_CC_CS)
 from `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` and C have "?t = Fun f (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 hence ts: "ts = (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 from `j<#ps \<in> poss ?s` have r0: "j < length ss" by simp
 hence "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append_take)
 with ts have "ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 let ?C' = "More f (take j ss) \<box> (drop (Suc j) ss)"
 from D_rstep have r1: "(ss!j,ts!j) \<in> rstep R" unfolding `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` by simp
 have "?s = ?C\<langle>l\<cdot>\<sigma>\<rangle>" unfolding `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` by simp
 also have "\<dots> = ?C'\<langle>?D\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>" unfolding C by simp
 finally have r2:"?C'\<langle>ss!j\<rangle> = ?s" unfolding `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` by simp
 have "?t = ?C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` by simp
 also have "\<dots> = ?C'\<langle>?D\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>" unfolding C by simp
 finally have r3:"?C'\<langle>ts!j\<rangle> = ?t" unfolding `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` by simp
 from r0 r1 r2 r3 show ?thesis by best
qed


lemma pair_fun_eq[simp]:
 fixes f :: "'a \<Rightarrow> 'b" and g :: "'b \<Rightarrow> 'a"
 shows "((\<lambda>(x,y). (x,f y)) \<circ> (\<lambda>(x,y). (x,g y))) = (\<lambda>(x,y). (x,(f \<circ> g) y))" (is "?f = ?g")
proof (rule ext)
 fix ab :: "'c * 'b"
 obtain a b where "ab = (a,b)" by force
 have "((\<lambda>(x,y). (x,f y))\<circ>(\<lambda>(x,y). (x,g y))) (a,b) = (\<lambda>(x,y). (x,(f\<circ>g) y)) (a,b)" by simp
 thus "?f ab = ?g ab" unfolding `ab = (a,b)` by simp
qed


lemma restrict_singleton:
  assumes "x \<in> subst_domain \<sigma>" shows "\<exists>t. \<sigma> |s {x} = (\<lambda>y. if y = x then t else Var y)"
proof -
  have "\<sigma> |s {x} = (\<lambda>y. if y = x then \<sigma> y else Var y)" by (simp add: subst_restrict_def)
  hence "\<sigma> |s {x} = (\<lambda>y. if y = x then \<sigma> x else Var y)" by (simp cong: if_cong)
  thus ?thesis by (rule exI[of _ "\<sigma> x"])
qed

definition rstep_r_c_s :: "('f,'v)rule \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)subst \<Rightarrow> ('f,'v)term ars"
where "rstep_r_c_s r C \<sigma> = {(s,t) | s t. s = C\<langle>fst r \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>snd r \<cdot> \<sigma>\<rangle>}"

lemma rstep_iff_rstep_r_c_s: "((s,t) \<in> rstep R) = (\<exists> l r C \<sigma>. (l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>)" (is "?left = ?right")
proof
  assume ?left
  then obtain l r p \<sigma> where step: "(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
    unfolding rstep_iff_rstep_r_p_s by blast
  obtain D where D: "D = ctxt_of_pos_term p s" by auto
  with step have Rrule: "(l,r) \<in> R" and s: "s = D\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = D\<langle>r \<cdot> \<sigma>\<rangle>" 
    unfolding rstep_r_p_s_def by (force simp: Let_def)+
  thus ?right unfolding rstep_r_c_s_def by auto
next
  assume ?right
  from this obtain l r C \<sigma> where "(l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>" by auto
  hence rule: "(l,r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    unfolding rstep_r_c_s_def by auto
  show ?left unfolding rstep_CC_CS
    by (simp add: s t, rule ctxt_closureI, rule subst_closureI, rule rule)
qed

lemma rstep_subset_characterization:
  "(rstep R \<subseteq> rstep S) = (\<forall> l r. (l,r) \<in> R \<longrightarrow> (\<exists> l' r' C \<sigma> . (l',r') \<in> S \<and> l = C\<langle>l' \<cdot> \<sigma>\<rangle> \<and> r = C\<langle>r' \<cdot> \<sigma>\<rangle>))" (is "?left = ?right")
proof
  assume ?right 
  show ?left
  proof (rule subsetI, simp add: split_paired_all)
    fix s t
    assume "(s,t) \<in> rstep R"
    then obtain l r C \<sigma> where step: "(l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>"
      unfolding rstep_iff_rstep_r_c_s by best
    hence Rrule: "(l,r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" 
      unfolding rstep_r_c_s_def by (force)+
    from Rrule `?right` obtain l' r' C' \<sigma>' where Srule: "(l',r') \<in> S" and l: "l = C'\<langle>l' \<cdot> \<sigma>'\<rangle>" and r: "r = C'\<langle>r' \<cdot> \<sigma>'\<rangle>"
      by (force simp: Let_def)+
    let ?D = "C \<circ>c (C' \<cdot>c \<sigma>)"
    let ?sig = "\<sigma>' \<circ>s \<sigma>"
    have s2: "s = ?D\<langle>l' \<cdot> ?sig\<rangle>" by (simp add: s l)
    have t2: "t = ?D\<langle>r' \<cdot> ?sig\<rangle>" by (simp add: t r)
    from s2 t2 have sStep: "(s,t) \<in> rstep_r_c_s (l',r') ?D ?sig" unfolding rstep_r_c_s_def by force
    with Srule show "(s,t) \<in> rstep S" by (simp only: rstep_iff_rstep_r_c_s, blast)
  qed
next
  assume ?left
  show ?right
  proof (rule ccontr)
    assume "\<not> ?right"
    from this obtain l r where "(l,r) \<in> R" and cond: "\<forall> l' r' C \<sigma>. (l',r') \<in> S \<longrightarrow> (l \<noteq> C\<langle>l' \<cdot> \<sigma>\<rangle> \<or> r \<noteq> C\<langle>r' \<cdot> \<sigma>\<rangle>)" by blast
    hence "(l,r) \<in> rstep R" by blast
    with `?left` have "(l,r) \<in> rstep S" by auto
    with cond show False by (simp only: rstep_iff_rstep_r_c_s, unfold rstep_r_c_s_def, force)
  qed
qed

lemma rstep_preserves_wf_terms:
  assumes "funas_trs R \<subseteq> F" and "s \<in> wf_terms F" and "(s,t) \<in> rstep R" and "wf_trs R"
  shows "t \<in> wf_terms F"
proof -
  from `(s,t) \<in> rstep R` obtain l r C \<sigma> where R: "(l,r) \<in> R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" using rstep_imp_C_s_r[of s t R] by fast
  from `s \<in> wf_terms F` have "funas_term s \<subseteq> F" by (induct s) auto
  from `funas_trs R \<subseteq> F` and R have "funas_term r \<subseteq> F" unfolding funas_defs by force
  from `wf_trs R` and R have "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by simp
  have "t \<in> wf_terms (funas_term t)" by (rule wf_terms_funas_term)
  have "funas_term s = funas_ctxt C \<union> funas_term(l\<cdot>\<sigma>)" unfolding s by simp
  hence funas_s: "funas_term s = funas_ctxt C \<union> funas_term l \<union> (\<Union>x\<in>vars_term l. funas_term(Var x \<cdot> \<sigma>))" using funas_term_subst_apply[of l \<sigma>] by auto
  have "funas_term t = funas_ctxt C \<union> funas_term(r\<cdot>\<sigma>)" unfolding t by simp
  hence "funas_term t = funas_ctxt C \<union> funas_term r \<union> (\<Union>x\<in>vars_term r. funas_term(Var x \<cdot> \<sigma>))" using funas_term_subst_apply[of r \<sigma>] by auto
  hence "funas_term t \<subseteq> funas_ctxt C \<union> funas_term r \<union> (\<Union>x\<in>vars_term l. funas_term(Var x \<cdot> \<sigma>))" using `vars_term r \<subseteq> vars_term l` by auto
  with `funas_term r \<subseteq> F` `funas_term s \<subseteq> F` have "funas_term t \<subseteq> F" unfolding funas_s by blast
  from wf_terms_subset[OF `funas_term t \<subseteq> F`] and `t \<in> wf_terms (funas_term t)` show ?thesis by auto
qed

lemma no_Var_rstep[simp]: fixes R::"('f,'v)trs" assumes "wf_trs R" and "(Var x,t) \<in> rstep R" shows "False"
using assms unfolding wf_trs_def rstep_iff_rstep_r_p_s rstep_r_p_s_def Let_def by force

lemma lhs_wf:
  assumes R: "(l,r) \<in> R" and subset: "funas_trs R \<subseteq> F" shows "l \<in> wf_terms F"
proof -
  have "funas_term l \<subseteq> funas_trs R" using R unfolding funas_trs_def funas_rule_def by auto
  hence "wf_terms (funas_term l) \<subseteq> wf_terms (funas_trs R)" by (rule wf_terms_subset)
  moreover have "l \<in> wf_terms (funas_term l)" by simp
  ultimately have "l \<in> wf_terms (funas_trs R)" by auto
  moreover from subset have "wf_terms (funas_trs R) \<subseteq> wf_terms (F)" by (rule wf_terms_subset)
  ultimately show ?thesis by auto
qed

lemma rhs_wf:
  assumes R: "(l,r) \<in> R" and subset: "funas_trs R \<subseteq> F" shows "r \<in> wf_terms F"
proof -
  have "funas_term r \<subseteq> funas_trs R" using R unfolding funas_trs_def funas_rule_def by auto
  hence "wf_terms (funas_term r) \<subseteq> wf_terms (funas_trs R)" by (rule wf_terms_subset)
  moreover have "r \<in> wf_terms (funas_term r)" by simp
  ultimately have "r \<in> wf_terms (funas_trs R)" by auto
  moreover from subset have "wf_terms (funas_trs R) \<subseteq> wf_terms F" by (rule wf_terms_subset)
  ultimately show ?thesis by auto
qed

lemma subst_dom: "s\<cdot>\<sigma> = s\<cdot>\<tau> \<Longrightarrow> vars_term t \<subseteq> vars_term s \<Longrightarrow> t\<cdot>\<sigma> = t\<cdot>\<tau>" by (induct t) (induct s,auto)

lemma subst_Var_imp_Var: assumes "s\<cdot>\<sigma> = Var x" shows "\<exists>y. s = Var y" using assms by (induct s) auto 

lemma supt_map_term[intro]: "t \<rhd> s \<Longrightarrow> map_term fg t \<rhd> map_term fg s"
proof (induct rule: supt.induct)
  case (arg s ss f)
  hence "map_term fg s \<in> set(map (map_term fg) ss)" by simp
  thus ?case unfolding map_term.simps by (rule supt.arg)
next
  case (subt s ss u f)
  hence "map_term fg s \<in> set(map (map_term fg) ss)" by simp
  from this and `map_term fg s \<rhd> map_term fg u` show ?case
    unfolding map_term.simps by (rule supt.subt)
qed


lemma nondef_root_imp_arg_step:
  assumes "(Fun f ss,t) \<in> rstep R"
      and wf: "\<forall>lr\<in>R. \<forall>x. fst lr \<noteq> Var x"
      and "(f,length ss) \<notin> defs R"
  shows "\<exists>i < length ss. (ss!i,t|_<i>) \<in> rstep R \<and> t = Fun f (take i ss@t|_<i>#drop (Suc i) ss)"
proof -
  from assms obtain l r p \<sigma> where rstep_r_p_s: "(Fun f ss,t) \<in> rstep_r_p_s R (l,r) p \<sigma>" unfolding rstep_iff_rstep_r_p_s by auto
  let ?C = "ctxt_of_pos_term p (Fun f ss)"
  from rstep_r_p_s have "p \<in> poss(Fun f ss)" and "(l,r) \<in> R"
    and "?C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
  have "\<exists>i q. p = i<#q"
  proof (cases p)
    case PCons thus ?thesis by auto
  next
    case Empty
    have "?C = \<box>" unfolding Empty by simp
    with `?C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss` have "l\<cdot>\<sigma> = Fun f ss" by simp
    have "\<forall>(l,r)\<in>R. \<exists>f ss. l = Fun f ss"
    proof (intro ballI impI)
      fix lr assume "lr \<in> R"
      with wf have "\<forall>x. fst lr \<noteq> Var x" by auto
      hence "\<exists>f ss. (fst lr) = Fun f ss" by (cases "fst lr") auto
      thus "(\<lambda>(l,r). \<exists>f ss. l = Fun f ss) lr" by auto
    qed
    with `(l,r) \<in> R` obtain g ts where "l = Fun g ts" unfolding wf_trs_def by best
    with `l\<cdot>\<sigma> = Fun f ss` have "f = g" by auto
    from `l\<cdot>\<sigma> = Fun f ss` and `l = Fun g ts` and `(l, r) \<in> R`
      have "(f, length ss) \<in> defs R" unfolding defs_def defined_def `f = g` by auto
    with `(f,length ss) \<notin> defs R` show ?thesis by simp
  qed
  then obtain i q where "p = i<#q" by auto
  from `p \<in> poss(Fun f ss)` have "i < length ss" and "q \<in> poss(ss!i)" unfolding `p = i<#q` by auto
  let ?D = "ctxt_of_pos_term q (ss!i)"
  have C: "?C = More f (take i ss) ?D (drop (Suc i) ss)" unfolding `p = i<#q` by auto
  from `?C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss` have "take i ss@?D\<langle>l\<cdot>\<sigma>\<rangle>#drop (Suc i) ss = ss" unfolding C by auto
  hence "(take i ss@?D\<langle>l\<cdot>\<sigma>\<rangle>#drop (Suc i) ss)!i = ss!i" by simp
  with `i < length ss` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!i" using nth_append_take[where xs = ss and i = i] by auto
  have t: "t = Fun f (take i ss@?D\<langle>r\<cdot>\<sigma>\<rangle>#drop (Suc i) ss)" unfolding `?C\<langle>r\<cdot>\<sigma>\<rangle> = t`[symmetric] C by simp
  from `i < length ss` have "t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>" unfolding t unfolding subt_at.simps using nth_append_take[where xs = ss and i = i] by auto
  from `q \<in> poss(ss!i)` and `(l,r) \<in> R` 
    and `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!i` and `t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>`[symmetric]
    have "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) q \<sigma>" unfolding rstep_r_p_s_def Let_def by auto
  hence "(ss!i,t|_<i>) \<in> rstep R" unfolding rstep_iff_rstep_r_p_s by auto
  from `i < length ss` and this and t show ?thesis unfolding `t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>`[symmetric] by auto
qed

lemma nondef_root_imp_arg_steps:
  assumes "(Fun f ss,t) \<in> (rstep R)^*"
      and wf: "\<forall>lr\<in>R. \<forall>x. fst lr \<noteq> Var x"
      and "(f,length ss) \<notin> defs R"
  shows "\<exists>ts. length ts = length ss \<and> t = Fun f ts \<and> (\<forall>i<length ss. (ss!i,ts!i) \<in> (rstep R)^*)"
proof -
  from assms obtain n where "(Fun f ss,t) \<in> (rstep R)^^n" using rtrancl_imp_rel_pow' by best
  thus ?thesis
  proof (induct n arbitrary: t)
    case 0 thus ?case by auto
  next
    case (Suc n)
    then obtain u where "(Fun f ss,u) \<in> (rstep R)^^n" and "(u,t) \<in> rstep R" by auto
    with Suc obtain ts where IH1: "length ts = length ss" and IH2: "u = Fun f ts" and IH3: "\<forall>i<length ss. (ss!i,ts!i) \<in> (rstep R)^*" by auto
    from `(u,t) \<in> rstep R` have "(Fun f ts,t) \<in> rstep R" unfolding `u = Fun f ts` .
    from nondef_root_imp_arg_step[OF this wf `(f,length ss) \<notin> defs R`[simplified IH1[symmetric]]]
      obtain j where "j < length ts"
      and "(ts!j,t|_<j>) \<in> rstep R"
      and B: "t = Fun f (take j ts@t|_<j>#drop (Suc j) ts)" (is "t = Fun f ?ts") by auto
    from `j < length ts` have "length ?ts = length ts" by auto
    hence A: "length ?ts = length ss" unfolding `length ts = length ss` .
    have C: "\<forall>i<length ss. (ss!i,?ts!i) \<in> (rstep R)^*"
    proof (intro allI, intro impI)
      fix i assume "i < length ss"
        from `i < length ss` and IH3 have "(ss!i,ts!i) \<in> (rstep R)^*" by auto
      have "i = j \<or> i \<noteq> j" by auto
      thus "(ss!i,?ts!i) \<in> (rstep R)^*"
      proof
        assume "i = j"
        from `j < length ts` have "j \<le> length ts" by simp
        from nth_append_take[OF this] have "?ts!j = t|_<j>" by simp
        from `(ts!j,t|_<j>) \<in> rstep R` have "(ts!i,t|_<i>) \<in> rstep R" unfolding `i = j` .
        with `(ss!i,ts!i) \<in> (rstep R)^*` show ?thesis unfolding `i = j` unfolding `?ts!j = t|_<j>` by auto
      next
        assume "i \<noteq> j"
        from `i < length ss` have "i \<le> length ts" unfolding `length ts = length ss` by simp
        from `j < length ts` have "j \<le> length ts" by simp
        from nth_append_take_drop_is_nth_conv[OF `i \<le> length ts` `j \<le> length ts` `i \<noteq> j`]
          have "?ts!i = ts!i" by simp
        with `(ss!i,ts!i) \<in> (rstep R)^*` show ?thesis by auto
      qed
    qed
    from A and B and C show ?case by blast
  qed
qed

lemma rstep_imp_nrrstep:
  assumes "\<not> is_var s" and "\<not> defined R (the (root s), num_args s)" and "\<forall>(l,r)\<in>R. \<not> is_var l"
    and "(s, t) \<in> rstep R"
  shows "(s, t) \<in> nrrstep R"
proof -
  from `\<not> is_var s` obtain f ss where s: "s = Fun f ss" by (cases s) auto
  from assms have undef: "(f, length ss) \<notin> defs R" by (simp add: defs_def s)
  from assms have non_var: "\<forall>lr\<in>R. \<forall>x. fst lr \<noteq> Var x" by auto
  from nondef_root_imp_arg_step[OF `(s, t) \<in> rstep R`[unfolded s] non_var undef]
    obtain i where "i < length ss" and step: "(ss ! i, t |_ <i>) \<in> rstep R"
      and t: "t = Fun f (take i ss @ t |_ <i> # drop (Suc i) ss)" by auto
  from step obtain C l r \<sigma> where "(l, r) \<in> R" and lhs: "ss ! i = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and rhs: "(t |_ <i>) = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  let ?C = "More f (take i ss) C (drop (Suc i) ss)"
  have "(l, r) \<in> R" by fact
  moreover have "?C \<noteq> \<box>" by simp
  moreover have "s = ?C\<langle>l \<cdot> \<sigma>\<rangle>"
  proof -
    have "s = Fun f (take i ss @ ss!i # drop (Suc i) ss)"
      using id_take_nth_drop[OF `i < length ss`] unfolding s by simp
    also have "\<dots> = ?C\<langle>l \<cdot> \<sigma>\<rangle>" by (simp add: lhs)
   finally show ?thesis .
  qed
  moreover have "t = ?C\<langle>r \<cdot> \<sigma>\<rangle>"
  proof -
    have "t = Fun f (take i ss @ t |_ <i> # drop (Suc i) ss)" by fact
    also have "\<dots> = Fun f (take i ss @ C\<langle>r \<cdot> \<sigma>\<rangle> # drop (Suc i) ss)" by (simp add: rhs)
    finally show ?thesis by simp
  qed
  ultimately show "(s, t) \<in> nrrstep R" unfolding nrrstep_def' by blast
qed

lemma rsteps_imp_nrrsteps:
  assumes "\<not> is_var s" and "\<not> defined R (the (root s), num_args s)"
    and no_vars: "\<forall>(l,r)\<in>R. \<not> is_var l"
    and "(s, t) \<in> (rstep R)^*"
  shows "(s, t) \<in> (nrrstep R)^*"
using `(s, t) \<in> (rstep R)^*`
proof (induct)
  case base show ?case by simp
next
  case (step u v)
  from assms obtain f ss where s: "s = Fun f ss" by (induct s) auto
  from nrrsteps_preserve_root[OF `(s, u) \<in> (nrrstep R)^*`[unfolded s]]
    obtain ts where u: "u = Fun f ts" by auto
  from nrrsteps_equiv_num_args[OF `(s, u) \<in> (nrrstep R)^*`[unfolded s]]
    have len: "length ss = length ts" unfolding u by simp
  have "\<not> is_var u" by (simp add: u)
  have undef: "\<not> defined R (the (root u), num_args u)"
    using `\<not> defined R (the (root s), num_args s)`
    unfolding s u by (simp add: len)
  have "(u, v) \<in> nrrstep R"
    using rstep_imp_nrrstep[OF `\<not> is_var u` undef no_vars] step
    by simp
  with step show ?case by auto
qed

lemma left_var_imp_not_SN:
  fixes R :: "('f,'v)trs" and t :: "('f,'v)term"
  assumes "(Var y,r) \<in> R" (is "(?y,_) \<in> _")
  shows "\<not> (SN_elt (rstep R) t)"
proof -
  have steps: "\<forall> t. \<exists> s. (t,s) \<in> rstep R"
  proof 
    fix t :: "('f,'v)term"
    let ?yt = "subst y t"
    let ?ryt = "subst y (r \<cdot> ?yt)"
    from assms have "(?y \<cdot> ?yt, r \<cdot> ?yt) \<in> CS R" by (rule  subst_closureI)
    hence "(\<box>\<langle>?y \<cdot> ?yt\<rangle>, \<box>\<langle>r \<cdot> ?yt\<rangle>) \<in> CC (CS R)" by (rule ctxt_closureI)
    hence "(t, r \<cdot> ?yt) \<in> rstep R" unfolding subst_def rstep_CC_CS by auto 
    thus "\<exists> s. (t,s) \<in> rstep R" ..
  qed
  show ?thesis 
  proof
    assume "SN_elt (rstep R) t"
    from this obtain s where norm: "(t,s) \<in> (rstep R)^!"
      using SN_elt_imp_normalizability[of "rstep R" t] by force
    from steps obtain u where "(s,u) \<in> rstep R" by auto
    with norm show False by auto
  qed
qed

lemma not_SN_subt_imp_not_SN:
 assumes ctxt: "ctxt_closed R" and SN: "\<not> SN_elt R t" and sub: "s \<unrhd> t"
 shows "\<not> SN_elt R s"
using SN_rel_imp_SN_subt[OF ctxt _ sub] SN by auto

lemma SN_args_imp_SN :
  assumes SN: "\<And> s. s \<in> set ss \<Longrightarrow> SN_elt (rstep R) s"
  and nvar: "\<forall> lr \<in> R. \<forall>  x.  fst lr \<noteq> Var x"
  and ndef: "(f,length ss) \<notin> defs R"
  shows "SN_elt (rstep R) (Fun f ss)"
proof (rule ccontr)
  assume nSN: "\<not> SN_elt (rstep R) (Fun f ss)"
  have SN: "\<And> s. Fun f ss \<rhd> s \<Longrightarrow> SN_elt (rstep R) s"
  proof -
    fix s
    assume "Fun f ss \<rhd> s"
    then obtain C where "C \<noteq> \<box>" and "Fun f ss = C\<langle>s\<rangle>" ..
    then obtain bef D aft where "ss = bef @ D\<langle>s\<rangle> # aft" by (cases C, auto)
    hence "D\<langle>s\<rangle> \<in> set ss" by simp
    from SN[OF this] not_SN_subt_imp_not_SN[OF ctxt_closed_rstep _ ctxt_imp_supteq[of D s]]
    show "SN_elt (rstep R) s" by auto
  qed
  let ?t = "Fun f ss"
  have t: "?t \<in> Tinf R" unfolding Tinf_def using nSN SN by best
  from nSN obtain S where "S 0 = ?t" and seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R" by auto
  have "\<exists>i. (S 0, S i) \<in> (nrrstep R)^* \<and> (S i, S (Suc i)) \<in> rrstep R"
  proof -
    from t have "SN_elt (nrrstep R) (S 0)"
      by (simp add: Tinf_imp_SN_elt_nrrstep `S 0 = ?t`)
    with seq union_iseq_SN_elt_imp_first_step[of S "rrstep R" "nrrstep R"]
      obtain j where first_step: "(S j, S (Suc j)) \<in> rrstep R"
      and nrsteps: "\<forall>i<j. (S i, S (Suc i)) \<in> nrrstep R"
      by (auto simp: rstep_iff_rrstep_or_nrrstep)
    from nrsteps have "(S 0, S j) \<in> (nrrstep R)^*" by (induct j) auto
    with first_step show ?thesis by auto
  qed
  then obtain j where nrseq: "(S 0, S j) \<in> (nrrstep R)^*"
    and rrstep: "(S j, S (Suc j)) \<in> rrstep R" by auto
  from rrstep_imp_rule_subst[OF rrstep]
    obtain l r \<sigma> where "(l, r) \<in> R" and "l\<cdot>\<sigma> = S j" and "r\<cdot>\<sigma> = S (Suc j)" by auto
  from nrseq have "(?t, S j) \<in> (nrrstep R)^*" by (simp add: `S 0 = ?t`)
  moreover from nrrsteps_preserve_root[OF this] nrrsteps_equiv_num_args[OF this] obtain ls'
    where "S j = Fun f ls'" and "length ss = length ls'" by auto
  ultimately have "l\<cdot>\<sigma> = Fun f ls'"
    by (auto simp: `l\<cdot>\<sigma> = S j`) 
  from nvar `(l,r) \<in> R` obtain g ls 
    where "l = Fun g ls" by (cases l, force+)
  with `l\<cdot>\<sigma> = Fun f ls'` `(l,r) \<in> R` `length ss = length ls'` ndef have "(Fun f ls,r) \<in> R" and "(f, length ls) \<notin> defs R" by auto
  thus False unfolding defs_def defined_def by force
qed


lemma Some_root_imp_Fun: "root t = Some f \<Longrightarrow> \<exists> ss. t = Fun f ss" by (induct t) auto


lemma map_rule_power:
  fixes f :: "'f \<Rightarrow> 'f" 
  shows "((map_rule f) ^^ n) = map_rule (f ^^ n)"
proof (rule sym, intro ext, clarify) 
  fix  l r
  show "map_rule (f ^^ n) (l,r) = (map_rule f ^^ n) (l,r)"
  proof (induct n)
    case 0
    show ?case
      by (simp, unfold id_def[symmetric] map_term_ident, simp)
  next
    case (Suc n)
    have "map_rule (f ^^ Suc n) (l,r) = map_rule f (map_rule (f ^^ n) (l,r))" 
      by (simp add: map_term_comp)
    also have "\<dots> = map_rule f ((map_rule f ^^ n) (l,r))" unfolding Suc ..
    also have "\<dots> = (map_rule f ^^ (Suc n)) (l,r)" by simp
    finally show ?case .
  qed
qed

lemma map_trs_power: fixes f :: "'f \<Rightarrow> 'f" 
  shows "((map_trs f) ^^ n) = map_trs (f ^^ n)"
proof (rule sym, rule ext, unfold map_trs.simps) 
  fix R :: "('f,'v)trs"
  obtain g where g: "(map_rule f :: ('f,'v)rule \<Rightarrow> ('f,'v)rule) = g" by simp
  have "map_rule (f ^^ n) ` R = (map_rule f ^^ n) ` R" unfolding map_rule_power ..
  also have "\<dots> = ((\<lambda> R. map_trs f R) ^^ n) R" unfolding map_trs.simps g
    by (induct n, force+)
  finally show "map_rule (f ^^ n) ` R = (map_trs f ^^ n) R" .
qed


fun
  sharp_term :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v)term \<Rightarrow> ('f,'v) term"
where
  "sharp_term shp (Var x) = Var x" |
  "sharp_term shp (Fun f ss) = Fun (shp f) ss"

definition
  DP :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs"
where
  "DP shp R = {(s,t). \<exists>l r h us. s = sharp_term shp l \<and> t = sharp_term shp (Fun h us)
     \<and> (l,r) \<in> R \<and> r \<unrhd> (Fun h us) \<and> defined R (h, length us) \<and> \<not> l \<rhd> Fun h us}"

lemma sharp_term_subst_distr[simp]:
  "sharp_term shp (Fun f ss \<cdot> \<sigma>) = sharp_term shp (Fun f ss) \<cdot> \<sigma>"
by simp
 
lemma nrrstep_imp_sharp_rstep:
  assumes "(s, t) \<in> nrrstep R"
  shows "(sharp_term shp s, sharp_term shp t) \<in> rstep R"
proof -
  from assms obtain l r i ps \<sigma> where "(s,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>" unfolding nrrstep_def by best
  hence "let C = ctxt_of_pos_term (i<#ps) s in (i<#ps) \<in> poss s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = t"
  (is "let C = ?C in _") unfolding rstep_r_p_s_def by simp
  hence "(l,r) \<in> R" and s: "?C\<langle>l\<cdot>\<sigma>\<rangle> = s" and t: "?C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding Let_def by auto
  from `(s,t) \<in> nrrstep R` obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by best
  hence C: "?C = More f (take i ss) (ctxt_of_pos_term ps (ss!i)) (drop (Suc i) ss)"
    (is "?C = More f ?ss1 ?D ?ss2") by simp
  let ?C' = "More (shp f) ?ss1 ?D ?ss2"
  from `(l,r) \<in> R` have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> CS R" by (rule subst_closureI)
  hence "(?C'\<langle>l \<cdot> \<sigma>\<rangle>, ?C'\<langle>r \<cdot> \<sigma>\<rangle>) \<in> CC(CS R)"
    by (rule ctxt_closureI)
  hence R_step: "(?C'\<langle>l \<cdot> \<sigma>\<rangle>, ?C'\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep R"
    by (simp add: rstep_CC_CS)
  from `?C\<langle>l\<cdot>\<sigma>\<rangle> = s` have "sharp_term shp s = sharp_term shp (?C\<langle>l \<cdot> \<sigma>\<rangle>)" by simp
  hence s: "sharp_term shp s = ?C'\<langle>l \<cdot> \<sigma>\<rangle>" using C by simp
  from `?C\<langle>r\<cdot>\<sigma>\<rangle> = t` have "sharp_term shp t = sharp_term shp (?C\<langle>r\<cdot>\<sigma>\<rangle>)" by simp
  hence t: "sharp_term shp t = ?C'\<langle>r \<cdot> \<sigma>\<rangle>" by (simp add: C)
  from R_step show ?thesis unfolding s t .
qed

lemma nrrsteps_imp_sharp_rsteps:
  "(s,t) \<in> (nrrstep R)^* \<Longrightarrow> (sharp_term shp s, sharp_term shp t) \<in> (rstep R)^*"
proof (induct rule: rtrancl_induct)
  case base thus ?case by simp
next
  case (step a b)
  from `(a,b) \<in> nrrstep R` have "(sharp_term shp a, sharp_term shp b) \<in> rstep R"
    by (rule nrrstep_imp_sharp_rstep)
   with step show ?case by auto
qed

lemma Tinf_lemma_DP:
  assumes "wf_trs R" "s \<in> Tinf R"
  shows "\<exists>t u. t \<in> Tinf R \<and> u \<in> Tinf R \<and>
    (sharp_term shp s, sharp_term shp t) \<in> (rstep R)^* \<and>
    (sharp_term shp t, sharp_term shp u) \<in> CS (DP shp R)"
proof -
  obtain l r u \<sigma> where "(l, r) \<in> R" and "r \<unrhd> u" and "\<not> is_var u"
    and nrseq: "(s,l\<cdot>\<sigma>) \<in> (nrrstep R)^*"
    and "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rrstep R" and "l\<cdot>\<sigma> \<in> Tinf R" and "u\<cdot>\<sigma> \<in> Tinf R"
    and "\<not> l\<cdot>\<sigma> \<rhd> (u\<cdot>\<sigma>)"
    using Tinf_lemma[where ?R = "R" and ?t = "s",OF `wf_trs R` `s \<in> Tinf R`] by auto
  let ?l = "sharp_term shp l"
  let ?r = "sharp_term shp u"
  from `\<not> is_var u` obtain f us where "u = Fun f us" by (cases u) auto
  hence "u\<cdot>\<sigma> = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) us)" (is "_ = Fun f ?us'") by simp
  hence "root u = root (u\<cdot>\<sigma>)" unfolding `u = Fun f us` by simp
  from `wf_trs R` and `u\<cdot>\<sigma> \<in> Tinf R` obtain h vs where "(u\<cdot>\<sigma>) = Fun h vs"
    and "defined R (h, length vs)"
    using Tinf_imp_def_root by best
  from `\<not> l\<cdot>\<sigma> \<rhd> (u\<cdot>\<sigma>)` have "\<not> l \<rhd> u" by (rule instance_no_subt_imp_no_subt)
  from nrseq have "(sharp_term shp s, sharp_term shp (l\<cdot>\<sigma>)) \<in> (rstep R)^*"
    by (rule nrrsteps_imp_sharp_rsteps)
  from `(l,r) \<in> R` and `wf_trs R` obtain g ts
    where "l = Fun g ts" unfolding wf_trs_def by blast
  from `u\<cdot>\<sigma> = Fun h vs` and `u = Fun f us` have "f = h" and "length us = length vs" by auto
  hence "defined R (f, length us)" using `defined R (h, length vs)` by simp
  from `(l, r) \<in> R` and `r \<unrhd> u` and `defined R (f, length us)` and `\<not> l \<rhd> u`
    have "(?l, ?r) \<in> DP shp R" unfolding DP_def `u = Fun f us` by auto
  hence "(?l\<cdot>\<sigma>, ?r\<cdot>\<sigma>) \<in> CS (DP shp R)" by auto
  hence "(sharp_term shp (l\<cdot>\<sigma>), sharp_term shp (u\<cdot>\<sigma>)) \<in> CS (DP shp R)"
    unfolding `l = Fun g ts` `u = Fun f us` sharp_term_subst_distr by simp
  with `l\<cdot>\<sigma> \<in> Tinf R` and `u\<cdot>\<sigma> \<in> Tinf R` and `(sharp_term shp s, sharp_term shp (l\<cdot>\<sigma>)) \<in> (rstep R)^*`
    show ?thesis by auto
qed


lemma Tinf_lemma_DP_explicit_subst:
  assumes "wf_trs R" and "s \<in> Tinf R"
  shows "\<exists>l r u \<sigma>. (l, r) \<in> DP shp R \<and> r\<cdot>\<sigma> = sharp_term shp u \<and> u \<in> Tinf R
    \<and> (sharp_term shp s, l\<cdot>\<sigma>) \<in> (rstep R)^*"
proof -
  from Tinf_lemma_DP[OF assms, of shp] obtain t u
    where t: "t \<in> Tinf R" and u: "u \<in> Tinf R"
    and steps: "(sharp_term shp s, sharp_term shp t) \<in> (rstep R)^*"
    and step: "(sharp_term shp t, sharp_term shp u) \<in> CS (DP shp R)" by auto
  from step obtain l r \<sigma> where lr: "(l,r) \<in> DP shp R" and tl: "sharp_term shp t = l \<cdot> \<sigma>" and ur: "sharp_term shp u = r \<cdot> \<sigma>" 
    by (induct, auto)
  show ?thesis
    by (rule exI[of _ l], rule exI[of _ r], rule exI[of _ u], rule exI[of _ \<sigma>], insert steps lr tl ur t u, auto)
qed


lemma finiteR_imp_finiteDP:
  assumes "finite R"
  shows "finite (DP shp R)"
proof -
 from assms have fS:
  "finite{(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R (h, length us)\<and> \<not>(l \<rhd> u))}"
  (is "finite ?S")
  by (rule finiteR_imp_finiteDP')
 let ?f = "\<lambda>(x,y,z). (sharp_term shp x, sharp_term shp z)"
 have eq1: "(\<Union>y\<in>?S. {x. x = ?f y}) = ?f ` ?S" by blast
 with fS have "finite(?f ` ?S)" by auto
 have "DP shp R = ?f ` ?S" (is "?DP = ?T")
 proof
  show "?DP \<subseteq> ?T"
  proof
   fix x assume "x \<in> ?DP"
   then obtain l r h us where "fst x = sharp_term shp l"
     and "snd x = sharp_term shp (Fun h us)" "(l,r) \<in> R" "r \<unrhd> Fun h us" "defined R (h, length us)" " \<not>(l \<rhd> Fun h us)" unfolding DP_def split_def by auto
   hence "(l,r,Fun h us) \<in> ?S" by auto
   thus "x \<in> ?T" unfolding eq1[symmetric]
   proof (rule UN_I)
    have "x = (fst x, snd x)" by simp
    hence "x = (sharp_term shp l, sharp_term shp (Fun h us))"
      unfolding `fst x = sharp_term shp l` `snd x = sharp_term shp (Fun h us)` .
    thus "x \<in> {x. x = ?f (l,r,Fun h us)}" by auto
   qed
  qed
 next
  show "?T \<subseteq> ?DP"
  proof
   fix x assume "x \<in> ?T"
   hence "x \<in> (\<Union>y\<in>?S. {x. x = ?f y})" unfolding eq1 .
   then obtain y where "y \<in> ?S" and "x = ?f y" by fast
   then obtain l r h us where "y = (l,r,Fun h us)" and dp: "(l,r) \<in> R" "r \<unrhd> Fun h us" "defined R (h, length us)" "\<not>(l \<rhd> Fun h us)" by blast
   with `x = ?f y` have "x = (sharp_term shp l, sharp_term shp (Fun h us))" by auto
   thus "x \<in> ?DP" using dp unfolding DP_def by auto
  qed
 qed
 with fS show ?thesis by auto
qed

lemma vars_sharp_eq_vars[simp]: "vars_term (sharp_term shp t) = vars_term t"
  by (induct t, auto)

lemma wf_trs_imp_wf_DP:
 assumes "wf_trs R"
 shows "wf_trs (DP shp R)"
unfolding wf_trs_def
proof (intro allI impI)
 fix s t
 assume "(s,t) \<in> DP shp R"
 then obtain l r h us where "s = sharp_term shp l" "t = sharp_term shp (Fun h us)" "(l,r) \<in> R" "r \<unrhd> (Fun h us)" "\<not>(l \<rhd> Fun h us)"
  unfolding DP_def by auto
 from `wf_trs R` and `(l,r) \<in> R` have "\<exists>f ss. l = Fun f ss"
  "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by auto
 from `\<exists>f ss. l = Fun f ss` obtain f ss where "l = Fun f ss" by auto
 hence "s = Fun (shp f) ss" unfolding `s = sharp_term shp l` by simp
 hence "\<exists>f ss. s = Fun f ss" by auto
 from `r \<unrhd> Fun h us` have "vars_term(Fun h us) \<le> vars_term r" by (induct rule: supteq.induct) auto
 hence "vars_term t \<subseteq> vars_term s" unfolding `s = sharp_term shp l`
   and `t = sharp_term shp (Fun h us)` vars_sharp_eq_vars using `vars_term r \<le> vars_term l` by simp
 from `\<exists>f ss. s = Fun f ss` `vars_term t \<subseteq> vars_term s`
 show "(\<exists>f ss. s = Fun f ss) \<and> vars_term t \<subseteq> vars_term s" by simp
qed

  
(* TODO: reformulate this lemma for map_term *)
(*
lemma plain_ctxt_term': assumes "(plain_term s) = (C\<langle>t\<rangle>)" shows "\<exists>D u. t = plain_term u \<and> C = plain_ctxt D"
using plain_ctxt_term[OF assms] by auto
*)


(* TODO: reform for map_term *)
(*
lemma plain_subst_term:
  assumes "plain_term s = t \<cdot> \<sigma>"
  shows "t = plain_term (unsharp_term t) \<and> t \<cdot> (\<sigma> |s (vars_term t)) = t \<cdot> (Subst(plain_term \<circ> unsharp_term \<circ> (case \<sigma> |s (vars_term t) of Subst s \<Rightarrow> s)))"
using assms proof (induct t arbitrary: s)
 case (Var x)
 obtain s' where "\<sigma> = Subst s'" by (cases \<sigma>) simp
 have "\<sigma> |s (vars_term(Var x)) = \<sigma> |s {x}" by simp
 have "Var x = plain_term (unsharp_term (Var x))" by simp
 show ?case
 proof (cases "x \<in> subst_domain \<sigma>")
 assume "x \<in> subst_domain \<sigma>"
 then obtain t where singleton:"\<sigma> |s {x} = Subst(\<lambda>z. if z = x then t else Var z)" (is "\<sigma> |s {x} = ?\<sigma> t") using restrict_singleton by best
 from `plain_term s = (Var x)\<cdot>\<sigma>` have "plain_term s = (Var x)\<cdot>(\<sigma> |s vars_term(Var x))" unfolding coincidence_lemma[where t = "Var x" and \<sigma> = \<sigma>] by simp
  hence "plain_term s = (Var x)\<cdot>(\<sigma> |s {x})" by simp
  hence "plain_term s = (Var x)\<cdot>(?\<sigma> t)" unfolding singleton by simp
  hence "plain_term s = t" by simp
  let ?\<tau> = "Subst(unsharp_term \<circ> (case \<sigma> |s {x} of Subst s \<Rightarrow> s))"
  have "\<sigma> |s vars_term(Var x) = ?\<sigma> t" unfolding `\<sigma> |s vars_term(Var x) = \<sigma> |s {x}` using singleton by simp
  also have "\<dots> = ?\<sigma> t" unfolding `\<sigma> |s {x} = ?\<sigma> t` by simp
  also have "\<dots> = ?\<sigma>(plain_term s)" unfolding `plain_term s = t` by simp
  also have "\<dots> = Subst(\<lambda>z. if z = x then plain_term s else plain_term (Var z))" by (simp cong: if_cong)
  also have "\<dots> = Subst(\<lambda>z. plain_term (if z = x then s else Var z))" unfolding if_distrib by simp
  also have "\<dots> = Subst(\<lambda>z. plain_term (unsharp_term (plain_term (if z = x then s else Var z))))" by simp
  also have "\<dots> = Subst(plain_term \<circ> unsharp_term \<circ> (\<lambda>z. plain_term (if z = x then s else Var z)))" unfolding o_def by simp
  also have "\<dots> = Subst(plain_term \<circ> unsharp_term \<circ> (\<lambda>z. if z = x then plain_term s else Var z))" unfolding if_distrib by (simp cong: if_cong)
  also have "\<dots> = Subst(plain_term \<circ> unsharp_term \<circ> (\<lambda>z. if z = x then t else Var z))" unfolding `plain_term s = t` by simp
  also have "\<dots> = Subst(plain_term \<circ> unsharp_term \<circ> (case Subst(\<lambda>z. if z = x then t else Var z) of Subst t \<Rightarrow> t))" unfolding subst_case_aux by simp
  also have "\<dots> = Subst(plain_term \<circ> unsharp_term \<circ> (case \<sigma> |s {x} of Subst t \<Rightarrow> t))" unfolding singleton by simp
  also have "\<dots> = Subst(plain_term \<circ> (case ?\<tau> of Subst t \<Rightarrow> t))" unfolding subst_case_aux unfolding o_assoc by simp
  finally have "\<sigma> |s vars_term(Var x) = Subst(plain_term \<circ> (case ?\<tau> of Subst t \<Rightarrow> t))" by simp
  hence "(Var x)\<cdot>(\<sigma> |s vars_term(Var x)) = (Var x)\<cdot>(Subst(plain_term \<circ> (case ?\<tau> of Subst t \<Rightarrow> t)))" unfolding `\<sigma> |s vars_term(Var x) = \<sigma> |s {x}` by simp
  with `Var x = plain_term (unsharp_term (Var x))` show ?thesis unfolding `\<sigma> |s vars_term(Var x) = \<sigma> |s {x}` unfolding subst_case_aux unfolding o_assoc by (rule conjI)
 next
  assume "x \<notin> subst_domain \<sigma>"
  have "\<sigma> |s vars_term(Var x) = \<sigma> |s {x}" by simp
  also have "\<dots> = Var" using `x \<notin> subst_domain \<sigma>` restrict_empty by best
  finally have "\<sigma> |s vars_term(Var x) = Subst(plain_term \<circ> unsharp_term \<circ> (case \<sigma> |s vars_term(Var x) of Subst t \<Rightarrow> t))" unfolding `\<sigma> |s vars_term(Var x) = \<sigma> |s {x}` unfolding `\<sigma> = Subst s'` unfolding o_def
    by (simp add: subst_empty_def)
  hence "(Var x)\<cdot>(\<sigma> |s vars_term(Var x)) = (Var x)\<cdot>(Subst(plain_term \<circ> unsharp_term \<circ> (case \<sigma> |s vars_term(Var x) of Subst t \<Rightarrow> t)))" by simp
  thus ?thesis using `Var x = plain_term (unsharp_term (Var x))` by simp
 qed
next
 case (Fun g ts)
 hence "plain_term s = Fun g (map (\<lambda>x. x\<cdot>\<sigma>) ts)" (is "_ = Fun g ?ts") by simp
 then obtain f ss where "s = Fun f ss" and "g = Plain f" by (induct s) auto
 hence "plain_term s = Fun (Plain f) (map plain_term ss)" (is "_ = Fun _ ?ss") by simp
 with `plain_term s = Fun g ?ts` have "?ss = ?ts" by simp
 hence "length ?ss = length ?ts" by simp
 hence "length ts = length ss" by simp
 obtain s' where "\<sigma> = Subst s'" by (cases \<sigma>) auto
 have ss:"\<forall>i<length ts. (map plain_term ss)!i = plain_term (ss!i)" unfolding `length ts = length ss` using nth_map by simp
 have ts:"\<forall>i<length ts. ?ts!i = (ts!i)\<cdot>\<sigma>" using nth_map by simp
 hence eq:"\<forall>i<length ts. plain_term (ss!i) = (ts!i)\<cdot>\<sigma>" using ss ts `?ss = ?ts` by simp
 have set:"\<forall>i<length ts. (ts!i) \<in> set ts" by simp
 have "length ts = length(map plain_term (map unsharp_term ts))" by simp
 have IH:"\<forall>i<length ts. (ts!i) = plain_term (unsharp_term (ts!i)) \<and>
          (ts!i)\<cdot>(\<sigma> |s vars_term(ts!i)) = (ts!i)\<cdot>(Subst(plain_term \<circ> unsharp_term \<circ> (case \<sigma> |s vars_term(ts!i) of Subst s \<Rightarrow> s)))" using Fun eq set by best
 hence "\<forall>i<length ts. (ts!i) = plain_term (unsharp_term (ts!i))" by simp
 hence "\<forall>i<length ts. (ts!i) = plain_term (map unsharp_term ts!i)" by simp
 hence "\<forall>i<length ts. (ts!i) = (map plain_term (map unsharp_term ts))!i" by simp
 hence "ts = map plain_term (map unsharp_term ts)" using list_eq_iff_nth_eq[where xs = ts and ys = "map plain_term (map unsharp_term ts)"] `length ts = length(map plain_term (map unsharp_term ts))` by best
 have "plain_term (unsharp_term (Fun g ts)) = Fun g (map plain_term (map unsharp_term ts))" unfolding `g = Plain f` by simp
 also have "\<dots> = Fun g ts" unfolding `ts = map plain_term (map unsharp_term ts)`[symmetric] by simp
 finally have one:"Fun g ts = plain_term (unsharp_term (Fun g ts))" by simp
 have "\<forall>i < length ts. vars_term(ts!i) \<subseteq> \<Union>set(map vars_term ts)"
 proof (induct ts)
  case Nil thus ?case by simp
 next
  case (Cons t ts)
  have zero:"vars_term((t#ts)!0) \<subseteq> \<Union>set(map vars_term (t#ts))" by auto
  show ?case
  proof
   fix i have "i = (0::nat) \<or> i \<noteq> 0" by auto
   thus "i < length(t#ts) \<longrightarrow> vars_term((t#ts)!i) \<subseteq> \<Union>set(map vars_term (t#ts))"
   proof
    assume "i = 0" thus ?thesis using zero by auto
   next
    assume "i \<noteq> 0" show ?thesis
    proof
     assume "i < length(t#ts)"
     with `i \<noteq> 0` have "((t#ts)!i) \<in> set ts" by (induct i) auto
     with Cons show "vars_term((t#ts)!i) \<subseteq> \<Union>set(map vars_term(t#ts))" by auto
    qed
   qed
  qed
 qed
 hence a:"\<And>i. i < length ts \<Longrightarrow> vars_term(ts!i) \<subseteq> vars_term(Fun g ts)" by simp
 hence b:"!!i. i < length ts \<Longrightarrow> (ts!i)\<cdot>(\<sigma> |s vars_term(ts!i)) = (ts!i)\<cdot>(\<sigma> |s vars_term(Fun g ts))" using coincidence_lemma' by best
 let ?un = "\<lambda>\<sigma>. Subst(unsharp_term \<circ> (case \<sigma> of Subst s \<Rightarrow> s))"
 let ?pl = "\<lambda>\<sigma>. Subst(plain_term \<circ> (case \<sigma> of Subst s \<Rightarrow> s))"
 from IH have A:"!!i. i < length ts \<Longrightarrow> (ts!i)\<cdot>(\<sigma> |s vars_term(ts!i)) = (ts!i)\<cdot>(?pl (?un (\<sigma> |s vars_term(ts!i))))" unfolding subst_case_aux unfolding o_assoc by simp
 have "Var x = unsharp_term (Var x)" by simp
 have c:"!!i. i < length ts \<Longrightarrow> (ts!i)\<cdot>(?pl (?un (\<sigma> |s vars_term(ts!i)))) = (ts!i)\<cdot>(?pl (?un (\<sigma> |s vars_term(Fun g ts))))" 
 proof -
  fix i assume "i < length ts"
  have "(ts!i)\<cdot>(?pl (?un (\<sigma> |s vars_term(ts!i)))) = (ts!i)\<cdot>(?pl ((?un \<sigma>) |s vars_term(ts!i)))" unfolding subst_restrict_f[OF map_term.simps(1)] by simp
  also have "\<dots> = (ts!i)\<cdot>((?pl (?un \<sigma>)) |s vars_term(ts!i))" unfolding subst_restrict_f[OF map_term.simps(1)] by simp
  also have "\<dots> = (ts!i)\<cdot>((?pl (?un \<sigma>)) |s vars_term(Fun g ts))" using a[OF `i < length ts`] coincidence_lemma'[where t = "ts!i" and V = "vars_term(Fun g ts)"] by best
  also have "\<dots> = (ts!i)\<cdot>(?pl ((?un \<sigma>) |s vars_term(Fun g ts)))" unfolding subst_restrict_f[OF map_term.simps(1)] by simp
  finally show "(ts!i)\<cdot>(?pl (?un (\<sigma> |s vars_term(ts!i)))) = (ts!i)\<cdot>(?pl (?un (\<sigma> |s vars_term(Fun g ts))))" unfolding subst_restrict_f[OF map_term.simps(1)] by simp
 qed
 from A b c have "\<forall>i < length ts. (ts!i)\<cdot>(\<sigma> |s (vars_term(Fun g ts))) = (ts!i)\<cdot>(?pl (?un (\<sigma> |s vars_term(Fun g ts))))" by simp
 hence d: "\<forall>i < length ts. map (\<lambda>x. x\<cdot>(\<sigma> |s (vars_term(Fun g ts)))) ts!i = map (\<lambda>x. x\<cdot>(?pl(?un(\<sigma> |s vars_term(Fun g ts))))) ts!i" by simp
 have e: "length(map (\<lambda>x. x\<cdot>(?pl (?un (\<sigma> |s vars_term(Fun g ts))))) ts) = length ts" by simp
 from d and e have f:"map (\<lambda>x. x\<cdot>(\<sigma> |s vars_term(Fun g ts))) ts = map (\<lambda>x. x\<cdot>(?pl(?un(\<sigma> |s vars_term(Fun g ts))))) ts"
  using list_eq_iff_nth_eq[where xs = "map (\<lambda>x. x\<cdot>(\<sigma> |s vars_term(Fun g ts))) ts"
                             and ys = "map (\<lambda>x. x\<cdot>(?pl(?un(\<sigma> |s vars_term(Fun g ts))))) ts"] by simp
 have "Fun g ts\<cdot>(\<sigma> |s vars_term(Fun g ts)) = Fun g (map (\<lambda>x. x\<cdot>(\<sigma> |s vars_term(Fun g ts))) ts)" by simp
 also have "\<dots> = Fun g (map (\<lambda>x. x\<cdot>(?pl(?un(\<sigma> |s vars_term(Fun g ts))))) ts)" unfolding f by simp
 finally have two:"Fun g ts\<cdot>(\<sigma> |s vars_term(Fun g ts)) = Fun g ts\<cdot>(?pl(?un(\<sigma> |s vars_term(Fun g ts))))" by simp
 from one and two show ?case unfolding subst_case_aux unfolding o_assoc by fast
qed
 
(* TODO: reform with map_term *)
lemma plain_subst_term':
 assumes "plain_term s = t\<cdot>\<sigma>"
 shows "\<exists>u \<tau>. t = plain_term u \<and> t\<cdot>(\<sigma> |s vars_term t) = t\<cdot>(Subst(plain_term \<circ> \<tau>))"
proof -
 from plain_subst_term[OF assms] have "t = plain_term (unsharp_term t)"
  and "t\<cdot>(\<sigma> |s vars_term t) = t\<cdot>(Subst(plain_term \<circ> (unsharp_term \<circ> (case \<sigma> |s vars_term t of Subst s \<Rightarrow> s))))" unfolding o_assoc by auto
 hence "\<exists>\<tau>. t = plain_term (unsharp_term t) \<and> t\<cdot>(\<sigma> |s vars_term t) = t\<cdot>(Subst(plain_term \<circ> \<tau>))" using exI[where x = "unsharp_term \<circ> (case \<sigma> |s vars_term t of Subst s \<Rightarrow> s)"] by auto
 thus ?thesis by auto
qed
*)


(*
(* TODO: reform with map_term *)
lemma plain_rstep_imp_plain_rhs:
 assumes "wf_trs R" and "(plain_term s,t) \<in> rstep(ID R)"
 shows "\<exists>u. t = plain_term u"
proof -
 from assms obtain l r p \<sigma> where "(plain_term s,t) \<in> rstep_r_p_s (ID R) (l,r) p \<sigma>"
  unfolding rstep_iff_rstep_r_p_s by best
 hence "(l,r) \<in> ID R" and "p \<in> poss(plain_term s)"
  and "(ctxt_of_pos_term p (plain_term s))\<langle>l\<cdot>\<sigma>\<rangle> = (plain_term s)" (is "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s")
  and "(ctxt_of_pos_term p (plain_term s))\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
 from `(l,r) \<in> ID R` obtain l' r' where "(l',r') \<in> R" and "l = plain_term l'" and "r = plain_term r'" unfolding map_trs.simps by auto
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = plain_term s` have "plain_term s = (?C\<langle>l\<cdot>\<sigma>\<rangle>)" by simp
 then obtain D u where "l\<cdot>\<sigma> = plain_term u" and "?C = plain_ctxt D" using plain_ctxt_term by best
 hence "plain_term u = l\<cdot>\<sigma>" by simp
 let ?un = "\<lambda>\<sigma>. Subst(unsharp_term \<circ> (case \<sigma> of Subst s \<Rightarrow> s))"
 let ?pl = "\<lambda>\<sigma>. Subst(plain_term \<circ> (case \<sigma> of Subst s \<Rightarrow> s))"
 obtain \<sigma>' where "\<forall>s. s\<cdot>\<sigma> = s\<cdot>\<sigma>'" by auto
 hence "plain_term u = l\<cdot>\<sigma>'" unfolding `plain_term u = l\<cdot>\<sigma>` by simp
 have "l = plain_term (unsharp_term l)" (is "l = plain_term ?l") and "l\<cdot>(\<sigma>' |s vars_term l) = l\<cdot>(?pl(?un(\<sigma>' |s vars_term l)))" (is "_ = l\<cdot>?\<tau>") using plain_subst_term[where \<sigma> = "\<sigma>'" and s = u and t = l, OF `plain_term u = l\<cdot>\<sigma>'`] unfolding o_assoc[symmetric] unfolding subst_case_aux by auto
 have "plain_term (?l\<cdot>(?un(\<sigma>' |s vars_term l))) = (plain_term ?l)\<cdot>(?pl(?un(\<sigma>' |s vars_term l)))" unfolding map_subst_distrib by (simp add: o_def)
 also have "\<dots> = l\<cdot>?\<tau>" unfolding `l = plain_term ?l`[symmetric] by simp
 also have "\<dots> = l\<cdot>(\<sigma>' |s vars_term l)" unfolding `l\<cdot>(\<sigma>' |s vars_term l) = l\<cdot>?\<tau>` by simp
 also have "\<dots> = l\<cdot>\<sigma>'" unfolding coincidence_lemma[symmetric] by simp
 finally have "plain_term (?l\<cdot>(?un(\<sigma>' |s vars_term l))) = l\<cdot>\<sigma>" unfolding `plain_term u = l\<cdot>\<sigma>`[symmetric] using `plain_term u = l\<cdot>\<sigma>'` by simp
 from `\<forall>s. s\<cdot>\<sigma> = s\<cdot>\<sigma>'` have "r\<cdot>\<sigma> = r\<cdot>\<sigma>'" by simp
 hence "r\<cdot>\<sigma>' = r\<cdot>(\<sigma>' |s vars_term r)" using coincidence_lemma[where t = r] by simp
 from `wf_trs R` have "wf_trs (ID R)" by simp
 with `(l,r) \<in> ID R` have "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by simp
 with `r\<cdot>\<sigma>' = r\<cdot>(\<sigma>' |s vars_term r)` have "r\<cdot>\<sigma> = r\<cdot>(\<sigma>' |s vars_term l)" unfolding `r\<cdot>\<sigma> = r\<cdot>\<sigma>'` using coincidence_lemma' by auto
 have A:"?C\<langle>r\<cdot>(\<sigma>' |s vars_term l)\<rangle> = t" using `?C\<langle>r\<cdot>\<sigma>\<rangle> = t` unfolding `r\<cdot>\<sigma> = r\<cdot>(\<sigma>' |s vars_term l)` by simp
 with `l\<cdot>(\<sigma>' |s vars_term l) = l\<cdot>(?pl(?un(\<sigma>' |s vars_term l)))` and `vars_term r \<subseteq> vars_term l` have "r\<cdot>(\<sigma>' |s vars_term l) = r\<cdot>(?pl(?un(\<sigma>' |s vars_term l)))" using subst_dom by auto
 also have "\<dots> = r\<cdot>(?pl((?un \<sigma>') |s vars_term l))"
  unfolding subst_restrict_f[OF map_term.simps(1)] by simp
 also have "\<dots> = r\<cdot>((?pl(?un \<sigma>')) |s vars_term l)" unfolding subst_restrict_f[OF map_term.simps(1)] by simp
 also have "\<dots> = r\<cdot>((?pl(?un \<sigma>')) |s vars_term r)" unfolding coincidence_lemma'[where t = r and V = "vars_term l", OF `vars_term r \<subseteq> vars_term l`] by simp
 also have "\<dots> = r\<cdot>(?pl((?un \<sigma>') |s vars_term r))" unfolding subst_restrict_f[OF map_term.simps(1)] by simp
 finally have "r\<cdot>(\<sigma>' |s vars_term l) = r\<cdot>(?pl(?un(\<sigma>' |s vars_term r)))" unfolding subst_restrict_f[OF map_term.simps(1)] by simp
 with A have "?C\<langle>r\<cdot>(?pl(?un(\<sigma>' |s vars_term r)))\<rangle> = t" by simp
 hence "?C\<langle>plain_term r'\<cdot>(?pl(?un(\<sigma>' |s vars_term r)))\<rangle> = t" unfolding `r = plain_term r'` by simp
 hence "?C\<langle>plain_term (r'\<cdot>(?un(\<sigma>' |s vars_term r)))\<rangle> = t" using map_subst_distrib[where t = r' and \<sigma> = "?un(\<sigma>' |s vars_term r)"] by (simp add: o_def)
 hence "(plain_ctxt D)\<langle>plain_term (r'\<cdot>(?un(\<sigma>' |s vars_term r)))\<rangle> = t" unfolding `?C = plain_ctxt D` by simp
 hence "plain_term (D\<langle>r'\<cdot>(?un(\<sigma>' |s vars_term r))\<rangle>) = t" unfolding map_ctxt_distrib by simp
 thus ?thesis by best
qed

(* TODO reform to map_term and inj fg *)
lemma plain_rstep_pow_imp_rstep_pow: 
 assumes "wf_trs R"
 shows "(plain_term s,plain_term t) \<in> rstep(ID R)^^n \<Longrightarrow> (s,t) \<in> (rstep R ^^n)"
proof (induct n arbitrary: s t)
 case 0 thus ?case using map_term_inj[OF Plain_inj] unfolding inj_on_def by auto
next
 case (Suc n s t)
 from rel_pow_Suc_D2[OF this(2)]
   obtain u where "(plain_term s,u) \<in> rstep(ID R)" and "(u,plain_term t) \<in> (rstep (ID R) ^^ n)" by best
 from `wf_trs R` and `(plain_term s,u) \<in> rstep(ID R)` obtain v where "u = plain_term v" using plain_rstep_imp_plain_rhs by best
 with `(u,plain_term t) \<in> (rstep (ID R) ^^ n)` have "(plain_term v,plain_term t) \<in> (rstep (ID R) ^^ n)" by simp
 with Suc have "(v,t) \<in> (rstep R ^^ n)" by simp
 from `(plain_term s,u) \<in> rstep(ID R)` have "(plain_term s,plain_term v) \<in> rstep(ID R)" unfolding `u = plain_term v` by simp
 with rstep_iff_map_rstep[OF Plain_inj] have "(s,v) \<in> rstep R" by best
 with `(v,t) \<in> (rstep R ^^ n)` show ?case using rel_pow_Suc_I2[where R = "rstep R" and x = s] by simp
qed
  
(* TODO reform to map_term *)
lemma plain_rsteps_imp_rsteps:
 assumes "wf_trs R"
 shows "(plain_term s,plain_term t) \<in> (rstep(ID R))^* \<Longrightarrow> (s,t) \<in> (rstep R)^*"
proof -
 assume "(plain_term s,plain_term t) \<in> (rstep (ID R))^*"
 then obtain n where "(plain_term s,plain_term t) \<in> (rstep (ID R) ^^ n)" using rtrancl_imp_rel_pow' by best
 hence "(s,t) \<in> (rstep R ^^ n)" using plain_rstep_pow_imp_rstep_pow[OF assms] by simp
 thus "(s,t) \<in> (rstep R)^*" using rel_pow_imp_rtrancl by best
qed
*)

lemma sharp_eq_imp_eq: assumes "inj shp" shows "sharp_term shp s = sharp_term shp t \<Longrightarrow> s = t"
proof (cases s)
  case (Var x) 
  assume "sharp_term shp s = sharp_term shp t" with Var show ?thesis by (induct t) auto
next
  case (Fun f ss)
  assume "sharp_term shp s = sharp_term shp t" 
  with Fun have "sharp_term shp (Fun f ss) = sharp_term shp t" by simp
  then obtain g ts where t: "t = Fun g ts" by (induct t) auto
  with `sharp_term shp s = sharp_term shp t` have "Fun (shp f) ss = Fun (shp g) ts" unfolding `s = Fun f ss` `t = Fun g ts`
    by simp
  hence "f = g" and "ss = ts" using `inj shp`[unfolded inj_on_def] by auto
  thus ?thesis unfolding Fun t by simp
qed

lemma SN_elt_imp_SN_elt_sharp:
 assumes SN: "SN_elt (rstep R) s" and s: "s = Fun f ss" and ndef: "\<not> defined R (shp f, length ss)"
 shows "SN_elt (rstep R) (sharp_term shp s)" 
  unfolding s sharp_term.simps
proof (rule SN_args_imp_SN)
  fix t
  assume "t \<in> set ss" thus "SN_elt (rstep R) t" using SN_imp_SN_arg[OF SN[unfolded s]]  by auto
next
  {
    fix x r
    assume "(Var x,r) \<in> R"
    from left_var_imp_not_SN[OF this] SN have False ..
  }
  thus "\<forall> lr \<in> R. \<forall> x. fst lr \<noteq> Var x" by auto
next
  show "(shp f, length ss) \<notin> defs R" using ndef unfolding defs_def by simp
qed

(* TODO delete?
lemma sharp_nrrstep_imp_rstep:
 assumes "wf_trs R" and "\<not> defined R (shp (the (root s)),num_args s)" and "(sharp_term shp s,sharp_term shp t) \<in> nrrstep R"
 shows "(s,t) \<in> rstep R"
proof -
 from assms obtain f ss where "sharp_term s = Fun f ss" using nrrstep_imp_Fun by best
 with assms have "(Fun f ss,sharp_term t) \<in> nrrstep(ID R)" by simp
 then obtain ts where "sharp_term t = Fun f ts" using nrrstep_preserves_root by best
 from `sharp_term s = Fun f ss` obtain f' ss' where "sharp_term s = sharp_term (Fun f' ss')" by (induct s) auto
 from `sharp_term t = Fun f ts` obtain g' ts' where "sharp_term t = sharp_term (Fun g' ts')" by (induct t) auto
 have "root(sharp_term s) = root(sharp_term t)" unfolding `sharp_term s = Fun f ss` `sharp_term t = Fun f ts` by simp
 hence "Sharp f' = Sharp g'"
   unfolding `sharp_term s = sharp_term (Fun f' ss')` `sharp_term t = sharp_term (Fun g' ts')` by simp
 hence "f' = g'" by simp
 from `(sharp_term s,sharp_term t) \<in> nrrstep(ID R)`
  have a:"(Fun (Sharp f') (map plain_term ss'),Fun (Sharp g') (map plain_term ts')) \<in> nrrstep(ID R)"
  unfolding `sharp_term s = sharp_term (Fun f' ss')` and `sharp_term t = sharp_term (Fun g' ts')` by simp
 hence "length(map plain_term ss') = length(map plain_term ts')" unfolding `f' = g'`
  using nrrstep_Fun_preserves_num_args[where ss = "map plain_term ss'"] by auto
 hence "length ss' = length ts'" by simp
 from a have "(Fun (Sharp f') (map plain_term ss'), Fun (Sharp f') (map plain_term ts')) \<in> nrrstep(ID R)"
  unfolding `Sharp f' = Sharp g'` by auto
 then obtain C i where "i < length(map plain_term ss')"
  and b:"(map plain_term ss'!i,map plain_term ts'!i) \<in> rstep(ID R)"
  and c:"C\<langle>map plain_term ss'!i\<rangle> = Fun (Sharp f') (map plain_term ss')"
  and d:"C\<langle>map plain_term ts'!i\<rangle> = Fun (Sharp g') (map plain_term ts')"
  unfolding `Sharp f' = Sharp g'`
  using nrrstep_Fun_imp_arg_rstep[where f = "Sharp g'" and ss = "map plain_term ss'"] 
  by best
 from `i < length(map plain_term ss')` have "map plain_term ss'!i = plain_term (ss'!i)" by simp
 from `i < length(map plain_term ss')` have "map plain_term ts'!i = plain_term (ts'!i)" unfolding `length(map plain_term ss') = length(map plain_term ts')` by simp
 from b have "(plain_term (ss'!i),plain_term (ts'!i)) \<in> rstep(ID R)"
  unfolding `map plain_term ss'!i = plain_term (ss'!i)` `map plain_term ts'!i = plain_term (ts'!i)` by simp
 hence "(ss'!i,ts'!i) \<in> rstep R" unfolding rstep_iff_map_rstep[OF Plain_inj, symmetric] by simp
 from `sharp_term s = sharp_term (Fun f' ss')` have "s = Fun f' ss'" by (rule sharp_eq_imp_eq)
 from `sharp_term t = sharp_term (Fun g' ts')` have "t = Fun g' ts'" by (rule sharp_eq_imp_eq)
 let ?C = "unsharp_ctxt C"
 have "?C\<langle>ss'!i\<rangle> = ?C\<langle>unsharp_term (plain_term (ss'!i))\<rangle>" by simp
 also have "\<dots> = unsharp_term (C\<langle>plain_term (ss'!i)\<rangle>)" unfolding map_ctxt_distrib by simp
 also have "\<dots> = unsharp_term (Fun (Sharp f') (map plain_term ss'))"
  using c unfolding `map plain_term ss'!i = plain_term (ss'!i)` by simp
 also have "\<dots> = Fun f' (map unsharp_term (map plain_term ss'))" by simp
 finally have e:"?C\<langle>ss'!i\<rangle> = Fun f' ss'" unfolding unsharp_plain by simp
 have "?C\<langle>ts'!i\<rangle> = ?C\<langle>unsharp_term (plain_term (ts'!i))\<rangle>" by simp
 also have "\<dots> = unsharp_term (C\<langle>plain_term (ts'!i)\<rangle>)" unfolding map_ctxt_distrib by simp
 also have "\<dots> = unsharp_term (Fun (Sharp g') (map plain_term ts'))"
  using d unfolding `map plain_term ts'!i = plain_term (ts'!i)` by simp
 also have "\<dots> = Fun g' (map unsharp_term (map plain_term ts'))" by simp
 finally have f:"?C\<langle>ts'!i\<rangle> = Fun g' ts'" unfolding unsharp_plain by simp
 from `(ss'!i,ts'!i) \<in> rstep R` have "(?C\<langle>ss'!i\<rangle>,?C\<langle>ts'!i\<rangle>) \<in> CC(rstep R)" by (rule ctxt_closureI)
 hence "(?C\<langle>ss'!i\<rangle>,?C\<langle>ts'!i\<rangle>) \<in> rstep R" by auto
 hence "(Fun f' ss',Fun g' ts') \<in> rstep R" unfolding e f by simp
 thus ?thesis unfolding `s = Fun f' ss'` `t = Fun g' ts'` by simp
qed
*)

lemma DP_step_in_R:
  fixes R::"('f,'v)trs" and v::"('f,'v)term \<Rightarrow> 'v" and shp :: "'f \<Rightarrow> 'f"
  assumes "(s,t) \<in> DP shp R" and inj: "inj shp" shows "\<exists>C. C \<in> wf_ctxts (funas_trs R) \<and> (sharp_term (the_inv shp) s,C\<langle>sharp_term (the_inv shp) t\<rangle>) \<in> R"
proof -
  let ?us = "sharp_term (the_inv shp)"
  let ?s = "sharp_term shp"
  from assms obtain l r f ts
    where s: "s = sharp_term shp l" and t: "t = sharp_term shp (Fun f ts)"
    and R: "(l,r) \<in> R" and sub: "r \<unrhd> Fun f ts" unfolding DP_def supt_supteq_conv by auto
  from sub obtain C where r: "r = C\<langle>Fun f ts\<rangle>" by auto
  have subset: "funas_trs R \<subseteq> funas_trs R" by simp
  from rhs_wf[OF R subset] have "r \<in> wf_terms (funas_trs R)" .
  hence "C\<langle>Fun f ts\<rangle> \<in> wf_terms (funas_trs R)" unfolding r .
  from ctxt_apply_wf_imp_ctxt_and_term_wf[OF this]
    have "C \<in> wf_ctxts (funas_trs R)" and "Fun f ts \<in> wf_terms (funas_trs R)" by auto
  from lhs_wf[OF R subset] have "l \<in> wf_terms (funas_trs R)" .
  have us: "?us s = l" unfolding s by (cases l, auto simp: the_inv_f_f[OF inj])
  have ut: "?us t = Fun f ts" unfolding t by (simp add: the_inv_f_f[OF inj])
  from R have "(?us s,C\<langle>?us t\<rangle>) \<in> R" unfolding us ut r .
  with `C \<in> wf_ctxts (funas_trs R)` show ?thesis by best
qed

lemma sharp_rrstep_imp_rstep:
 assumes rrstep: "(sharp_term shp s,sharp_term shp t) \<in> CS(DP shp R)" and "inj shp" and "wf_trs R" shows "\<exists>C. (s,C\<langle>t\<rangle>) \<in> rstep R"
proof -
 from `wf_trs R` have "wf_trs (DP shp R)" by (rule wf_trs_imp_wf_DP)
 from rrstep obtain l r \<sigma> where "(l,r) \<in> DP shp R" and ss: "sharp_term shp s = l\<cdot>\<sigma>" and st: "sharp_term shp t = r\<cdot>\<sigma>"
   by (induct, auto)
 from `(l,r) \<in> DP shp R` obtain l' r' h' us' where l: "l = sharp_term shp l'" and r: "r = sharp_term shp (Fun h' us')" (is "r = sharp_term shp ?u'")
  and l'r': "(l',r') \<in> R" and "r' \<unrhd> (Fun h' us')" and "\<not> l' \<rhd> Fun h' us'"
  unfolding DP_def by auto
 from `wf_trs R` and `(l',r') \<in> R` obtain f' ss' where l': "l' = Fun f' ss'" using wf_trs_imp_lhs_Fun by best
 from `r' \<unrhd> ?u'` obtain C where r': "r' = C\<langle>?u'\<rangle>" by best
 have ss: "sharp_term shp s = (Fun (shp f') ss') \<cdot> \<sigma>" unfolding ss l l' by simp
 then obtain f where s: "s = (Fun f ss') \<cdot> \<sigma>" by (cases s, auto)
 from s ss have "shp f = shp f'" by simp
 with `inj shp`[unfolded inj_on_def] have f: "f = f'" by simp
 have ts: "sharp_term shp t = (Fun (shp h') us') \<cdot> \<sigma>" unfolding st r by simp
 then obtain h where t: "t = (Fun h us') \<cdot> \<sigma>" by (cases t, auto)
 from t ts have "shp h = shp h'" by simp
 with `inj shp`[unfolded inj_on_def] have h: "h = h'" by simp
 show ?thesis
   by (rule exI[of _ "C \<cdot>c \<sigma>"], unfold s t f h, rule rstepI[OF l'r', of _ \<box> \<sigma>], unfold l' r', simp, simp) 
qed

(*
lemma rstepID_imp_rstep:
 assumes a:"wf_trs R" and b: "(sharp_term shp s,sharp_term shp t) \<in> rstep R"
 shows "(s,t) \<in> rstep R"
proof -
 from b and a have "(sharp_term shp s,sharp_term shp t) \<in> nrrstep R" by (rule rstepID_imp_nrrstepID)
 with c show ?thesis by (rule sharp_nrrstep_imp_rstep)
qed
*)


(* 
lemma rstepID_imp_plain_rstep:
 assumes "(s,t) \<in> rstep(ID R)"
 shows "(plain_term (unsharp_term s),plain_term (unsharp_term t)) \<in> rstep(ID R)"
proof -
  from assms have "(unsharp_term s, unsharp_term t) \<in> rstep (map_trs unsharp_fun (ID R))" (is "(?us,?ut) \<in> rstep ?uR")
    by (rule rstep_imp_map_rstep)
  hence "(plain_term ?us, plain_term ?ut) \<in> rstep (map_trs Plain ?uR)" (is "_ \<in> rstep ?PuR")
    by (rule rstep_imp_map_rstep)
  moreover have "?PuR = ID R"
    by (simp add: map_trs_comp o_def)
  ultimately show ?thesis by simp
qed
*)

(*
lemma SN_elt_rstep_iff_SN_elt_rstepID:
 fixes R :: "('f,'v)trs" and s :: "('f,'v)term"
 shows "(SN_elt (rstep R) s::bool) = (SN_elt (rstep(ID R)) (plain_term s)::bool)"
proof
 assume "SN_elt (rstep R) s"
 show "SN_elt (rstep (ID R)) (plain_term s)"
 proof (rule ccontr)
  assume "\<not>SN_elt (rstep(ID R)) (plain_term s)"
  then obtain S where start: "S 0 = plain_term s" and seq: "\<forall>i. (S i,S(Suc i)) \<in> rstep(ID R)"
   unfolding SN_elt_def by auto
  let ?S = "\<lambda>i. plain_term (unsharp_term (S i))"
  let ?T = "\<lambda>i. unsharp_term (S i)"
  from seq have "\<forall>i. (?S i,?S(Suc i)) \<in> rstep(ID R)" using rstepID_imp_plain_rstep by force
  hence seq': "\<forall>i. (?T i,?T(Suc i)) \<in> rstep R" unfolding rstep_iff_map_rstep[OF Plain_inj, symmetric] by simp
  hence "?T 0 = s" unfolding `S 0 = plain_term s` by simp
  with seq' have "\<not>(SN_elt (rstep R) s)" unfolding SN_elt_def by auto
  with `SN_elt (rstep R) s` show False by simp
 qed
qed (rule SN_map_imp_SN)
*)

(*
lemma supt_sharp[simp]: assumes "t \<rhd> s" shows "sharp_term t \<rhd> plain_term s"
using assms proof (cases rule: supt.cases)
  case (arg ss f)
  hence "plain_term s \<in> set(map plain_term ss)" by simp
  thus ?thesis unfolding `t = Fun f ss` unfolding sharp_term.simps by (rule supt.arg)
next
  case (subt u ss f)
  hence "plain_term u \<in> set(map plain_term ss)" and "plain_term u \<rhd> plain_term s"
  using supt_map_term by auto
  thus ?thesis unfolding subt sharp_term.simps by (rule supt.subt)
qed

lemma map_unsharp_plain_id[simp]: "map (unsharp_term \<circ> plain_term) ss = ss" by (simp add: o_def)

lemma unsharp_sharp_id[simp]: "unsharp_term (sharp_term s) = s"
proof (cases s)
 case (Var x) thus ?thesis unfolding Var by simp
next
 case (Fun f ss) thus ?thesis unfolding Fun by simp
qed
*)

(*
lemma TinfID_imp_Tinf: assumes "sharp_term t \<in> Tinf (ID R)" shows "t \<in> Tinf R"
proof -
  from assms have supt:"\<forall>s. (sharp_term t) \<rhd> s \<longrightarrow> SN_elt (rstep(ID R)) s"
    and "\<not>SN_elt (rstep(ID R)) (sharp_term t)" unfolding Tinf_def by auto
  have "\<forall>s. t \<rhd> s \<longrightarrow> SN_elt (rstep R) s"
  proof (intro allI impI)
    fix s assume "t \<rhd> s"
    hence "sharp_term t \<rhd> plain_term s" by (rule supt_sharp)
    with supt have "SN_elt (rstep(ID R)) (plain_term s)" by simp
    thus "SN_elt (rstep R) s" unfolding SN_elt_rstep_iff_SN_elt_rstepID[symmetric] by simp
  qed
  from `\<not>(SN_elt (rstep(ID R)) (sharp_term t))` have "\<not>(SN_elt (rstep R) t)"
    using SN_elt_imp_SN_elt_sharp by auto
  with `\<forall>s. t \<rhd> s \<longrightarrow> SN_elt (rstep R) s` show ?thesis unfolding Tinf_def by simp
qed


lemma supt_plain'[intro 2]: assumes "plain_term t \<rhd> plain_term s" shows "t \<rhd> s" using supt_map_term[OF assms, of unsharp_fun]
  by simp

lemma supt_sharp'[intro 2]: assumes "sharp_term t \<rhd> plain_term s" shows "t \<rhd> s" using supt_map_term[OF assms, of unsharp_fun]
  by simp

lemma supt_plain_conv[iff]: "(plain_term s \<rhd> plain_term t) = (s \<rhd> t)"
using iffI[OF supt_plain' supt_map_term] by (simp add: supt_supteq_conv)

lemma supt_sharp_conv[iff]: "(sharp_term s \<rhd> plain_term t) = (s \<rhd> t)"
using iffI[OF supt_sharp' supt_sharp] by (simp add: supt_supteq_conv)
*)
 
lemma reverseTrs[iff]: "(rstep R)^-1 = rstep (R^-1)"
proof - 
  have oneDir: "\<And>R. (rstep R)^-1 \<subseteq> rstep (R^-1)" 
  proof
    fix R st
    show "(st \<in> (rstep R)^-1) \<Longrightarrow> (st \<in> rstep (R^-1))"
    proof (cases st)
      case (Pair s t)
      assume "st \<in> (rstep R)^-1"
      with Pair have "(t,s) \<in> rstep R" by simp
      hence "(s,t) \<in> rstep (R^-1)" 
      proof
        fix l r C \<sigma> assume in_R: "(l, r) \<in> R" and t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>" and s: "s = C\<langle>r\<cdot>\<sigma>\<rangle>"
        from in_R show "(s,t) \<in> rstep (R^-1)" unfolding s t by auto
      qed
      with Pair show ?thesis by simp
    qed
  qed
  hence one: "(rstep R)^-1 \<subseteq> rstep (R^-1)" by blast
  from oneDir have "(rstep (R^-1))^-1 \<subseteq> rstep ((R^-1) ^-1)" by blast
  with one show ?thesis by force
qed

(*
lemma no_sharp_rrstep:
 fixes R :: "('f,'v)trs"
 assumes "wf_trs R" shows "(sharp_term s,t) \<notin> rrstep(ID R)"
proof
 assume "(sharp_term s,t) \<in> rrstep(ID R)"
 then obtain l r :: "('f shp,'v)term" and \<sigma> where step: "(sharp_term s,t) \<in> rstep_r_p_s (ID R) (l,r) \<epsilon> \<sigma>"
  unfolding rrstep_def by auto
 let ?C = "ctxt_of_pos_term \<epsilon> (sharp_term s)"
 have "?C = \<box>" by simp
 with step have "(l,r) \<in> ID R" and "l\<cdot>\<sigma> = sharp_term s" and "r\<cdot>\<sigma> = t" unfolding rstep_r_p_s_def Let_def by auto
 from `(l,r) \<in> ID R` obtain u v where "(u,v) \<in> R" and "l = plain_term u" unfolding map_trs.simps by auto
 from assms have "wf_trs (ID R)" by simp
 from `(l,r) \<in> ID R` and `wf_trs (ID R)` obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
 from `(u,v) \<in> R` and `wf_trs R` obtain g ts where "u = Fun g ts" unfolding wf_trs_def by best
 from `l\<cdot>\<sigma> = sharp_term s` obtain h us where "s = Fun h us"
   unfolding `l = Fun f ss` subst_apply_term.simps by (induct s) auto
 from `l\<cdot>\<sigma> = sharp_term s` have "f = Sharp h" unfolding `l = Fun f ss` `s = Fun h us` by simp
 hence "l = Fun (Sharp h) ss" using `l = Fun f ss` by simp
 with `l = plain_term u` show False unfolding `u = Fun g ts` by simp
qed

lemma poss_unsharp_conv[iff]: "(p \<in> poss(unsharp_term s)) = (p \<in> poss s)"
proof (induct s arbitrary: p)
 case (Var x) thus ?case by simp
next
 case (Fun f ss) thus ?case by (cases p) auto
qed
*)

lemma ctxt_of_pos_term_map_term_conv[iff]:
 assumes "p \<in> poss s"
 shows "map_ctxt fg (ctxt_of_pos_term p s) = (ctxt_of_pos_term p (map_term fg s))"
using assms proof (induct s arbitrary: p)
 case (Var x) thus ?case by simp
next
 case (Fun f ss) thus ?case
 proof (cases p)
  case Empty thus ?thesis by simp
 next
  case (PCons i q)
  with `p \<in> poss(Fun f ss)` have "i < length ss" and "q \<in> poss(ss!i)" unfolding PCons poss.simps by auto
  hence "ss!i \<in> set ss" by simp
  with Fun and `q \<in> poss(ss!i)`
  have IH: "map_ctxt fg(ctxt_of_pos_term q (ss!i)) = ctxt_of_pos_term q (map_term fg (ss!i))" by simp
  have "map_ctxt fg(ctxt_of_pos_term p (Fun f ss)) = map_ctxt fg(ctxt_of_pos_term (i<#q) (Fun f ss))" unfolding PCons by simp
  also have "\<dots> = map_ctxt fg(More f (take i ss) (ctxt_of_pos_term q (ss!i)) (drop (Suc i) ss))" by simp
  also have "\<dots> = More (fg f) (map (map_term fg) (take i ss)) (map_ctxt fg(ctxt_of_pos_term q (ss!i))) (map (map_term fg) (drop (Suc i) ss))" by simp
  also have "\<dots> = More (fg f) (map (map_term fg) (take i ss)) (ctxt_of_pos_term q (map_term fg (ss!i))) (map (map_term fg) (drop (Suc i) ss))" unfolding IH by simp
  also have "\<dots> = More (fg f) (take i (map (map_term fg) ss)) (ctxt_of_pos_term q (map (map_term fg) ss!i)) (drop (Suc i) (map (map_term fg) ss))" unfolding nth_map[OF `i < length ss`,symmetric] take_map drop_map nth_map by simp
  finally show ?thesis unfolding PCons by simp
 qed
qed

(*
lemma nrrstep_imp_unsharp_nrrstep:
 assumes "(s, t) \<in> nrrstep (ID R)" shows "(unsharp_term s, unsharp_term t) \<in> nrrstep R"
proof -
  from assms obtain l r i p \<sigma>
    where step: "(s,t) \<in> rstep_r_p_s (ID R) (l,r) (i<#p) \<sigma>" unfolding nrrstep_def by auto
  let ?C = "ctxt_of_pos_term (i<#p) s"
  from step have "i<#p \<in> poss s" and "(l,r) \<in> ID R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = s"
    and "?C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
  from `(l,r) \<in> ID R` have "(unsharp_term l,unsharp_term r) \<in> R" (is "(?l,?r) \<in> R")
    unfolding map_trs.simps by auto
  from `i<#p \<in> poss s` have "i<#p \<in> poss(unsharp_term s)" by simp
  let ?D = "unsharp_ctxt ?C"
  let ?\<sigma> = "unsharp_subst \<sigma>"
  from `i<#p \<in> poss s` have ctxt: "?D = ctxt_of_pos_term (i<#p) (unsharp_term s)" by simp
  have "?D\<langle>?l\<cdot>?\<sigma>\<rangle> = unsharp_term (?C\<langle>l\<cdot>\<sigma>\<rangle>)" by simp
  hence lhs: "?D\<langle>?l\<cdot>?\<sigma>\<rangle> = unsharp_term s" unfolding `?C\<langle>l\<cdot>\<sigma>\<rangle> = s` by simp
  have "?D\<langle>?r\<cdot>?\<sigma>\<rangle> = unsharp_term (?C\<langle>r\<cdot>\<sigma>\<rangle>)" by simp
  hence rhs: "?D\<langle>?r\<cdot>?\<sigma>\<rangle> = unsharp_term t" unfolding `?C\<langle>r\<cdot>\<sigma>\<rangle> = t` by simp
  from `i<#p \<in> poss(unsharp_term s)` and `(?l,?r) \<in> R` and lhs and rhs
    show ?thesis unfolding nrrstep_def ctxt rstep_r_p_s_def Let_def by force
qed

lemma Tinf_imp_SN_sharp:
  fixes R :: "('f, 'v) trs"
  assumes "wf_trs R" and "s \<in> Tinf R"
  shows "SN_elt (rstep (ID R)) (sharp_term s)"
proof (rule ccontr)
 assume nontermin: "\<not>(SN_elt (rstep (ID R)) (sharp_term s))"
 then obtain S where "S 0 = sharp_term s" and seq:"\<forall>i. (S i,S(Suc i)) \<in> rstep(ID R)" unfolding SN_elt_def by auto
 hence union: "\<forall>i. (S i,S(Suc i)) \<in> rrstep(ID R) \<union> nrrstep(ID R)" unfolding rstep_iff_rrstep_or_nrrstep by simp
 from seq have step: "(sharp_term s,S(Suc 0)) \<in> rstep(ID R)" unfolding `S 0 = sharp_term s`[symmetric] by simp
 then obtain l r p \<sigma> where step': "(sharp_term s,S(Suc 0)) \<in> rstep_r_p_s (ID R) (l,r) p \<sigma>" unfolding rstep_iff_rstep_r_p_s by auto
 let ?C = "ctxt_of_pos_term p (sharp_term s)"
 from step' have "p \<in> poss(sharp_term s)" and "(l,r) \<in> ID R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = sharp_term s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = S(Suc 0)" unfolding rstep_r_p_s_def Let_def by auto
 from `wf_trs R` have "wf_trs (ID R)" by simp
 with `(l,r) \<in> ID R` obtain g ts where "l = Fun g ts" unfolding wf_trs_def by best
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = sharp_term s` obtain C where "C\<langle>l\<cdot>\<sigma>\<rangle> = sharp_term s" by auto
 obtain f ss where "s = Fun f ss" using `C\<langle>l\<cdot>\<sigma>\<rangle> = sharp_term s` unfolding `l = Fun g ts` by (induct s,cases C) auto
 have "sharp_term s = Fun (Sharp f) (map plain_term ss)" (is "_ = Fun _ ?ss")
   unfolding `s = Fun f ss` by simp
 hence "S 0 = Fun (Sharp f) ?ss" unfolding `S 0 = sharp_term s` by simp
 have "\<forall>u. sharp_term s \<rhd> u \<longrightarrow> SN_elt (rstep(ID R)) u"
 proof (intro allI impI)
  fix u assume "sharp_term s \<rhd> u"
  from subt_of_sharp_imp_plain[OF this]
    obtain v where "u = plain_term v" by auto
  from `sharp_term s \<rhd> u` have "s \<rhd> v" unfolding `u = plain_term v` supt_sharp_conv .
  with `s \<in> Tinf R` have "SN_elt (rstep R) v" unfolding Tinf_def supt_supteq_conv by auto
  thus "SN_elt (rstep(ID R)) u" unfolding `u = plain_term v` unfolding SN_elt_rstep_iff_SN_elt_rstepID[symmetric] by simp
 qed
 with nontermin have "sharp_term s \<in> Tinf (ID R)" unfolding Tinf_def supt_supteq_conv by auto
 hence "SN_elt (nrrstep(ID R)) (sharp_term s)" by (rule Tinf_imp_SN_elt_nrrstep)
 hence first: "SN_elt (nrrstep(ID R)) (S 0)" unfolding `S 0 = sharp_term s` by simp
 obtain i where "(S i,S(Suc i)) \<in> rrstep(ID R)"
  and prefix: "\<forall>j<i. (S j,S(Suc j)) \<in> nrrstep(ID R) \<and> (S j,S(Suc j)) \<notin> rrstep(ID R)"
  using union_iseq_SN_elt_imp_first_step[of S, OF union first] by auto
 have less: "\<forall>j<i.\<exists>us. S j = Fun (Sharp f) us"
 proof (intro allI impI)
  fix j assume "j < i" thus "\<exists>us. S j = Fun (Sharp f) us"
  proof (induct j)
   case 0 thus ?case using `S 0 = Fun (Sharp f) ?ss` by best
  next
   case (Suc j)
   then obtain us' where "S j = Fun (Sharp f) us'" by auto
   have "(S j,S(Suc j)) \<in> nrrstep(ID R)" using prefix and Suc by auto
   then obtain us'' where "S(Suc j) = Fun (Sharp f) us''" using nrrstep_preserves_root unfolding `S j = Fun (Sharp f) us'` by best
   thus ?case by auto
  qed
 qed
 have "\<exists>us. S i = Fun (Sharp f) us"
 proof (cases i)
  case 0 thus ?thesis using `S 0 = Fun (Sharp f) ?ss` by auto
 next
  case (Suc n)
  from less have "\<exists>us. S n = Fun (Sharp f) us" unfolding Suc by auto
  then obtain vs where "S n = Fun (Sharp f) vs" by auto
  have "(S n,S(Suc n)) \<in> nrrstep(ID R)" using prefix unfolding Suc by auto
  then obtain vs' where "S i = Fun (Sharp f) vs'" using nrrstep_preserves_root unfolding Suc `S n = Fun (Sharp f) vs` by fast
  thus ?thesis by auto
 qed
 then obtain us where "S i = Fun (Sharp f) us" by auto
 from `(S i,S(Suc i)) \<in> rrstep(ID R)` have "(Fun (Sharp f) us,S(Suc i)) \<in> rrstep(ID R)" unfolding `S i = Fun (Sharp f) us` by auto
 then obtain l r \<sigma> where "(Fun (Sharp f) us,S(Suc i)) \<in> rstep_r_p_s (ID R) (l,r) \<epsilon> \<sigma>" unfolding rrstep_def by auto
 hence "(l,r) \<in> ID R" and "l\<cdot>\<sigma> = Fun (Sharp f) us" unfolding rstep_r_p_s_def Let_def by auto
 then obtain l' r' where "(l',r') \<in> R" and "l = plain_term l'" unfolding map_trs.simps by auto
 with `wf_trs R` obtain g vs where "l' = Fun g vs" unfolding wf_trs_def by best
 from `l\<cdot>\<sigma> = Fun (Sharp f) us` show False unfolding `l = plain_term l'` `l' = Fun g vs` by simp
qed
*)

lemma var_rewrite_imp_not_SN: assumes sn: "SN_elt (rstep R) u" and step: "(t,s) \<in> rstep R" 
  shows "\<not> (is_var t)" 
using assms
proof (cases t)
  case (Fun f ts) thus ?thesis by simp
next
  case (Var x)
  from step obtain l r p \<sigma> where "(Var x, s) \<in> rstep_r_p_s R (l,r) p \<sigma>" unfolding Var rstep_iff_rstep_r_p_s by best
  hence "l \<cdot> \<sigma> = Var x" and rule: "(l,r) \<in> R" unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from this obtain y where "l = Var y" (is "_ = ?y") by (cases l, auto)
  with rule have "(?y, r) \<in> R" by auto
  hence "\<not> (SN_elt (rstep R) u)" by (rule left_var_imp_not_SN)
  with sn show ?thesis by blast
qed

lemma rstep_empty[simp]: "rstep {} = {}"
proof (rule set_eqI, simp, rule notI, simp only: split_paired_all)
  fix l r
  assume "(l,r) \<in> rstep {}"
  thus False by (simp only: rstep_iff_rstep_r_p_s rstep_r_p_s_def, auto)
qed

lemma rstep_union: "rstep (R \<union> S) = rstep R \<union> rstep S"
proof (rule set_eqI, simp only: split_paired_all)
  fix s t
  show "((s,t) \<in> rstep (R \<union> S)) = ((s,t) \<in> rstep R \<union> rstep S)"
    by (simp only: Un_iff, simp only: rstep_iff_rstep_r_p_s, unfold rstep_r_p_s_def, force simp: Let_def)
qed

lemma rstep_id: "rstep Id = Id"
  by (rule set_eqI, simp only: split_paired_all, simp only: rstep_iff_rstep_r_p_s, unfold rstep_r_p_s_def, auto simp: Let_def, 
    rule exI, rule exI[where x = \<epsilon>], auto, rule exI[where x = Var], auto)

lemma rstep_imp_wfrstep:
  assumes "(s, t) \<in> rstep_r_p_s R (l, r) p \<sigma>"
  shows "(l \<cdot> (clean_subst (funas_trs R) \<sigma>), r \<cdot> (clean_subst (funas_trs R) \<sigma>)) \<in> wfrstep R"
proof -
  let ?l = "l \<cdot> \<sigma>"
  let ?r = "r \<cdot> \<sigma>"
  let ?u = "clean_term (funas_trs R) ?l"
  let ?v = "clean_term (funas_trs R) ?r"
  let ?C = "ctxt_of_pos_term p s"
  let ?\<sigma> = "(clean_subst (funas_trs R) \<sigma>)"
  from assms have "p \<in> poss s" and "(l, r) \<in> R"
    and "?C\<langle>l \<cdot> \<sigma>\<rangle> = s" and "?C\<langle>r \<cdot> \<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
  from `(l,r) \<in> R` have "funas_term l \<subseteq> funas_trs R"
    unfolding funas_trs_def funas_rule_def by auto
  moreover have "l \<in> wf_terms (funas_term l)" by simp
  ultimately have "l \<in> wf_terms (funas_trs R)"
    using wf_terms_subset[of "funas_term l" "funas_trs R"] by best
  from `(l,r) \<in> R` have "funas_term r \<subseteq> funas_trs R"
    unfolding funas_trs_def funas_rule_def by auto
  moreover have "r \<in> wf_terms (funas_term r)" by simp
  ultimately have "r \<in> wf_terms (funas_trs R)"
    using wf_terms_subset[of "funas_term r" "funas_trs R"] by best
  have "?u = l\<cdot>?\<sigma>" using `l \<in> wf_terms (funas_trs R)` by simp
  have "?v = r\<cdot>?\<sigma>" using `r \<in> wf_terms (funas_trs R)` by simp
  moreover have "(?u, ?v) \<in> wfrstep R"
  proof -
    have "?u \<in> wf_terms (funas_trs R)" by simp
    moreover have "?v \<in> wf_terms (funas_trs R)" by simp
    moreover have "(?u, ?v) \<in> rstep R"
    proof -
      from `(l,r) \<in> R` have "(?u,?v) \<in> CS R" unfolding `?u = l\<cdot>?\<sigma>` `?v = r\<cdot>?\<sigma>` by auto
      thus ?thesis using ctxt_closureI[of ?u ?v "CS R" \<box>] unfolding rstep_CC_CS by simp
    qed
    ultimately show ?thesis unfolding wfrstep_def by force
  qed
  ultimately show ?thesis unfolding `?u = l\<cdot>?\<sigma>` `?v = r\<cdot>?\<sigma>` by auto
qed

lemma rstep_imp_clean_rstep_or_Id:
  fixes R::"('f, 'v) trs" and F::"'f sig"
  defines [simp]: "c \<equiv> clean_term F"
  assumes subset: "funas_trs R \<subseteq> F"
    and step: "(s, t) \<in> rstep R"
  shows "(c s, c t) \<in> rstep R \<union> Id"
proof -
  from rstep_imp_C_s_r[OF step] obtain C l r \<sigma>
    where R: "(l,r) \<in> R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>"
    and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
  from s t show ?thesis
  proof (induct C arbitrary: s t)
    case Hole
    let ?c = "\<lambda>\<sigma>. (c \<circ> \<sigma>)"
    from clean_subst_apply_term[OF lhs_wf[OF R subset]]
      have s: "c(s) = l\<cdot>?c(\<sigma>)" unfolding Hole by simp
    from clean_subst_apply_term[OF rhs_wf[OF R subset]]
      have t: "c(t) = r\<cdot>?c(\<sigma>)" unfolding Hole by simp
    from R have "(c(s),c(t)) \<in> CS R" unfolding s t by auto
    hence "(c(s),c(t)) \<in> rstep R" using ctxt_closureI[where C=\<box>] unfolding rstep_CC_CS by simp
    thus ?case ..
  next
    case (More f ss1 D ss2)
    hence IH: "(c(D\<langle>l\<cdot>\<sigma>\<rangle>),c(D\<langle>r\<cdot>\<sigma>\<rangle>)) \<in> rstep R \<union> Id" by simp
    show ?case
    proof (cases "(f,Suc(length(ss1@ss2))) \<in> F")
      case False
      with More have "c(s) = Var(SOME(x::'v). True)" by simp
      moreover from False More have "c(t) = Var(SOME(x::'v). True)" by simp
      ultimately have "(c(s),c(t)) \<in> Id" by simp
      thus ?thesis ..
    next
      case True
      let ?s = "c(D\<langle>l\<cdot>\<sigma>\<rangle>)"
      let ?t = "c(D\<langle>r\<cdot>\<sigma>\<rangle>)"
      from IH show ?thesis
      proof
        assume "(?s,?t) \<in> Id" with True More show ?thesis by auto
      next
        let ?C = "More f (map c ss1) \<box> (map c ss2)"
        assume "(?s,?t) \<in> rstep R"
        from rstep.ctxt[OF this,where C="?C"]
          have "(c(s),c(t)) \<in> rstep R" unfolding More using True by simp
        thus ?thesis ..
      qed
    qed
  qed
qed

lemma rsteps_imp_clean_rsteps:
  fixes R::"('f, 'v) trs" and F::"'f sig"
  defines [simp]: "c \<equiv> clean_term F"
  assumes subset: "funas_trs R \<subseteq> F"
  and steps: "(s,t) \<in> (rstep R)^*"
  shows "(c s, c t) \<in> (rstep R)^*"
using steps proof (induct rule: rtrancl_induct)
  case base show ?case by simp
next
  case (step y z)
  from subset have "funas_trs R \<subseteq> F" by auto
  from rstep_imp_clean_rstep_or_Id[OF this `(y,z) \<in> rstep R`]
  have "(c(y),c(z)) \<in> rstep R \<union> Id" by (simp only: c_def)
  thus ?case
  proof
    assume "(c y,c z) \<in> Id" with step show ?thesis by simp
  next
    assume "(c y,c z) \<in> rstep R" with step show ?thesis by auto
  qed
qed

(*
lemma funas_term_plain[simp]:
  "(f, n) \<in> funas_term (plain_term t) \<Longrightarrow> (!f!, n) \<in> funas_term t"
by (induct t) auto

lemma funasID_funas:
  fixes R::"('f,'v)trs" assumes "(f,n) \<in> funas_trs(ID R)" shows "(!f!,n) \<in> funas_trs R"
proof -
  from assms obtain l r::"('f shp,'v)term"
    where ID_R: "(l,r) \<in> ID R" and "(f,n) \<in> funas_rule(l,r)"
    unfolding funas_trs_def by auto
  from ID_R obtain l' r'::"('f,'v)term" where "l = plain_term l'" and "r = plain_term r'" and "(l',r') \<in> R"
    unfolding map_trs.simps by auto
  from `(f,n) \<in> funas_rule(l,r)` have "(f,n) \<in> funas_term l \<or> (f,n) \<in> funas_term r" unfolding funas_rule_def by auto
  thus ?thesis
  proof
    assume "(f,n) \<in> funas_term l"
    from `(f,n) \<in> funas_term l` and `l = plain_term l'` have "(!f!,n) \<in> funas_term l'" by simp
    with `(l',r') \<in> R` show ?thesis unfolding funas_trs_def funas_rule_def by force
  next
    assume "(f,n) \<in> funas_term r"
    from `(f,n) \<in> funas_term r` and `r = plain_term r'` have "(!f!,n) \<in> funas_term r'" by simp
    with `(l',r') \<in> R` show ?thesis unfolding funas_trs_def funas_rule_def by force
  qed
qed


lemma funas_term_sharp[simp]:
  fixes t::"('f,'v)term" assumes "(f,n) \<in> funas_term(sharp_term t)" shows "(!f!,n) \<in> funas_term t"
proof (cases t)
  case (Var x) with assms show ?thesis by simp
next
  case (Fun g ts)
  from assms have t: "sharp_term t = Fun (Sharp g) (map plain_term ts)" unfolding Fun by simp
  hence "funas_term(sharp_term t) = {(Sharp g,length(map plain_term ts))} \<union> \<Union>set(map funas_term (map plain_term ts))" by simp
  with assms have "(f,n) = (Sharp g,length(map plain_term ts)) \<or> (f,n) \<in> \<Union>set(map funas_term (map plain_term ts))" (is "?A \<or> ?B") by auto
  thus ?thesis
  proof
    assume ?A with assms show ?thesis unfolding Fun by simp
  next
    assume ?B
    hence "(f,n) \<in> \<Union>set(map (funas_term \<circ> plain_term) ts)" by simp
    then obtain t where "t \<in> set ts" and "(f,n) \<in> funas_term(plain_term t)" unfolding o_def by auto
    hence "(!f!,n) \<in> funas_term t" by simp
    with `t \<in> set ts` show ?thesis unfolding Fun by auto
  qed
qed

lemma funasDP_funas:
  fixes R::"('f,'v)trs" assumes "(f,n) \<in> funas_trs(DP R)" shows "(!f!,n) \<in> funas_trs R"
proof -
  from assms obtain l u::"('f shp,'v)term"
    where DP_R: "(l,u) \<in> DP R" and "(f,n) \<in> funas_rule(l,u)"
    unfolding funas_trs_def by auto
  from DP_R obtain l' r'::"('f,'v)term" and g ts where "l = sharp_term l'" and "u = sharp_term (Fun g ts)" and "(l',r') \<in> R"
    and "r' \<unrhd> Fun g ts" 
    unfolding DP_def supt_supteq_conv by auto
  from `(f,n) \<in> funas_rule(l,u)` have "(f,n) \<in> funas_term l \<or> (f,n) \<in> funas_term u" unfolding funas_rule_def by auto
  thus ?thesis
  proof
    assume "(f,n) \<in> funas_term l"
    with `l = sharp_term l'` have "(!f!,n) \<in> funas_term l'" by simp
    with `(l',r') \<in> R` show ?thesis unfolding funas_trs_def funas_rule_def by force
  next
    assume "(f,n) \<in> funas_term u"
    from `r' \<unrhd> Fun g ts` have subs: "funas_term(Fun g ts) \<subseteq> funas_term(r')" by (rule supteq_imp_funas_subset)
    from `(f,n) \<in> funas_term u` have "(f,n) \<in> funas_term(Fun (Sharp g) (map plain_term ts))" unfolding `u = sharp_term (Fun g ts)` by simp
    hence "(f,n) = (Sharp g,length(map plain_term ts)) \<or> (f,n) \<in> \<Union>set(map funas_term (map plain_term ts))" (is "?A \<or> ?B") by auto
    thus ?thesis
    proof
      assume ?A
      with subs have "(!f!,n) \<in> funas_term r'" by simp
      thus ?thesis using `(l',r') \<in> R` unfolding funas_trs_def funas_rule_def by force
    next
      assume ?B
      then obtain t where "t \<in> set ts" and "(f,n) \<in> funas_term(plain_term t)" unfolding map_map[symmetric] o_def by auto
      hence "(!f!,n) \<in> funas_term t" by simp
      with `t \<in> set ts` have "(!f!,n) \<in> funas_term(Fun g ts)" by auto
      with `r' \<unrhd> Fun g ts` have "(!f!,n) \<in> funas_term r'" using supteq_imp_funas_subset by best
      thus ?thesis using `(l',r') \<in> R` unfolding funas_trs_def funas_rule_def by force
    qed
  qed
qed

lemma unsharp_clean_conv[simp]:
  fixes R::"('f,'v)trs"
  defines "c \<equiv> clean_term (funas_trs R)"
    and "c' \<equiv> clean_term ({(Plain f,n) | f n. (f,n) \<in> funas_trs R} \<union> {(Sharp f,n) | f n. (f,n) \<in> funas_trs R})" (is "_ \<equiv> clean_term ?F")
  shows "unsharp_term (c' t) = c(unsharp_term t)"
proof (induct t)
  case (Var x) show ?case by (simp add: c_def c'_def)
next
  case (Fun f ts)
  hence IH: "\<And>s. s \<in> set ts \<Longrightarrow> unsharp_term (c' s) = c(unsharp_term s)" by simp
  show ?case
  proof (cases "(!f!,length ts) \<in> funas_trs R")
    case False thus ?thesis unfolding c_def c'_def by auto
  next
    case True show ?thesis
    proof (cases f)
      case (Sharp g)
      with True have "(f,length ts) \<in> ?F" by simp
      with IH show ?thesis unfolding c'_def c_def Sharp by simp
    next
      case (Plain g)
      with True have "(f,length ts) \<in> ?F" by simp
      with IH show ?thesis unfolding c'_def c_def Plain by auto
    qed
  qed
qed

lemma rstepID_imp_rstep':
  assumes "(s,t) \<in> rstep(ID R)" shows "(unsharp_term s,unsharp_term t) \<in> rstep R"
proof -
  from rstep_imp_C_s_r[OF assms] obtain C \<sigma> l r
    where ID_R: "(l,r) \<in> ID R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>"
    and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
  from ID_R obtain l' r' where R: "(l',r') \<in> R" and l: "l = plain_term l'" and r: "r = plain_term r'"
    unfolding map_trs.simps by auto
  let ?C = "unsharp_ctxt C"
  let ?\<sigma> = "case \<sigma> of Subst s \<Rightarrow> Subst(\<lambda> x. (unsharp_term (s x)))"
  have us: "unsharp_term s = ?C\<langle>l'\<cdot>?\<sigma>\<rangle>" unfolding s l
    unfolding map_ctxt_distrib map_subst_distrib by (cases \<sigma>) (auto)
  have ut: "unsharp_term t = ?C\<langle>r'\<cdot>?\<sigma>\<rangle>" unfolding t r unfolding map_ctxt_distrib map_subst_distrib by (cases \<sigma>) (auto)
  from R have "(l'\<cdot>?\<sigma>,r'\<cdot>?\<sigma>) \<in> CS R" using subst_closureI by auto
  thus ?thesis unfolding us ut using ctxt_closureI[where C="?C"] unfolding rstep_CC_CS by auto
qed
*)

lemma map_rule_id[simp]: "map_rule id = id"
proof (intro ext)
  fix lr
  show "map_rule id lr = id lr" by (cases lr, auto)
qed

lemma map_trs_id[simp]: "map_trs id = id"
  by (intro ext, auto simp: map_trs.simps)

(*
lemma rstepIDs_imp_rsteps:
  assumes "(s,t) \<in> (rstep(ID R))^*" shows "(unsharp_term s,unsharp_term t) \<in> (rstep R)^*"
using rsteps_imp_map_rsteps[OF assms, of unsharp_fun]
by (simp add: map_trs_comp o_def map_trs_id[unfolded id_def]) 
*)

lemma wfrstep_trancl:
  assumes "wf_trs R"
  shows "(wfrstep R)^+ = {(x,y) | x y. (x,y) \<in> (rstep R)^+ \<and> x \<in> wf_terms (funas_trs R)
    \<and> y \<in> wf_terms (funas_trs R)}" (is "?S = ?R")
proof
  show "?R \<subseteq> ?S"
  proof (rule subsetI2)
    have subset: "funas_trs R \<subseteq> funas_trs R" by simp
    fix s t assume "(s,t) \<in> ?R"
    hence "(s,t) \<in> (rstep R)^+" and "s \<in> wf_terms (funas_trs R)" by auto
    hence "(s,t) \<in> ?S \<and> t \<in> wf_terms (funas_trs R)"
    proof (induct rule: trancl.induct)
      case (r_into_trancl a b)
      hence "a \<in> wf_terms (funas_trs R)" and "(a,b) \<in> rstep R" using r_into_trancl by auto
      from rstep_preserves_wf_terms[OF subset `a \<in> wf_terms (funas_trs R)`
        `(a,b) \<in> rstep R` assms] have "b \<in> wf_terms (funas_trs R)" .
      with `a \<in> wf_terms (funas_trs R)` `(a,b) \<in> rstep R`
        have "(a,b) \<in> wfrstep R" unfolding wfrstep_def by blast
      thus ?case using `b \<in> wf_terms (funas_trs R)` by simp
    next
      case (trancl_into_trancl a b c)
      hence IH: "(a,b) \<in> ?S" and "b \<in> wf_terms (funas_trs R)" by auto
      from rstep_preserves_wf_terms[OF subset `b \<in> wf_terms (funas_trs R)`
        `(b,c) \<in> rstep R` assms] have "c \<in> wf_terms (funas_trs R)" .
      with `b \<in> wf_terms (funas_trs R)` `(b,c) \<in> rstep R` have "(b,c) \<in> wfrstep R" 
        unfolding wfrstep_def by blast
      with `(a,b) \<in> (wfrstep R)^+` and `c \<in> wf_terms (funas_trs R)` show ?case by auto
    qed
    thus "(s,t) \<in> (wfrstep R)^+" by simp
  qed
next
  show "?S \<subseteq> ?R"
  proof (rule subsetI2)
    fix s t assume "(s,t) \<in> ?S" thus "(s,t) \<in> ?R"
    by (induct rule: trancl.induct) (auto simp: wfrstep_def)
  qed
qed

(*
lemma DP_step_in_R:
  fixes R::"('f,'v)trs" and v::"('f,'v)term \<Rightarrow> 'v"
  assumes "(s,t) \<in> DP R" shows "\<exists>C. C \<in> wf_ctxts (funas_trs R) \<and> (unsharp_term s,C\<langle>unsharp_term t\<rangle>) \<in> R"
proof -
  from assms obtain l r f ts
    where s: "s = sharp_term l" and t: "t = sharp_term (Fun f ts)"
    and R: "(l,r) \<in> R" and sub: "r \<unrhd> Fun f ts" unfolding DP_def supt_supteq_conv by auto
  from sub obtain C where r: "r = C\<langle>Fun f ts\<rangle>" by auto
  have subset: "funas_trs R \<subseteq> funas_trs R" by simp
  from rhs_wf[OF R subset] have "r \<in> wf_terms (funas_trs R)" .
  hence "C\<langle>Fun f ts\<rangle> \<in> wf_terms (funas_trs R)" unfolding r .
  from ctxt_apply_wf_imp_ctxt_and_term_wf[OF this]
    have "C \<in> wf_ctxts (funas_trs R)" and "Fun f ts \<in> wf_terms (funas_trs R)" by auto
  from lhs_wf[OF R subset] have "l \<in> wf_terms (funas_trs R)" .
  have us: "unsharp_term s = l" unfolding s by simp
  have ut: "unsharp_term t = Fun f ts" unfolding t unsharp_sharp_id[of "Fun f ts"] ..
  from R have "(unsharp_term s,C\<langle>unsharp_term t\<rangle>) \<in> R" unfolding us ut r .
  with `C \<in> wf_ctxts (funas_trs R)` show ?thesis by best
qed

lemma funas_DP_ID_subset:
  "funas_trs(DP R) \<union> funas_trs(ID R)
     \<subseteq> {(Plain f,n) | f n. (f,n) \<in> funas_trs R} \<union> {(Sharp f,n) | f n. (f,n) \<in> funas_trs R}"
     (is "?F \<union> ?G \<subseteq> ?H \<union> ?I")
proof (rule Un_least)
  show "?F \<subseteq> ?H \<union> ?I"
  proof (rule subsetI2)
    fix f n assume "(f,n) \<in> ?F" hence "(!f!,n) \<in> funas_trs R" by (rule funasDP_funas)
    thus "(f,n) \<in> ?H \<union> ?I" by (cases f) auto
  qed
next
  show "?G \<subseteq> ?H \<union> ?I"
  proof (rule subsetI2)
    fix f n assume "(f,n) \<in> ?G" hence "(!f!,n) \<in> funas_trs R" by (rule funasID_funas)
    thus "(f,n) \<in> ?H \<union> ?I" by (cases f) auto
  qed
qed
*)

lemma lhs_var_imp_wfrstep_not_SN: assumes "(Var x,r) \<in> R" shows "\<not> SN(wfrstep R)"
proof -
  let ?\<sigma> = "(\<lambda>x. r)"
  let ?t = "\<lambda>i. r\<cdot>(?\<sigma> ^^ i)"
  from rhs_wf[OF assms] have wf_r: "r \<in> wf_terms (funas_trs R)" by fast
  have clean: "?\<sigma> = (\<lambda>x. clean_term (funas_trs R) r)"
    using wf_terms_clean_term_ident[OF wf_r] by (simp add: o_def)
  have "\<forall>i. ?t i \<in> wf_terms (funas_trs R)"
  proof
    fix i show "?t i \<in> wf_terms (funas_trs R)"
    proof (induct i)
      case 0 from wf_r show ?case by (simp)
    next
      case (Suc i)
      from subst_apply_term_clean_wf_terms[OF Suc,of "Subst (\<lambda>x. r)"]
        have "r\<cdot>?\<sigma> ^^ i \<cdot> ?\<sigma> \<in> wf_terms (funas_trs R)" by simp
      hence "r\<cdot>?\<sigma> ^^ i \<circ>s ?\<sigma> ^^ Suc 0 \<in> wf_terms (funas_trs R)"
        by (simp)
      thus ?case unfolding subst_power_compose_distrib[symmetric] by simp
    qed
  qed
  moreover have "\<forall>i. (?t i,?t(Suc i)) \<in> rstep R"
  proof
    fix i
    from subst_closureI[OF assms,of "?\<sigma> ^^ Suc i"] have "(?t i,?t(Suc i)) \<in> CS R" by (simp)
    from ctxt_closureI[OF this,of \<box>] show "(?t i,?t(Suc i)) \<in> rstep R" unfolding rstep_CC_CS by simp
  qed
  ultimately have "\<forall>i. (?t i,?t(Suc i)) \<in> wfrstep R" unfolding wfrstep_def by blast
  thus ?thesis unfolding SN_defs by best
qed

lemma rhs_free_vars_imp_wfrstep_not_SN:
  assumes R: "(l,r) \<in> R" and free: "\<not> vars_term r \<subseteq> vars_term l" shows "\<not> SN_elt (wfrstep R) l"
proof -
  from free obtain x where x: "x \<in> vars_term r - vars_term l" by auto
  hence "x \<in> vars_term r" by simp
  from var_imp_supteq[OF this] have "r \<unrhd> Var x" .
  then obtain C where r: "C\<langle>Var x\<rangle> = r" by auto
  let ?\<sigma> = "\<lambda>y. if y = x then l else Var y"
  let ?t = "\<lambda>i. ((C \<cdot>c ?\<sigma>)^i)\<langle>l\<rangle>"
  from rhs_wf[OF R] have wf_r: "r \<in> wf_terms (funas_trs R)" by fast
  from lhs_wf[OF R] have wf_l: "l \<in> wf_terms (funas_trs R)" by fast
  from ctxt_apply_wf_imp_ctxt_and_term_wf[OF wf_r[unfolded r[symmetric]]]
    have wf_C: "C \<in> wf_ctxts (funas_trs R)" by simp
  have clean: "?\<sigma> = (\<lambda>y. clean_term (funas_trs R) (?\<sigma> y))"
    unfolding if_distrib wf_terms_clean_term_ident[OF wf_l] clean_term.simps by simp
  from x have neq: "\<forall>y\<in>vars_term l. y \<noteq> x" by auto
  have l: "l\<cdot>?\<sigma> = l"
  proof -
    have lsigma: "l\<cdot>?\<sigma> = l\<cdot>(?\<sigma> |s vars_term l)" unfolding coincidence_lemma[symmetric] by simp
    have aux: "(\<lambda>y. if y \<in> vars_term l then if y = x then l else Var y else Var y) = Var"
    proof
      fix y show "(if y \<in> vars_term l then if y = x then l else Var y else Var y) = Var y" using neq by auto
    qed
    have "l\<cdot>(?\<sigma> |s vars_term l) = l" unfolding subst_restrict_def unfolding aux by simp
    thus ?thesis unfolding lsigma .
  qed
  have rsigma: "r\<cdot>?\<sigma> = (C \<cdot>c ?\<sigma>)\<langle>l\<rangle>" unfolding r[symmetric] by simp
  have "\<forall>i. ?t i \<in> wf_terms (funas_trs R)"
  proof
    fix i show "?t i \<in> wf_terms (funas_trs R)"
    proof (induct i)
      case 0 with wf_l show ?case by (simp add: one_ctxt_def)
    next
      case (Suc i)
      from subst_apply_ctxt_clean_subst_wf_ctxts[OF wf_C,of "(\<lambda>y. ?\<sigma> y)"]
        have "C \<cdot>c ?\<sigma> \<in> wf_ctxts (funas_trs R)" by (simp add: o_def clean del: clean_term_if)
      with Suc show ?case by (simp add: times_ctxt_def)
    qed
  qed
  moreover have "\<forall>i. (?t i,?t(Suc i)) \<in> rstep R"
  proof
    fix i
    from subst_closureI[OF R,of "?\<sigma>"] have "(l,(C \<cdot>c ?\<sigma>)\<langle>l\<rangle>) \<in> CS R" unfolding l rsigma .
    from ctxt_closureI[OF this,of "(C \<cdot>c ?\<sigma>)^i"]
      have "(((C \<cdot>c ?\<sigma>)^i)\<langle>l\<rangle>, ((C \<cdot>c ?\<sigma>)^i \<circ>c (C \<cdot>c ?\<sigma>)^Suc 0)\<langle>l\<rangle>) \<in> rstep R"
      unfolding rstep_CC_CS
      by (simp add: one_ctxt_def times_ctxt_def)
    thus "(?t i,?t(Suc i)) \<in> rstep R" unfolding ctxt_power_compose_distr[symmetric] by simp
  qed
  ultimately have steps: "\<forall>i. (?t i,?t(Suc i)) \<in> wfrstep R" unfolding wfrstep_def by blast
  show ?thesis unfolding SN_defs 
    by (simp, intro exI[of _ ?t], simp only: steps, simp)
qed

lemma lhs_var_imp_rstep_not_SN: assumes "(Var x,r) \<in> R" shows "\<not> SN(rstep R)"
using lhs_var_imp_wfrstep_not_SN[OF assms] unfolding wfrstep_def SN_defs by blast

lemma rhs_free_vars_imp_rstep_not_SN:
  assumes "(l,r) \<in> R" and "\<not> vars_term r \<subseteq> vars_term l"
  shows "\<not> SN_elt (rstep R) l"
using rhs_free_vars_imp_wfrstep_not_SN[OF assms] unfolding wfrstep_def SN_defs by blast

lemma free_right_rewrite_imp_not_SN:
  fixes l r::"('f,'v)term"
  assumes step: "(t,s) \<in> rstep_r_p_s R (l,r) p \<sigma>" 
  and vars: "\<not> vars_term l \<supseteq> vars_term r" 
  shows "\<not> SN_elt (rstep R) t" 
proof
  assume SN: "SN_elt (rstep R) t"
  let ?C = "ctxt_of_pos_term p t"
  from step have left: "?C\<langle>l \<cdot> \<sigma>\<rangle> = t" (is "?t = t") and right: "?C\<langle>r \<cdot> \<sigma>\<rangle> = s" and pos: "p \<in> poss t"
    and rule: "(l,r) \<in> R"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from rhs_free_vars_imp_rstep_not_SN[OF rule vars] have nSN:"\<not> SN_elt (rstep R) l" by simp
  from SN_imp_SN_subt[OF SN ctxt_imp_supteq[of ?C "l \<cdot> \<sigma>", simplified left]]
  have SN: "SN_elt (rstep R) (l \<cdot> \<sigma>)" .
  from SNinstance_imp_SN[OF SN] nSN show False by simp
qed

lemma not_SN_elt_rstep_subst_apply_term[intro]:
  assumes "\<not> SN_elt (rstep R) t" shows "\<not> SN_elt (rstep R) (t \<cdot> \<sigma>)"
using assms unfolding SN_elt_def by best

lemma SN_rstep_imp_wf_trs: assumes "SN (rstep R)" shows "wf_trs R"
proof (rule ccontr)
  assume "\<not> wf_trs R"
  then obtain l r where R: "(l,r) \<in> R"
    and not_wf: "(\<forall>f ts. l \<noteq> Fun f ts) \<or> \<not>(vars_term r \<subseteq> vars_term l)" unfolding wf_trs_def
    by auto
  from not_wf have "\<not> SN (rstep R)"
  proof
    assume free: "\<not> vars_term r \<subseteq> vars_term l"
    from rhs_free_vars_imp_rstep_not_SN[OF R free] show ?thesis unfolding SN_defs by auto
  next
    assume "\<forall>f ts. l \<noteq> Fun f ts"
    then obtain x where l:"l = Var x" by (cases l) auto
    with R have "(Var x,r) \<in> R" unfolding l by simp
    from lhs_var_imp_rstep_not_SN[OF this] show ?thesis by simp
  qed
  with assms show False by blast
qed
  
lemma SN_wfrstep_imp_wf_trs: assumes "SN (wfrstep R)" shows "wf_trs R"
proof (rule ccontr)
  assume "\<not> wf_trs R"
  then obtain l r where R: "(l,r) \<in> R"
    and not_wf: "(\<forall>f ts. l \<noteq> Fun f ts) \<or> \<not>(vars_term r \<subseteq> vars_term l)" unfolding wf_trs_def
    by auto
  from not_wf have "\<not> SN (wfrstep R)"
  proof
    assume free: "\<not> vars_term r \<subseteq> vars_term l"
    from rhs_free_vars_imp_wfrstep_not_SN[OF R free] show ?thesis unfolding SN_def by auto
  next
    assume "\<forall>f ts. l \<noteq> Fun f ts"
    then obtain x where l:"l = Var x" by (cases l) auto
    with R have "(Var x,r) \<in> R" unfolding l by simp
    from lhs_var_imp_wfrstep_not_SN[OF this] show ?thesis by simp
  qed
  with assms show False by blast
qed

lemma rstep_cases'[consumes 1, case_names root nonroot]:
assumes rstep: "(s,t) \<in> rstep R"
  and root: "\<And>l r \<sigma>. (l,r) \<in> R \<Longrightarrow> l\<cdot>\<sigma> = s \<Longrightarrow> r\<cdot>\<sigma> = t \<Longrightarrow> P"
  and nonroot: "\<And>f ss1 u ss2 v. s = Fun f (ss1@u#ss2) \<Longrightarrow> t = Fun f (ss1@v#ss2) \<Longrightarrow> (u,v) \<in> rstep R \<Longrightarrow> P"
shows "P"
proof -
  from rstep_imp_C_s_r[OF rstep] obtain C \<sigma> l r
    where R: "(l,r) \<in> R" and s: "C\<langle>l\<cdot>\<sigma>\<rangle> = s" and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t" by fast
  show ?thesis proof (cases C)
    case Hole
    from s t have "l\<cdot>\<sigma> = s" and "r\<cdot>\<sigma> = t" by (auto simp: Hole)
    with R show ?thesis by (rule root)
  next
    case (More f ss1 D ss2)
    let ?u = "D\<langle>l\<cdot>\<sigma>\<rangle>"
    let ?v  = "D\<langle>r\<cdot>\<sigma>\<rangle>"
    have "s = Fun f (ss1@?u#ss2)" by (simp add: More s[symmetric])
    moreover have "t = Fun f (ss1@?v#ss2)" by (simp add: More t[symmetric])
    moreover have "(?u,?v) \<in> rstep R" by (intro rstep.intros) (rule R)
    ultimately show ?thesis by (rule nonroot)
  qed
qed

lemma NF_var: assumes wf: "wf_trs R" shows "(Var x,t) \<notin> rstep R"
proof
  assume "(Var x,t) \<in> rstep R"
  from rstep_imp_C_s_r[OF this] obtain C l r \<sigma>
    where R: "(l,r) \<in> R" and lhs: "Var x = C\<langle>l\<cdot>\<sigma>\<rangle>" by fast
  from lhs have "Var x = l\<cdot>\<sigma>" by (induct C) auto
  then obtain y where l: "l = Var y" by (induct l) auto
  from wf R obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
  with l show False by simp
qed

lemma rstep_cases_Fun'[consumes 2, case_names root nonroot]:
assumes wf: "wf_trs R"
  and rstep: "(Fun f ss,t) \<in> rstep R"
  and root': "\<And>ls r \<sigma>. (Fun f ls,r) \<in> R \<Longrightarrow> map (\<lambda>t. t\<cdot>\<sigma>) ls = ss \<Longrightarrow> r\<cdot>\<sigma> = t \<Longrightarrow> P"
  and nonroot': "\<And>i u. i < length ss \<Longrightarrow> t = Fun f (take i ss@u#drop (Suc i) ss) \<Longrightarrow> (ss!i,u) \<in> rstep R \<Longrightarrow> P"
shows "P"
using rstep proof (cases rule: rstep_cases')
  case (root l r \<sigma>)
  with wf obtain g ls where l: "l = Fun g ls" unfolding wf_trs_def by best
  from root have [simp]: "g = f" unfolding l by simp
  from root have "(Fun f ls,r) \<in> R" and "map (\<lambda>t. t\<cdot>\<sigma>) ls = ss" and "r\<cdot>\<sigma> = t" unfolding l by auto
  thus ?thesis by (rule root')
next
  case (nonroot g ss1 u ss2 v)
  hence [simp]: "g = f" and args: "ss = ss1@u#ss2" by auto
  let ?i = "length ss1"
  from args have ss1: "take ?i ss = ss1" by simp
  from args have "drop ?i ss = u#ss2" by simp
  hence "drop (Suc 0) (drop ?i ss) = ss2" by simp
  hence ss2: "drop (Suc ?i) ss = ss2" by simp
  from args have len: "?i < length ss" by simp
  from id_take_nth_drop[OF len] have "ss = take ?i ss @ ss!?i # drop (Suc ?i) ss" by simp
  hence u: "ss!?i = u" unfolding args unfolding ss1[unfolded args] ss2[unfolded args] by simp
  from nonroot have "t = Fun f (take ?i ss@v#drop (Suc ?i) ss)" unfolding ss1 ss2 by simp
  moreover from nonroot have "(ss!?i,v) \<in> rstep R" unfolding u by simp
  ultimately show ?thesis by (rule nonroot'[OF len])
qed

lemma rstep_preserves_undefined_root:
  assumes "wf_trs R" and "\<not> defined R (f, length ss)" and "(Fun f ss, t) \<in> rstep R"
  shows "\<exists>ts. length ts = length ss \<and> t = Fun f ts"
proof -
  from `wf_trs R` and `(Fun f ss, t) \<in> rstep R` show ?thesis
  proof (cases rule: rstep_cases_Fun')
    case (root ls r \<sigma>)
    hence "defined R (f, length ss)" by (auto simp: defined_def)
    with `\<not> defined R (f, length ss)` show ?thesis by simp
  next
    case (nonroot i u) thus ?thesis by simp
  qed
qed

lemma funas_DP_subset:
  "funas_trs(DP shp R)
     \<subseteq> funas_trs R \<union> {(shp f,n) | f n. (f,n) \<in> funas_trs R}"
     (is "?F \<subseteq> ?H \<union> ?I")
proof (rule subsetI2)
  fix f n 
  assume "(f,n) \<in> ?F"
  then obtain s t where st: "(s,t) \<in> DP shp R" and "(f,n) \<in> funas_rule (s,t)" 
    unfolding funas_trs_def by auto
  then obtain u where fn: "(f,n) \<in> funas_term u" and u: "u = s \<or> u = t" unfolding funas_rule_def
    by auto
  from st[unfolded DP_def] obtain l r uu where lr: "(l,r) \<in> R" and s: "s = sharp_term shp l" and t: "t = sharp_term shp uu" and uu: "r \<unrhd> uu"
    by force
  from fn u[unfolded s t] obtain v where fn: "(f,n) \<in> funas_term (sharp_term shp v)" and v: "v = l \<or> v = uu" by auto
  from fn have fn: "(f,n) \<in> funas_term v \<union> {(shp f,n) | f n. (f,n) \<in> funas_term v}"
    by (cases v, auto)
  from uu obtain C where r: "r = C\<langle>uu\<rangle>" ..  
  have "funas_term uu \<subseteq> funas_term r" unfolding r by simp
  with v have "funas_term v \<subseteq> funas_rule (l,r)" unfolding funas_rule_def by auto
  hence subset: "funas_term v \<subseteq> funas_trs R" using lr unfolding funas_trs_def by auto
  with fn show "(f,n) \<in> ?H \<union> ?I" by auto
qed

lemma rstep_ctxt_imp_nrrstep: assumes step: "(s,t) \<in> rstep R" and C: "C \<noteq> \<box>" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> nrrstep R"
proof -
  from step[unfolded rstep_iff_rstep_r_c_s rstep_r_c_s_def]
  obtain l r D \<sigma> where lr: "(l,r) \<in> R" and s: "s = D\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = D\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  show ?thesis
    by (unfold nrrstep_def', rule, clarify, rule exI[of _ l], rule exI[of _ r], rule exI[of _ "C \<circ>c D"], rule exI[of _ \<sigma>], simp add: lr s t, cases C, auto simp: C)
qed

lemma rsteps_ctxt_imp_nrrsteps: assumes steps: "(s,t) \<in> (rstep R)^*" and C: "C \<noteq> \<box>" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> (nrrstep R)^*"
  using steps
proof (induct)
  case base show ?case by simp
next
  case (step t u)
  from rstep_ctxt_imp_nrrstep[OF step(2) C] step(3) show ?case by simp
qed

lemma nrrstep_mono: "rR \<subseteq> R \<Longrightarrow> nrrstep rR \<subseteq> nrrstep R"
unfolding nrrstep_def rstep_r_p_s_def by (force simp: Let_def)


lemma clean_term_Var_id[simp]: "clean_term F \<circ> Var = Var"
proof
  fix x show "(clean_term F \<circ> Var) x = Var x" unfolding o_def by simp
qed

lemma singleton_subst_restrict[simp]: "{x/s} |s {x} = {x/s}"
  unfolding subst_def subst_restrict_def by (rule ext) simp

lemma singleton_subst_map[simp]:
shows "f \<circ> {x/s} = (\<lambda>y. if x = y then f s else f (Var y))"
proof -
  have "(f \<circ> {x/s}) = (f \<circ> (\<lambda>y. if x = y then s else Var y))"
    unfolding subst_def by simp
  also have "\<dots> = (\<lambda>y. if x = y then f s else f(Var y))" unfolding o_def if_distrib by simp
  finally show ?thesis .
qed

lemma subst_restrict_vars[simp]:
shows "(((\<lambda>z. if z \<in> V then f z else g z)) |s V) = f |s V"
unfolding subst_restrict_def
proof (intro ext)
  fix x show "(if x \<in> V then if x \<in> V then f x else g x else Var x)
    = (if x \<in> V then f x else Var x)" by simp
qed

lemma subst_restrict_restrict[simp]:
assumes "V \<inter> W = {}"
shows "((\<lambda>z. if z \<in> V then f z else g z) |s W) = g |s W"
unfolding subst_restrict_def
proof (intro ext)
  fix x show "(if x \<in> W then if x \<in> V then f x else g x else Var x)
    = (if x \<in> W then g x else Var x)" using assms by auto
qed

lemma CC_O_distrib: "CC (R O S) \<subseteq> CC R O CC S"
proof
  fix s t
  assume "(s,t) \<in> CC (R O S)"
  thus "(s,t) \<in> CC R O CC S"
  proof(induct)
    case (ctxt_closureI l r C)
    then obtain m where "(l,m) \<in> R" and "(m,r) \<in> S" by auto
    hence "(C\<langle>l\<rangle>,C\<langle>m\<rangle>) \<in> CC R" and "(C\<langle>m\<rangle>,C\<langle>r\<rangle>) \<in> CC S" by auto
    thus ?case by auto
  qed
qed

lemma CC_rtrancl_distrib: "CC (R^*) \<subseteq> (CC R)^*"
proof
  fix s t
  assume "(s,t) \<in> CC (R^*)"
  thus "(s,t) \<in> (CC R)^*"
  proof(induct)
    case (ctxt_closureI l r C)
    thus ?case 
    proof (induct, simp)
      case (step y z)
      from step(2) have "(C\<langle>y\<rangle>,C\<langle>z\<rangle>) \<in> CC R" by auto
      with step(3) have "(C\<langle>l\<rangle>,C\<langle>z\<rangle>) \<in> (CC R)^* O CC R" by auto
      thus ?case by auto
    qed
  qed
qed    

lemma rstep_rstep: "rstep (rstep R) = rstep R"
proof -
  have "CC (CS (rstep R)) = rstep R" by (simp only: subst_closed_rstep_equal_rstep ctxt_closed_rstep_id)
  thus ?thesis unfolding rstep_CC_CS .
qed

lemma rstep_trancl_distrib: "rstep (R^+) \<subseteq> (rstep R)^+"
proof
  fix s t
  assume "(s,t) \<in> rstep (R^+)"
  thus "(s,t) \<in> (rstep R)^+"
  proof
    fix l r C \<sigma>
    assume lr: "(l,r) \<in> R^+" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    from lr have "(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> (rstep R)^+" 
    proof(induct)
      case (base r)
      thus ?case by auto
    next
      case (step r rr)
      from step(2) have "(C\<langle>r \<cdot> \<sigma>\<rangle>, C\<langle>rr \<cdot> \<sigma>\<rangle>) \<in> (rstep R)" by auto
      with step(3) show ?case by auto
    qed
    thus "(s,t) \<in> (rstep R)^+" unfolding s t .
  qed 
qed


definition instance_rule :: "('f,'v)rule \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where [code del]: "instance_rule lr st = (\<exists> \<sigma>. fst lr = fst st \<cdot> \<sigma> \<and> snd lr = snd st \<cdot> \<sigma>)"
(* instance rule has implementation in Substitution.thy *)

definition eq_rule_mod_vars :: "('f,'v)rule \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "eq_rule_mod_vars lr st \<equiv> instance_rule lr st \<and> instance_rule st lr"

notation eq_rule_mod_vars ("(_/ =v _)" [51,51] 50)

definition
  left_linear_trs :: "('f, 'v) trs \<Rightarrow> bool"
where
  "left_linear_trs R \<equiv> \<forall>(l, r)\<in>R. linear_term l"

lemma left_linear_mono: assumes "left_linear_trs S" and "R \<subseteq> S" shows "left_linear_trs R"
  using assms unfolding left_linear_trs_def by auto

lemma left_linear_map_trs[simp]:  "left_linear_trs (map_trs f R) = left_linear_trs R"
  unfolding left_linear_trs_def by (auto simp: map_trs.simps)

lemma left_linear_weak_match_rstep:
  assumes rstep: "(u, v) \<in> rstep R"
    and weak_match: "weak_match s u"
    and ll: "left_linear_trs R"
  shows "\<exists>t. (s, t) \<in> rstep R \<and> weak_match t v"
using weak_match 
proof (induct rule: rstep_induct[OF rstep])
  case (1 C sig l r)
  from 1(2) show ?case
  proof (induct C arbitrary: s)
    case (More f bef C aft s)
    let ?n = "Suc (length bef + length aft)" 
    let ?m = "length bef"
    from More(2) obtain ss where s: "s = Fun f ss" and lss: "?n = length ss" and wm: "(\<forall>i<length ss. weak_match (ss ! i) ((bef @ C\<langle>l \<cdot> sig\<rangle> # aft) ! i))"  by (cases s, auto)
    from lss wm[THEN spec, of ?m] have "weak_match (ss ! ?m) C\<langle>l \<cdot> sig\<rangle>" by auto
    from More(1)[OF this] obtain t where wmt:  "weak_match t C\<langle>r \<cdot> sig\<rangle>" and step: "(ss ! ?m,t) \<in> rstep R" by auto
    from lss have mss: "?m < length ss" by simp
    let ?tsi = "\<lambda> t. take ?m ss @ t # drop (Suc ?m) ss"
    let ?ts = "?tsi t"
    let ?ss = "?tsi (ss ! ?m)"
    from id_take_nth_drop[OF mss] 
    have lts: "length ?ts = ?n" using lss by auto
    show ?case
    proof (rule exI[of _ "Fun f ?ts"], intro conjI)
      have "weak_match (Fun f ?ts) (More f bef C aft)\<langle>r \<cdot> sig\<rangle> = 
        weak_match (Fun f ?ts) (Fun f (bef @ C\<langle>r \<cdot> sig\<rangle> # aft))" by simp
      also have "\<dots> = True" proof (simp only: weak_match.simps lts, simp, intro allI impI)
        fix i
        assume i: "i < ?n"
        show "weak_match (?ts ! i) ((bef @ C\<langle>r \<cdot> sig\<rangle> # aft) ! i)"
        proof (cases "i = ?m")
          case True
          have "weak_match (?ts ! i) ((bef @ C\<langle>r \<cdot> sig\<rangle> # aft) ! i) = weak_match t C\<langle>r \<cdot> sig\<rangle>"
            using True mss by (simp add: nth_append)
          thus ?thesis using wmt by simp
        next
          case False
          have eq: "?ts ! i = ss ! i \<and> (bef @ C\<langle>r \<cdot> sig\<rangle> # aft) ! i = (bef @ C\<langle>l \<cdot> sig\<rangle> # aft) ! i"
          proof (cases "i < ?m")
            case True
            thus ?thesis by (simp add: nth_append lss[symmetric])
          next
            case False
            with `i \<noteq> ?m` i have "\<exists> j. i = Suc (?m + j) \<and> j < length aft" by arith
            then obtain j where i: "i = Suc (?m + j)" and j: "j < length aft" by auto
            hence id: " (Suc (length bef + j) - min (Suc (length bef + length aft)) (length bef)) = Suc j" by simp
            from j show ?thesis by (simp add: nth_append i id lss[symmetric])
          qed
          thus ?thesis using wm[THEN spec, of i] i[unfolded lss] by (simp) 
        qed
      qed
      finally show "weak_match (Fun f ?ts) (More f bef C aft)\<langle>r \<cdot> sig\<rangle>" by simp
    next
      have "s = Fun f ?ss" unfolding s using id_take_nth_drop[OF mss, symmetric] by simp
      also have "\<dots> = (More f (take ?m ss) \<box> (drop (Suc ?m) ss))\<langle>(ss ! ?m)\<rangle>" (is "_ = ?C\<langle>_\<rangle>") by simp
      finally have s: "s = ?C\<langle>ss ! ?m\<rangle>" .
      have t: "Fun f ?ts = ?C\<langle>t\<rangle>" by simp
      from rstep.ctxt[OF step]
      show "(s, Fun f ?ts) \<in> rstep R"
        unfolding s t .
    qed
  next
    case (Hole s)
    from ll 1(1) have "linear_term l" unfolding left_linear_trs_def by auto
    from linear_weak_match[OF this Hole[simplified] refl] obtain \<tau> where 
      "s = l \<cdot> \<tau>" and "(\<forall> x \<in> vars_term l . weak_match (Var x \<cdot> \<tau>) (Var x \<cdot> sig))"
      by auto
    then obtain tau where s: "s = l \<cdot> tau" and wm: "(\<forall> x \<in> vars_term l . weak_match (tau x) (Var x \<cdot> sig))"
      by (auto)
    let ?delta = "(\<lambda> x. if x \<in> vars_term l then tau x else Var x \<cdot> sig)"
    show ?case
    proof (rule exI[of _ "r \<cdot> ?delta"], rule conjI)
      have "s = l \<cdot> (tau |s (vars_term l))" unfolding s by (rule coincidence_lemma)
      also have "\<dots> = l \<cdot> (?delta  |s (vars_term l))" by simp
      also have "\<dots> = l \<cdot> ?delta" by (rule coincidence_lemma[symmetric])
      finally have s: "s = l \<cdot> ?delta" .
      from 1(1) have step: "(l \<cdot> ?delta, r \<cdot> ?delta) \<in> rstep R" by auto
      thus "(s, r \<cdot> ?delta) \<in> rstep R" unfolding s .
    next
      have "weak_match (r \<cdot> ?delta) (r \<cdot> sig)"
      proof (induct r)
        case (Fun f ss)
        from this[unfolded set_conv_nth]
        show ?case by (force)
      next
        case (Var x)
        show ?case
        proof (cases "x \<in> vars_term l")
          case True
          with wm Var show ?thesis by simp
        next
          case False
          show ?thesis by (simp add: Var False weak_match_refl)
        qed
      qed
      thus "weak_match (r \<cdot> ?delta) (\<box> \<langle>(r \<cdot> sig)\<rangle>)" by simp
    qed
  qed
qed


fun S where
  "S R s t 0 = s"
| "S R s t (Suc i) = (SOME u. (S R s t i,u) \<in> rstep R \<and> weak_match u (t(Suc i)))"

lemma weak_match_SN:
  assumes wm: "weak_match s t"
    and ll: "left_linear_trs R"
    and SN: "SN_elt (rstep R) s"
  shows "SN_elt (rstep R) t"
proof
  fix f
  assume t0: "f 0 = t" and seq: "\<forall> i. (f i, f (Suc i)) \<in> rstep R"  
  let ?s = "S R s f"
  let ?P = "\<lambda>i u. (?s i, u) \<in> rstep R \<and> weak_match u (f (Suc i))"
  have "\<forall>i. (?s i, ?s (Suc i)) \<in> rstep R \<and> weak_match (?s (Suc i)) (f (Suc i))"
  proof
    fix i show "(?s i, ?s (Suc i)) \<in> rstep R \<and> weak_match (?s (Suc i)) (f (Suc i))"
    proof (induct i)
      case 0
      from seq have ini: "(f 0, f (Suc 0)) \<in> rstep R" by simp
      hence "(t, f (Suc 0)) \<in> rstep R" unfolding t0 .
      from someI_ex[OF left_linear_weak_match_rstep[OF this wm ll]]
      show ?case by simp
    next
      case (Suc i)
      hence IH1: "(?s i, ?s (Suc i)) \<in> rstep R"
        and IH2: "weak_match (?s (Suc i)) (f (Suc i))" by auto
      from seq have nxt: "(f (Suc i), f (Suc (Suc i))) \<in> rstep R" by simp
      from someI_ex[OF left_linear_weak_match_rstep[OF this IH2 ll]]
      have "\<exists>u. ?P (Suc i) u" by auto
      from someI_ex[OF this] 
      show ?case by simp
    qed
  qed
  moreover have "?s 0 = s" by simp
  ultimately have "\<not> SN_elt (rstep R) s" by best
  with SN show False by simp
qed

hide_const S


lemma left_linear_SN_elt_imp_clean_SN_elt:
  assumes ll: "left_linear_trs R"
  and SN: "SN_elt (rstep R) s"
  shows "SN_elt (rstep R) (clean_term F s)"
  by (rule weak_match_SN[OF clean_weak_match ll SN])  

lemma lhs_notin_NF_rstep: "(l, r) \<in> R \<Longrightarrow> l \<notin> NF (rstep R)" by auto

lemma NF_instance:
  assumes "(t \<cdot> \<sigma>) \<in> NF (rstep R)" shows "t \<in> NF (rstep R)"
using assms by auto

lemma NF_subterm:
  assumes "t \<in> NF (rstep R)" and "t \<unrhd> s"
  shows "s \<in> NF (rstep R)"
proof (rule ccontr)  
  assume "\<not> ?thesis"
  then obtain u rhs where "(s, u) \<in> rstep R" by auto
  from `t \<unrhd> s` obtain C where "t = C\<langle>s\<rangle>" by auto
  with `(s, u) \<in> rstep R` have "(t, C\<langle>u\<rangle>) \<in> rstep R" by auto
  hence "t \<notin> NF (rstep R)" by auto
  with assms show False by simp
qed

abbreviation
  lhss :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set"
where
  "lhss R \<equiv> fst ` R"

abbreviation
  rhss :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set"
where
  "rhss R \<equiv> snd ` R"


definition map_trs_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('g,'v)trs"
where "map_trs_wa fg R \<equiv> (\<lambda>(l,r). (map_term_wa fg l, map_term_wa fg r)) ` R"

lemma map_term_wa_compose: "map_term_wa gh (map_term_wa fg t) = map_term_wa (\<lambda> f n. gh (fg f n) n) t"
  by (induct t, auto)

lemma map_trs_wa_compose: "map_trs_wa gh (map_trs_wa fg R) = map_trs_wa (\<lambda> f n. gh (fg f n) n) R" (is "?L = map_trs_wa ?fgh R")
proof -
  have "map_trs_wa ?fgh R = {(map_term_wa ?fgh l, map_term_wa ?fgh r)| l r. (l,r) \<in> R}" unfolding map_trs_wa_def by auto
  also have "\<dots> = {(map_term_wa gh (map_term_wa fg l), map_term_wa gh (map_term_wa fg r)) | l r. (l,r) \<in> R}" unfolding map_term_wa_compose ..
  finally show ?thesis unfolding map_trs_wa_def by force
qed

lemma map_trs_wa_id: assumes id: "\<And> f n. fg f n = f"
  shows "map_trs_wa fg R = R"
  unfolding map_trs_wa_def map_term_wa_id[OF id] by auto


lemma map_trs_wa_rstep: assumes step:"(s,t) \<in> rstep R"
  shows "(map_term_wa fg s,map_term_wa fg t) \<in> rstep (map_trs_wa fg R)"
using step
proof (induct)
  case (IH C \<sigma> l r)
  show ?case unfolding map_trs_wa_def
    by (rule rstepI[where l = "map_term_wa fg l" and r = "map_term_wa fg r" and C = "map_ctxt_wa fg C"], auto simp: IH)
qed

definition map_rules_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f,'v)rules \<Rightarrow> ('g,'v)rules"
where "map_rules_wa fg R \<equiv> map (\<lambda>(l,r). (map_term_wa fg l, map_term_wa fg r)) R"

lemma map_rules_wa: "set (map_rules_wa fg R) = map_trs_wa fg (set R)"
  unfolding map_rules_wa_def map_trs_wa_def by auto

lemma rstep_ground: 
  assumes wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and ground: "ground s"
  and step: "(s,t) \<in> rstep R"
  shows "ground t"
using step ground
proof (induct)
  case (IH C \<sigma> l r)
  from wf_trs[OF IH(1)] IH(2)
  show ?case by auto
qed

lemma wf_terms_into_R: assumes step: "(s,t) \<in> rstep R" and s: "s \<in> wf_terms F" and t:"t \<in> wf_terms F" shows "(s,t) \<in> rstep ({(l,r). (l,r) \<in> R \<and> l \<in> wf_terms F \<and> r \<in> wf_terms F})"
using assms
proof (induct)
  case (IH C \<sigma> l r)
  from IH(2) have l: "l \<in> wf_terms F" by auto
  from IH(3) have r: "r \<in> wf_terms F" by auto
  from IH(1) l r show ?case
    by auto
qed

lemma rsteps_preserve_wf_terms: assumes F: "funas_trs R \<subseteq> F" and s: "s \<in> wf_terms F"
  and steps: "(s,t) \<in> (rstep R)^*" and wf: "wf_trs R"
  shows "t \<in> wf_terms F"
  using steps 
proof (induct)
  case base thus ?case by (rule s)
next
  case (step t u)
  show ?case by (rule rstep_preserves_wf_terms[OF F step(3) step(2) wf])
qed

definition compact :: "('f,'v)term set \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
  where "compact L R \<equiv> \<exists> F. finite F \<and> {t | s t. s \<in> L \<and> (s,t) \<in> (rstep R)^*} \<subseteq> wf_terms F"

definition locally_terminating :: "('f,'v)trs \<Rightarrow> bool"
  where "locally_terminating R \<equiv> \<forall> F. finite F \<longrightarrow> SN {(s,t). (s,t) \<in> rstep R \<and> s \<in> wf_terms F \<and> t \<in> wf_terms F}"

lemma locally_terminating_compact: 
  assumes lterm: "locally_terminating R"
  and compact: "compact L R"
  and t: "t \<in> L"
  shows "SN_elt (rstep R) t"
proof -
  from compact[unfolded compact_def] t obtain F where
    F: "finite F"
    and wf_terms: "{s. (t,s) \<in> (rstep R)^*} \<subseteq> wf_terms F" (is "?L \<subseteq> _")
    by auto
  from lterm[unfolded locally_terminating_def] F
  have SN: "SN {(s,t). (s,t) \<in> rstep R \<and> s \<in> wf_terms F \<and> t \<in> wf_terms F}" (is "SN ?R")
    by auto
  show "SN_elt (rstep R) t"
  proof
    fix f 
    assume t: "f 0 = t" and steps: "\<forall>i. (f i, f (Suc i)) \<in> rstep R" 
    {
      fix i
      have "f i \<in> ?L"
      proof (induct i)
        case 0
        thus ?case unfolding t by auto
      next
        case (Suc n)
        thus ?case using steps[THEN spec[of _ n]] by auto
      qed
      with wf_terms have "f i \<in> wf_terms F" by auto
    } note wf_terms = this
    {
      fix i
      have "(f i, f (Suc i)) \<in> ?R" 
        using wf_terms[of i] wf_terms[of "Suc i"] steps by auto
    }
    with SN show False unfolding SN_defs by auto
  qed    
qed

(* FIXME: move to Implementatino theory *)
subsection {* Output *}

fun
  shows_rule :: "('f::show, 'v::show) rule \<Rightarrow> shows"
where
  "shows_rule (l, r) = (shows l +@+ '' -> '' +#+ shows r)"

definition
  shows_rules :: "('f::show, 'v::show) rules \<Rightarrow> shows"
where
  "shows_rules trs \<equiv> shows_list_gen shows_rule [] [] [newline] [] trs +@+ shows_nl"

definition
  shows_p :: "('f::show, 'v::show) rules \<Rightarrow> shows"
where
  "shows_p P  \<equiv> ''dependency pairs:'' +#+ shows_nl +@+ shows_rules P"

definition
  shows_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows"
where
  "shows_trs R \<equiv> ''rewrite system:'' +#+ shows_nl +@+ shows_rules R"

definition
  shows_dpp :: "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows"
where
  "shows_dpp P R \<equiv> shows_p P +@+ shows_trs R"


end
